/*
 * This file is part of Cadmium.
 * Copyright (C) 2007-2010 Xavier Clerc.
 *
 * Cadmium is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * Cadmium is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package fr.x9c.cadmium.primitives.cadmium;

import java.io.InterruptedIOException;
import java.lang.reflect.Constructor;

import fr.x9c.cadmium.kernel.Block;
import fr.x9c.cadmium.kernel.CodeRunner;
import fr.x9c.cadmium.kernel.Fail;
import fr.x9c.cadmium.kernel.FalseExit;
import fr.x9c.cadmium.kernel.Primitive;
import fr.x9c.cadmium.kernel.PrimitiveProvider;
import fr.x9c.cadmium.kernel.Value;

/**
 * This class provides primitives related to {@link java.lang.reflect.Constructor}.
 *
 * @author <a href="mailto:cadmium@x9c.fr">Xavier Clerc</a>
 * @version 1.0
 * @since 1.0
 */
@PrimitiveProvider
public final class Constructors {

    /**
     * No instance of this class.
     */
    private Constructors() {
    } // end empty constructor

    /**
     * Executes {@link java.lang.reflect.Constructor#getClass()} method.
     * @param ctxt context
     * @param cstr object to execute method on
     * @return result of method execution
     * @throws Fail.Exception if Java method throws an exception
     */
    @Primitive
    public static Value cadmium_constr_get_declaring_class(final CodeRunner ctxt,
                                                           final Value cstr)
        throws Fail.Exception {
        try {
            final Constructor c = (Constructor) cstr.asBlock().asCustom();
            return Cadmium.createObject(c.getDeclaringClass());
        } catch (final Exception e) {
            Cadmium.fail(ctxt, e);
            return Value.UNIT; // never reached
        } // end try/catch
    } // end method 'cadmium_constr_get_declaring_class(CodeRunner, Value)'

    /**
     * Executes {@link java.lang.reflect.Constructor#getExceptionTypes()} method.
     * @param ctxt context
     * @param cstr object to execute method on
     * @return result of method execution
     * @throws Fail.Exception if Java method throws an exception
     */
    @Primitive
    public static Value cadmium_constr_get_exception_types(final CodeRunner ctxt,
                                                           final Value cstr)
        throws Fail.Exception {
        try {
            final Constructor c = (Constructor) cstr.asBlock().asCustom();
            return Cadmium.createObjectList(c.getExceptionTypes());
        } catch (final Exception e) {
            Cadmium.fail(ctxt, e);
            return Value.UNIT; // never reached
        } // end try/catch
    } // end method 'cadmium_constr_get_exception_types(CodeRunner, Value)'

    /**
     * Executes {@link java.lang.reflect.Constructor#getModifiers()} method.
     * @param ctxt context
     * @param cstr object to execute method on
     * @return result of method execution
     * @throws Fail.Exception if Java method throws an exception
     */
    @Primitive
    public static Value cadmium_constr_get_modifiers(final CodeRunner ctxt,
                                                     final Value cstr)
        throws Fail.Exception {
        try {
            final Constructor c = (Constructor) cstr.asBlock().asCustom();
            return Cadmium.createModifiersList(c.getModifiers());
        } catch (final Exception e) {
            Cadmium.fail(ctxt, e);
            return Value.UNIT; // never reached
        } // end try/catch
    } // end method 'cadmium_constr_get_modifiers(CodeRunner, Value)'

    /**
     * Executes {@link java.lang.reflect.Constructor#getParameterTypes()} method.
     * @param ctxt context
     * @param cstr object to execute method on
     * @return result of method execution
     * @throws Fail.Exception if Java method throws an exception
     */
    @Primitive
    public static Value cadmium_constr_get_parameter_types(final CodeRunner ctxt,
                                                           final Value cstr)
        throws Fail.Exception {
        try {
            final Constructor c = (Constructor) cstr.asBlock().asCustom();
            return Cadmium.encodeTypeList(c.getParameterTypes());
        } catch (final Exception e) {
            Cadmium.fail(ctxt, e);
            return Value.UNIT; // never reached
        } // end try/catch
    } // end method 'cadmium_constr_get_parameter_types(CodeRunner, Value)'

    /**
     * Executes {@link java.lang.reflect.Constructor#newInstance(java.lang.Object...)} method.
     * @param ctxt context
     * @param cstr object to execute method on
     * @param params parameter for method execution
     * @return result of method execution
     * @throws Fail.Exception if Java method throws an exception
     */
    @Primitive
    public static Value cadmium_constr_new_instance(final CodeRunner ctxt,
                                                    final Value cstr,
                                                    final Value params)
        throws Fail.Exception, FalseExit {
        try {
            final Constructor c = (Constructor) cstr.asBlock().asCustom();
            final Object[] args = Cadmium.decodeObjectList(params);
            return Cadmium.createObject(c.newInstance(args));
        } catch (final Exception e) {
            final Throwable cause = e.getCause();
            if ((cause instanceof InterruptedException)
                || (cause instanceof InterruptedIOException)) {
                final FalseExit fe =
                    FalseExit.createFromContext(ctxt.getContext());
                fe.fillInStackTrace();
                throw fe;
            } // end if
            Cadmium.fail(ctxt, e);
            return Value.UNIT; // never reached
        } // end try/catch
    } // end method 'cadmium_constr_new_instance(CodeRunner, Value, Value)'

    /**
     * Casts an object to a constructor.
     * @param ctxt context
     * @param obj object to cast
     * @return object as constructor
     * @throws Fail.Exception if cast fails
     */
    @Primitive
    public static Value cadmium_constr_as_constr(final CodeRunner ctxt,
                                                 final Value obj)
        throws Fail.Exception {
        try {
            final Block objBlock = obj.asBlock();
            objBlock.setCustom((Constructor) objBlock.asCustom());
            return obj;
        } catch (final Exception e) {
            Cadmium.fail(ctxt, e);
            return Value.UNIT; // never reached
        } // end try/catch
    } // end method 'cadmium_constr_as_constr(CodeRunner, Value)'

} // end class 'Constructors'
