/*
 * This file is part of Cadmium.
 * Copyright (C) 2007-2010 Xavier Clerc.
 *
 * Cadmium is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * Cadmium is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package fr.x9c.cadmium.primitives.cadmiumjdbc;

import java.sql.Ref;

import fr.x9c.cadmium.kernel.Block;
import fr.x9c.cadmium.kernel.CodeRunner;
import fr.x9c.cadmium.kernel.Fail;
import fr.x9c.cadmium.kernel.Primitive;
import fr.x9c.cadmium.kernel.PrimitiveProvider;
import fr.x9c.cadmium.kernel.Value;
import fr.x9c.cadmium.primitives.cadmium.Cadmium;

/**
 * This class provides the primitives related to the <tt>java.sql.Ref</tt> class.
 *
 * @author <a href="mailto:cadmium@x9c.fr">Xavier Clerc</a>
 * @version 1.1
 * @since 1.1
 */
@PrimitiveProvider
public final class Refs {

    /**
     * No instance of this class.
     */
    private Refs() {
    } // end empty constructor

    /**
     * Returns the fully-qualified SQL name of the structured type that the
     * passed reference points to.
     * @param ctxt context
     * @param ref ref
     * @return the fully-qualified SQL name of the structured type that the
     *         passed reference points to
     * @throws Fail.Exception if any error occurs
     */
    @Primitive
    public static Value cadmiumjdbc_ref_get_base_type_name(final CodeRunner ctxt,
                                                           final Value ref)
        throws Fail.Exception {
        try {
            final Ref r = (Ref) ref.asBlock().asCustom();
            final Block res = Block.createString(r.getBaseTypeName());
            return Value.createFromBlock(res);
        } catch (final Exception e) {
            Cadmium.fail(ctxt, e);
            return Value.UNIT; // never reached
        } // end try/catch
    } // end method 'cadmiumjdbc_ref_get_base_type_name(CodeRunner, Value)'

    /**
     * Returns the structured type instance that the passed reference points to.
     * @param ctxt context
     * @param ref ref
     * @return the structured type instance that the passed reference points to
     * @throws Fail.Exception if any error occurs
     */
    @Primitive
    public static Value cadmiumjdbc_ref_get_object(final CodeRunner ctxt,
                                                   final Value ref)
        throws Fail.Exception {
        try {
            final Ref r = (Ref) ref.asBlock().asCustom();
            return Cadmium.createObject(r.getObject());
        } catch (final Exception e) {
            Cadmium.fail(ctxt, e);
            return Value.UNIT; // never reached
        } // end try/catch
    } // end method 'cadmiumjdbc_ref_get_object(CodeRunner, Value)'

    /**
     * Changes the structured type instance that the passed reference points to.
     * @param ctxt context
     * @param ref ref
     * @param o new reference value
     * @return <i>unit</i>
     * @throws Fail.Exception if any error occurs
     */
    @Primitive
    public static Value cadmiumjdbc_ref_set_object(final CodeRunner ctxt,
                                                   final Value ref,
                                                   final Value o)
        throws Fail.Exception {
        try {
            final Ref r = (Ref) ref.asBlock().asCustom();
            r.setObject(o.asBlock().asCustom());
            return Value.UNIT;
        } catch (final Exception e) {
            Cadmium.fail(ctxt, e);
            return Value.UNIT; // never reached
        } // end try/catch
    } // end method 'cadmiumjdbc_ref_set_object(CodeRunner, Value, Value)'

} // end class 'Refs'
