/*
 * This file is part of Cadmium.
 * Copyright (C) 2007-2010 Xavier Clerc.
 *
 * Cadmium is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * Cadmium is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package fr.x9c.cadmium.primitives.cadmiumjdbc;

import java.sql.Struct;

import fr.x9c.cadmium.kernel.Block;
import fr.x9c.cadmium.kernel.CodeRunner;
import fr.x9c.cadmium.kernel.Fail;
import fr.x9c.cadmium.kernel.Primitive;
import fr.x9c.cadmium.kernel.PrimitiveProvider;
import fr.x9c.cadmium.kernel.Value;
import fr.x9c.cadmium.primitives.cadmium.Cadmium;

/**
 * This class provides the primitives related to the <tt>java.sql.Struct</tt> class.
 *
 * @author <a href="mailto:cadmium@x9c.fr">Xavier Clerc</a>
 * @version 1.1
 * @since 1.1
 */
@PrimitiveProvider
public final class Structs {

    /**
     * No instance of this class.
     */
    private Structs() {
    } // end empty constructor

    /**
     * Returns the ordred values of the attributes of the passed structured type.
     * @param ctxt context
     * @param struct structure
     * @return the ordred values of the attributes of the passed structured type
     * @throws Fail.Exception if any error occurs
     */
    @Primitive
    public static Value cadmiumjdbc_struct_get_attributes(final CodeRunner ctxt,
                                                          final Value struct)
        throws Fail.Exception {
        try {
            final Struct s = (Struct) struct.asBlock().asCustom();
            final Object[] attrs = s.getAttributes();
            final int len = attrs.length;
            Value res = Value.EMPTY_LIST;
            for (int i = len - 1; i >= 0; i--) {
                final Value v = Cadmium.createObject(attrs[i]);
                final Block cons = Block.createBlock(Block.TAG_CONS, v, res);
                res = Value.createFromBlock(cons);
            } // end for
            return res;
        } catch (final Exception e) {
            Cadmium.fail(ctxt, e);
            return Value.UNIT; // never reached
        } // end try/catch
    } // end method 'cadmiumjdbc_struct_get_attributes(CodeRunner, Value)'

    /**
     * Returns the SQL type name of the passed structured type.
     * @param ctxt context
     * @param struct structure
     * @return the SQL type name of the passed structured type
     * @throws Fail.Exception if any error occurs
     */
    @Primitive
    public static Value cadmiumjdbc_struct_get_sql_type_name(final CodeRunner ctxt,
                                                             final Value struct)
        throws Fail.Exception {
        try {
            final Struct s = (Struct) struct.asBlock().asCustom();
            final Block res = Block.createString(s.getSQLTypeName());
            return Value.createFromBlock(res);
        } catch (final Exception e) {
            Cadmium.fail(ctxt, e);
            return Value.UNIT; // never reached
        } // end try/catch
    } // end method 'cadmiumjdbc_struct_get_sql_type_name(CodeRunner, Value)'

} // end class 'Structs'
