/*
 * This file is part of Cadmium.
 * Copyright (C) 2007-2010 Xavier Clerc.
 *
 * Cadmium is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * Cadmium is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package fr.x9c.cadmium.primitives.cadmiumservlet;

import java.io.InputStream;
import java.io.OutputStream;
import java.util.Enumeration;
import java.util.Locale;
import java.util.Map;
import java.util.Set;

import javax.servlet.ServletContext;
import javax.servlet.http.Cookie;
import javax.servlet.http.HttpServlet;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import javax.servlet.http.HttpSession;

import fr.x9c.cadmium.kernel.Block;
import fr.x9c.cadmium.kernel.Channel;
import fr.x9c.cadmium.kernel.CodeRunner;
import fr.x9c.cadmium.kernel.Custom;
import fr.x9c.cadmium.kernel.Fail;
import fr.x9c.cadmium.kernel.Primitive;
import fr.x9c.cadmium.kernel.PrimitiveProvider;
import fr.x9c.cadmium.kernel.Value;
import fr.x9c.cadmium.primitives.cadmium.Cadmium;
import fr.x9c.cadmium.support.Helper;

/**
 * This class provides the primitives allowing use of <tt>javax.servlet</tt>
 * and <tt>javax.servlet.http</tt> classes from Cadmium.
 *
 * @author <a href="mailto:cadmium@x9c.fr">Xavier Clerc</a>
 * @version 1.1
 * @since 1.1
 */
@PrimitiveProvider
public final class Servlets {

    /**
     * No instance of this class.
     */
    private Servlets() {
    } // end empty constructor

    /**
     * Constructs a cookie.
     * @param ctxt context
     * @param name cookie name
     * @param value cookie value
     * @return new cookie
     * @throws Fail.Exception if <i>name</i> contains illegal characters
     * @throws Fail.Exception if any error occurs
     */
    @Primitive
    public static Value cadmiumservlet_cookie_make(final CodeRunner ctxt,
                                                   final Value name,
                                                   final Value value)
        throws Fail.Exception {
        try {
            final Cookie res = new Cookie(name.asBlock().asString(),
                                          value.asBlock().asString());
            return Cadmium.createObject(res);
        } catch (final IllegalArgumentException iae) {
            Fail.invalidArgument("Servlet.Cookie.make");
            return Value.UNIT; // never reached
        } catch (final Exception e) {
            Cadmium.fail(ctxt, e);
            return Value.UNIT; // never reached
        } // end try/catch
    } // end method 'cadmiumservlet_cookie_make(CodeRunner, Value, Value)'

    /**
     * Clones a cookie.
     * @param ctxt context
     * @param cookie cookie
     * @return a copy of the passed cookie
     * @throws Fail.Exception if any error occurs
     */
    @Primitive
    public static Value cadmiumservlet_cookie_clone(final CodeRunner ctxt,
                                                    final Value cookie)
        throws Fail.Exception {
        try {
            final Cookie c = (Cookie) cookie.asBlock().asCustom();
            return Cadmium.createObject((Cookie) c.clone());
        } catch (final Exception e) {
            Cadmium.fail(ctxt, e);
            return Value.UNIT; // never reached
        } // end try/catch
    } // end method 'cadmiumservlet_cookie_clone(CodeRunner, Value)'

    /**
     * Returns the comment of the passed cookie.
     * @param ctxt context
     * @param cookie cookie
     * @return the comment of the passed cookie
     * @throws Fail.Exception if any error occurs
     */
    @Primitive
    public static Value cadmiumservlet_cookie_get_comment(final CodeRunner ctxt,
                                                          final Value cookie)
        throws Fail.Exception {
        try {
            final Cookie c = (Cookie) cookie.asBlock().asCustom();
            final String res = c.getComment();
            return Value.createFromBlock(Block.createString(res));
        } catch (final Exception e) {
            Cadmium.fail(ctxt, e);
            return Value.UNIT; // never reached
        } // end try/catch
    } // end method 'cadmiumservlet_cookie_get_comment(CodeRunner, Value)'

    /**
     * Returns the domain of the passed cookie.
     * @param ctxt context
     * @param cookie cookie
     * @return the domain of the passed cookie
     * @throws Fail.Exception if any error occurs
     */
    @Primitive
    public static Value cadmiumservlet_cookie_get_domain(final CodeRunner ctxt,
                                                         final Value cookie)
        throws Fail.Exception {
        try {
            final Cookie c = (Cookie) cookie.asBlock().asCustom();
            final String res = c.getDomain();
            return Value.createFromBlock(Block.createString(res));
        } catch (final Exception e) {
            Cadmium.fail(ctxt, e);
            return Value.UNIT; // never reached
        } // end try/catch
    } // end method 'cadmiumservlet_cookie_get_domain(CodeRunner, Value)'

    /**
     * Returns the maximum age of the passed cookie.
     * @param ctxt context
     * @param cookie cookie
     * @return the maximum age of the passed cookie
     * @throws Fail.Exception if any error occurs
     */
    @Primitive
    public static Value cadmiumservlet_cookie_get_max_age(final CodeRunner ctxt,
                                                          final Value cookie)
        throws Fail.Exception {
        try {
            final Cookie c = (Cookie) cookie.asBlock().asCustom();
            final int res = c.getMaxAge();
            return Helper.createInt32(res);
        } catch (final Exception e) {
            Cadmium.fail(ctxt, e);
            return Value.UNIT; // never reached
        } // end try/catch
    } // end method 'cadmiumservlet_cookie_get_max_age(CodeRunner, Value)'

    /**
     * Returns the name of the passed cookie.
     * @param ctxt context
     * @param cookie cookie
     * @return the name of the passed cookie
     * @throws Fail.Exception if any error occurs
     */
    @Primitive
    public static Value cadmiumservlet_cookie_get_name(final CodeRunner ctxt,
                                                       final Value cookie)
        throws Fail.Exception {
        try {
            final Cookie c = (Cookie) cookie.asBlock().asCustom();
            final String res = c.getName();
            return Value.createFromBlock(Block.createString(res));
        } catch (final Exception e) {
            Cadmium.fail(ctxt, e);
            return Value.UNIT; // never reached
        } // end try/catch
    } // end method 'cadmiumservlet_cookie_get_name(CodeRunner, Value)'

    /**
     * Returns the path of the passed cookie.
     * @param ctxt context
     * @param cookie cookie
     * @return the path of the passed cookie
     * @throws Fail.Exception if any error occurs
     */
    @Primitive
    public static Value cadmiumservlet_cookie_get_path(final CodeRunner ctxt,
                                                       final Value cookie)
        throws Fail.Exception {
        try {
            final Cookie c = (Cookie) cookie.asBlock().asCustom();
            final String res = c.getPath();
            return Value.createFromBlock(Block.createString(res));
        } catch (final Exception e) {
            Cadmium.fail(ctxt, e);
            return Value.UNIT; // never reached
        } // end try/catch
    } // end method 'cadmiumservlet_cookie_get_path(CodeRunner, Value)'

    /**
     * Tests whether the cookie is sent using a secure protocol.
     * @param ctxt context
     * @param cookie cookie
     * @return <tt>true</tt> if the cookie is sent using a secure protocol,
     *         <tt>false</tt> otherwise
     * @throws Fail.Exception if any error occurs
     */
    @Primitive
    public static Value cadmiumservlet_cookie_get_secure(final CodeRunner ctxt,
                                                         final Value cookie)
        throws Fail.Exception {
        try {
            final Cookie c = (Cookie) cookie.asBlock().asCustom();
            final boolean res = c.getSecure();
            return res ? Value.TRUE : Value.FALSE;
        } catch (final Exception e) {
            Cadmium.fail(ctxt, e);
            return Value.UNIT; // never reached
        } // end try/catch
    } // end method 'cadmiumservlet_cookie_get_secure(CodeRunner, Value)'

    /**
     * Returns the value of the passed cookie.
     * @param ctxt context
     * @param cookie cookie
     * @return the value of the passed cookie
     * @throws Fail.Exception if any error occurs
     */
    @Primitive
    public static Value cadmiumservlet_cookie_get_value(final CodeRunner ctxt,
                                                        final Value cookie)
        throws Fail.Exception {
        try {
            final Cookie c = (Cookie) cookie.asBlock().asCustom();
            final String res = c.getValue();
            return Value.createFromBlock(Block.createString(res));
        } catch (final Exception e) {
            Cadmium.fail(ctxt, e);
            return Value.UNIT; // never reached
        } // end try/catch
    } // end method 'cadmiumservlet_cookie_get_value(CodeRunner, Value)'

    /**
     * Returns the version of the passed cookie.
     * @param ctxt context
     * @param cookie cookie
     * @return the version of the passed cookie
     * @throws Fail.Exception if any error occurs
     */
    @Primitive
    public static Value cadmiumservlet_cookie_get_version(final CodeRunner ctxt,
                                                          final Value cookie)
        throws Fail.Exception {
        try {
            final Cookie c = (Cookie) cookie.asBlock().asCustom();
            final int res = c.getVersion();
            return Helper.createInt32(res);
        } catch (final Exception e) {
            Cadmium.fail(ctxt, e);
            return Value.UNIT; // never reached
        } // end try/catch
    } // end method 'cadmiumservlet_cookie_get_version(CodeRunner, Value)'

    /**
     * Sets the comment of the passed cookie.
     * @param ctxt context
     * @param cookie cookie
     * @param value new comment
     * @return <i>unit</i>
     * @throws Fail.Exception if any error occurs
     */
    @Primitive
    public static Value cadmiumservlet_cookie_set_comment(final CodeRunner ctxt,
                                                          final Value cookie,
                                                          final Value value)
        throws Fail.Exception {
        try {
            final Cookie c = (Cookie) cookie.asBlock().asCustom();
            c.setComment(value.asBlock().asString());
            return Value.UNIT;
        } catch (final Exception e) {
            Cadmium.fail(ctxt, e);
            return Value.UNIT; // never reached
        } // end try/catch
    } // end method 'cadmiumservlet_cookie_set_comment(CodeRunner, Value, Value)'

    /**
     * Sets the domain of the passed cookie.
     * @param ctxt context
     * @param cookie cookie
     * @param value new domain
     * @return <i>unit</i>
     * @throws Fail.Exception if any error occurs
     */
    @Primitive
    public static Value cadmiumservlet_cookie_set_domain(final CodeRunner ctxt,
                                                         final Value cookie,
                                                         final Value value)
        throws Fail.Exception {
        try {
            final Cookie c = (Cookie) cookie.asBlock().asCustom();
            c.setDomain(value.asBlock().asString());
            return Value.UNIT;
        } catch (final Exception e) {
            Cadmium.fail(ctxt, e);
            return Value.UNIT; // never reached
        } // end try/catch
    } // end method 'cadmiumservlet_cookie_set_domain(CodeRunner, Value, Value)'

    /**
     * Sets the maximum age of the passed cookie.
     * @param ctxt context
     * @param cookie cookie
     * @param value new maximum age
     * @return <i>unit</i>
     * @throws Fail.Exception if any error occurs
     */
    @Primitive
    public static Value cadmiumservlet_cookie_set_max_age(final CodeRunner ctxt,
                                                          final Value cookie,
                                                          final Value value)
        throws Fail.Exception {
        try {
            final Cookie c = (Cookie) cookie.asBlock().asCustom();
            c.setMaxAge(value.asBlock().asInt32());
            return Value.UNIT;
        } catch (final Exception e) {
            Cadmium.fail(ctxt, e);
            return Value.UNIT; // never reached
        } // end try/catch
    } // end method 'cadmiumservlet_cookie_set_max_age(CodeRunner, Value, Value)'

    /**
     * Sets the path of the passed cookie.
     * @param ctxt context
     * @param cookie cookie
     * @param value new path
     * @return <i>unit</i>
     * @throws Fail.Exception if any error occurs
     */
    @Primitive
    public static Value cadmiumservlet_cookie_set_path(final CodeRunner ctxt,
                                                       final Value cookie,
                                                       final Value value)
        throws Fail.Exception {
        try {
            final Cookie c = (Cookie) cookie.asBlock().asCustom();
            c.setPath(value.asBlock().asString());
            return Value.UNIT;
        } catch (final Exception e) {
            Cadmium.fail(ctxt, e);
            return Value.UNIT; // never reached
        } // end try/catch
    } // end method 'cadmiumservlet_cookie_set_path(CodeRunner, Value, Value)'

    /**
     * Sets whether the passed cookie should be sent using a secure protocol.
     * @param ctxt context
     * @param cookie cookie
     * @param value whether the cookie should be sent using a secure protocol
     * @return <i>unit</i>
     * @throws Fail.Exception if any error occurs
     */
    @Primitive
    public static Value cadmiumservlet_cookie_set_secure(final CodeRunner ctxt,
                                                         final Value cookie,
                                                         final Value value)
        throws Fail.Exception {
        try {
            final Cookie c = (Cookie) cookie.asBlock().asCustom();
            c.setSecure(value == Value.TRUE);
            return Value.UNIT;
        } catch (final Exception e) {
            Cadmium.fail(ctxt, e);
            return Value.UNIT; // never reached
        } // end try/catch
    } // end method 'cadmiumservlet_cookie_set_secure(CodeRunner, Value, Value)'

    /**
     * Sets the value of the passed cookie.
     * @param ctxt context
     * @param cookie cookie
     * @param value new value
     * @return <i>unit</i>
     * @throws Fail.Exception if any error occurs
     */
    @Primitive
    public static Value cadmiumservlet_cookie_set_value(final CodeRunner ctxt,
                                                        final Value cookie,
                                                        final Value value)
        throws Fail.Exception {
        try {
            final Cookie c = (Cookie) cookie.asBlock().asCustom();
            c.setValue(value.asBlock().asString());
            return Value.UNIT;
        } catch (final Exception e) {
            Cadmium.fail(ctxt, e);
            return Value.UNIT; // never reached
        } // end try/catch
    } // end method 'cadmiumservlet_cookie_set_value(CodeRunner, Value, Value)'

    /**
     * Sets the version of the passed cookie.
     * @param ctxt context
     * @param cookie cookie
     * @param value new version
     * @return <i>unit</i>
     * @throws Fail.Exception if any error occurs
     */
    @Primitive
    public static Value cadmiumservlet_cookie_set_version(final CodeRunner ctxt,
                                                          final Value cookie,
                                                          final Value value)
        throws Fail.Exception {
        try {
            final Cookie c = (Cookie) cookie.asBlock().asCustom();
            c.setVersion(value.asBlock().asInt32());
            return Value.UNIT;
        } catch (final Exception e) {
            Cadmium.fail(ctxt, e);
            return Value.UNIT; // never reached
        } // end try/catch
    } // end method 'cadmiumservlet_cookie_set_version(CodeRunner, Value, Value)'

    /**
     * Returns an attribute of the passed context.
     * @param ctxt context
     * @param context servlet context
     * @param name attribute name
     * @return the value of the attribute whose name is passed
     * @throws Fail.Exception if no such attribute exists
     * @throws Fail.Exception if any error occurs
     */
    @Primitive
    public static Value cadmiumservlet_context_get_attribute(final CodeRunner ctxt,
                                                             final Value context,
                                                             final Value name)
        throws Fail.Exception {
        try {
            final ServletContext sc =
                (ServletContext) context.asBlock().asCustom();
            final Object res = sc.getAttribute(name.asBlock().asString());
            if (res == null) {
                Fail.raiseNotFound();
            } // end if
            return Cadmium.createObject(res);
        } catch (final Fail.Exception fe) {
            throw fe;
        } catch (final Exception e) {
            Cadmium.fail(ctxt, e);
            return Value.UNIT; // never reached
        } // end try/catch
    } // end method 'cadmiumservlet_context_get_attribute(CodeRunner, Value, Value)'

    /**
     * Returns the list of attribute names.
     * @param ctxt context
     * @param context servlet context
     * @return the list of attribute names
     * @throws Fail.Exception if any error occurs
     */
    @Primitive
    public static Value cadmiumservlet_context_get_attribute_names(final CodeRunner ctxt,
                                                                   final Value context)
        throws Fail.Exception {
        try {
            final ServletContext sc =
                (ServletContext) context.asBlock().asCustom();
            final Enumeration e = sc.getAttributeNames();
            Value res = Value.EMPTY_LIST;
            while (e.hasMoreElements()) {
                final String s = (String) e.nextElement();
                final Block b = Block.createBlock(Block.TAG_CONS,
                                                  Value.createFromBlock(Block.createString(s)),
                                                  res);
                res = Value.createFromBlock(b);
            } // end while
            return res;
        } catch (final Exception e) {
            Cadmium.fail(ctxt, e);
            return Value.UNIT; // never reached
        } // end try/catch
    } // end method 'cadmiumservlet_context_get_attribute_names(CodeRunner, Value)'

    /**
     * Returns the context for the passed path.
     * @param ctxt context
     * @param context servlet context
     * @param path path
     * @return the context for the passed path.
     * @throws Fail.Exception if no such context exists
     * @throws Fail.Exception if any error occurs
     */
    @Primitive
    public static Value cadmiumservlet_context_get_context(final CodeRunner ctxt,
                                                           final Value context,
                                                           final Value path)
        throws Fail.Exception {
        try {
            final ServletContext sc =
                (ServletContext) context.asBlock().asCustom();
            final ServletContext res = sc.getContext(path.asBlock().asString());
            if (res == null) {
                Fail.raiseNotFound();
            } // end if
            return Cadmium.createObject(res);
        } catch (final Fail.Exception fe) {
            throw fe;
        } catch (final Exception e) {
            Cadmium.fail(ctxt, e);
            return Value.UNIT; // never reached
        } // end try/catch
    } // end method 'cadmiumservlet_context_get_context(CodeRunner, Value, Value)'

    /**
     * Returns the context path.
     * @param ctxt context
     * @param context servlet context
     * @return the context path
     * @throws Fail.Exception if any error occurs
     */
    @Primitive
    public static Value cadmiumservlet_context_get_context_path(final CodeRunner ctxt,
                                                                final Value context)
        throws Fail.Exception {
        try {
            final ServletContext sc =
                (ServletContext) context.asBlock().asCustom();
            final String res = sc.getContextPath();
            return Value.createFromBlock(Block.createString(res));
        } catch (final Exception e) {
            Cadmium.fail(ctxt, e);
            return Value.UNIT; // never reached
        } // end try/catch
    } // end method 'cadmiumservlet_context_get_context_path(CodeRunner, Value)'

    /**
     * Returns the value of an init parameter.
     * @param ctxt context
     * @param context servlet context
     * @param name init parameter name
     * @return the value of the init parameter whose name is passed
     * @throws Fail.Exception if no such parameter exists
     * @throws Fail.Exception if any error occurs
     */
    @Primitive
    public static Value cadmiumservlet_context_get_init_parameter(final CodeRunner ctxt,
                                                                  final Value context,
                                                                  final Value name)
        throws Fail.Exception {
        try {
            final ServletContext sc =
                (ServletContext) context.asBlock().asCustom();
            final String res = sc.getInitParameter(name.asBlock().asString());
            if (res == null) {
                Fail.raiseNotFound();
            } // end if
            return Value.createFromBlock(Block.createString(res));
        } catch (final Fail.Exception fe) {
            throw fe;
        } catch (final Exception e) {
            Cadmium.fail(ctxt, e);
            return Value.UNIT; // never reached
        } // end try/catch
    } // end method 'cadmiumservlet_context_get_init_parameter(CodeRunner, Value, Value)'

    /**
     * Returns the list of the parameter names.
     * @param ctxt context
     * @param context servlet context
     * @return the list of the parameter names
     * @throws Fail.Exception if any error occurs
     */
    @Primitive
    public static Value cadmiumservlet_context_get_init_parameter_names(final CodeRunner ctxt,
                                                                        final Value context)
        throws Fail.Exception {
        try {
            final ServletContext sc =
                (ServletContext) context.asBlock().asCustom();
            final Enumeration e = sc.getInitParameterNames();
            Value res = Value.EMPTY_LIST;
            while (e.hasMoreElements()) {
                final String s = (String) e.nextElement();
                final Block b = Block.createBlock(Block.TAG_CONS,
                                                  Value.createFromBlock(Block.createString(s)),
                                                  res);
                res = Value.createFromBlock(b);
            } // end while
            return res;
        } catch (final Exception e) {
            Cadmium.fail(ctxt, e);
            return Value.UNIT; // never reached
        } // end try/catch
    } // end method 'cadmiumservlet_context_get_init_parameter_names(CodeRunner, Value)'

    /**
     * Returns the major version of the Servlet API of the container.
     * @param ctxt context
     * @param context servlet context
     * @return the major version of the Servlet API of the container
     * @throws Fail.Exception if any error occurs
     */
    @Primitive
    public static Value cadmiumservlet_context_get_major_version(final CodeRunner ctxt,
                                                                 final Value context)
        throws Fail.Exception {
        try {
            final ServletContext sc =
                (ServletContext) context.asBlock().asCustom();
            final int res = sc.getMajorVersion();
            return Helper.createInt32(res);
        } catch (final Exception e) {
            Cadmium.fail(ctxt, e);
            return Value.UNIT; // never reached
        } // end try/catch
    } // end method 'cadmiumservlet_context_get_major_version(CodeRunner, Value)'

    /**
     * Returns the MIME type of the passed file.
     * @param ctxt context
     * @param context servlet context
     * @param file file
     * @return the MIME type of the passed file
     * @throws Fail.Exception if the MIME type of the passed file is unknown
     * @throws Fail.Exception if any error occurs
     */
    @Primitive
    public static Value cadmiumservlet_context_get_mime_type(final CodeRunner ctxt,
                                                             final Value context,
                                                             final Value file)
        throws Fail.Exception {
        try {
            final ServletContext sc =
                (ServletContext) context.asBlock().asCustom();
            final String res = sc.getMimeType(file.asBlock().asString());
            if (res == null) {
                Fail.raiseNotFound();
            } // end if
            return Value.createFromBlock(Block.createString(res));
        } catch (final Fail.Exception fe) {
            throw fe;
        } catch (final Exception e) {
            Cadmium.fail(ctxt, e);
            return Value.UNIT; // never reached
        } // end try/catch
    } // end method 'cadmiumservlet_context_get_mime_type(CodeRunner, Value, Value)'

    /**
     * Returns the minor version of the Servlet API of the container.
     * @param ctxt context
     * @param context servlet context
     * @return the minor version of the Servlet API of the container
     * @throws Fail.Exception if any error occurs
     */
    @Primitive
    public static Value cadmiumservlet_context_get_minor_version(final CodeRunner ctxt,
                                                                 final Value context)
        throws Fail.Exception {
        try {
            final ServletContext sc =
                (ServletContext) context.asBlock().asCustom();
            final int res = sc.getMinorVersion();
            return Helper.createInt32(res);
        } catch (final Exception e) {
            Cadmium.fail(ctxt, e);
            return Value.UNIT; // never reached
        } // end try/catch
    } // end method 'cadmiumservlet_context_get_minor_version(CodeRunner, Value)'

    /**
     * Returns the real path for the passed virtual path.
     * @param ctxt context
     * @param context servlet context
     * @param path virtual path
     * @return the real path for the passed virtual path
     * @throws Fail.Exception if path translation fails
     * @throws Fail.Exception if any error occurs
     */
    @Primitive
    public static Value cadmiumservlet_context_get_real_path(final CodeRunner ctxt,
                                                             final Value context,
                                                             final Value path)
        throws Fail.Exception {
        try {
            final ServletContext sc =
                (ServletContext) context.asBlock().asCustom();
            final String res = sc.getRealPath(path.asBlock().asString());
            if (res == null) {
                Fail.invalidArgument("Servlet.Context.get_real_path");
            } // end if
            return Value.createFromBlock(Block.createString(res));
        } catch (final Fail.Exception fe) {
            throw fe;
        } catch (final Exception e) {
            Cadmium.fail(ctxt, e);
            return Value.UNIT; // never reached
        } // end try/catch
    } // end method 'cadmiumservlet_context_get_real_path(CodeRunner, Value, Value)'

    /**
     * Returns the URL for the resource at passed path.
     * @param ctxt context
     * @param context servlet context
     * @param path resource path
     * @return the URL for the resource at passed path
     * @throws Fail.Exception if no such resource exist
     * @throws Fail.Exception if any error occurs
     */
    @Primitive
    public static Value cadmiumservlet_context_get_resource_as_url(final CodeRunner ctxt,
                                                                   final Value context,
                                                                   final Value path)
        throws Fail.Exception {
        try {
            final ServletContext sc =
                (ServletContext) context.asBlock().asCustom();
            final String res =
                sc.getResource(path.asBlock().asString()).toString();
            if (res == null) {
                Fail.invalidArgument("Servlet.Context.get_resource_as_url");
            } // end if
            return Value.createFromBlock(Block.createString(res));
        } catch (final Fail.Exception fe) {
            throw fe;
        } catch (final Exception e) {
            Cadmium.fail(ctxt, e);
            return Value.UNIT; // never reached
        } // end try/catch
    } // end method 'cadmiumservlet_context_get_resource_as_url(CodeRunner, Value, Value)'

    /**
     * Returns the channel for the resource at passed path.
     * @param ctxt context
     * @param context servlet context
     * @param path resource path
     * @return the channel for the resource at passed path
     * @throws Fail.Exception if no such resource exist
     * @throws Fail.Exception if any error occurs
     */
    @Primitive
    public static Value cadmiumservlet_context_get_resource_as_stream(final CodeRunner ctxt,
                                                                      final Value context,
                                                                      final Value path)
        throws Fail.Exception {
        try {
            final ServletContext sc =
                (ServletContext) context.asBlock().asCustom();
            final InputStream res =
                sc.getResourceAsStream(path.asBlock().asString());
            if (res == null) {
                Fail.invalidArgument("Servlet.Context.get_resource_as_stream");
            } // end if
            final Channel ch = new Channel(res);
            ctxt.getContext().addChannel(ch);
            final Block b = Block.createCustom(Custom.CHANNEL_SIZE,
                                               Custom.CHANNEL_OPS);
            b.setCustom(ch);
            return Value.createFromBlock(b);
        } catch (final Fail.Exception fe) {
            throw fe;
        } catch (final Exception e) {
            Cadmium.fail(ctxt, e);
            return Value.UNIT; // never reached
        } // end try/catch
    } // end method 'cadmiumservlet_context_get_resource_as_stream(CodeRunner, Value, Value)'

    /**
     * Returns the resource paths for passed path.
     * @param ctxt context
     * @param context servlet context
     * @param path resource path
     * @return the resource paths for passed path.
     * @throws Fail.Exception if no resource exist in the passed path
     * @throws Fail.Exception if any error occurs
     */
    @Primitive
    public static Value cadmiumservlet_context_get_resource_paths(final CodeRunner ctxt,
                                                                  final Value context,
                                                                  final Value path)
        throws Fail.Exception {
        try {
            final ServletContext sc =
                (ServletContext) context.asBlock().asCustom();
            final Set paths = sc.getResourcePaths(path.asBlock().asString());
            if (paths == null) {
                Fail.invalidArgument("Servlet.Context.get_resource_paths");
            } // end if
            Value res = Value.EMPTY_LIST;
            if (paths != null) {
                for (Object o : paths) {
                    final String s = (String) o;
                    final Block b = Block.createBlock(Block.TAG_CONS,
                                                      Value.createFromBlock(Block.createString(s)),
                                                      res);
                    res = Value.createFromBlock(b);
                } // end for
            } // end if
            return res;
        } catch (final Fail.Exception fe) {
            throw fe;
        } catch (final Exception e) {
            Cadmium.fail(ctxt, e);
            return Value.UNIT; // never reached
        } // end try/catch
    } // end method 'cadmiumservlet_context_get_resource_paths(CodeRunner, Value, Value)'

    /**
     * Returns the server info string.
     * @param ctxt context
     * @param context servlet context
     * @return the server info string
     * @throws Fail.Exception if any error occurs
     */
    @Primitive
    public static Value cadmiumservlet_context_get_server_info(final CodeRunner ctxt,
                                                               final Value context)
        throws Fail.Exception {
        try {
            final ServletContext sc =
                (ServletContext) context.asBlock().asCustom();
            final String res = sc.getServerInfo();
            return Value.createFromBlock(Block.createString(res));
        } catch (final Exception e) {
            Cadmium.fail(ctxt, e);
            return Value.UNIT; // never reached
        } // end try/catch
    } // end method 'cadmiumservlet_context_get_server_info(CodeRunner, Value)'

    /**
     * Returns the servlet context name.
     * @param ctxt context
     * @param context servlet context
     * @return the servlet context name
     * @throws Fail.Exception if any error occurs
     */
    @Primitive
    public static Value cadmiumservlet_context_get_servlet_context_name(final CodeRunner ctxt,
                                                                        final Value context)
        throws Fail.Exception {
        try {
            final ServletContext sc =
                (ServletContext) context.asBlock().asCustom();
            final String res = sc.getServletContextName();
            return Value.createFromBlock(Block.createString(res != null ? res : ""));
        } catch (final Exception e) {
            Cadmium.fail(ctxt, e);
            return Value.UNIT; // never reached
        } // end try/catch
    } // end method 'cadmiumservlet_context_get_servlet_context_name(CodeRunner, Value)'

    /**
     * Logs a message.
     * @param ctxt context
     * @param context servlet context
     * @param msg message to log
     * @return <i>unit</i>
     * @throws Fail.Exception if any error occurs
     */
    @Primitive
    public static Value cadmiumservlet_context_log(final CodeRunner ctxt,
                                                   final Value context,
                                                   final Value msg)
        throws Fail.Exception {
        try {
            final ServletContext sc =
                (ServletContext) context.asBlock().asCustom();
            sc.log(msg.asBlock().asString());
            return Value.UNIT;
        } catch (final Exception e) {
            Cadmium.fail(ctxt, e);
            return Value.UNIT; // never reached
        } // end try/catch
    } // end method 'cadmiumservlet_context_log(CodeRunner, Value, Value)'

    /**
     * Logs a message.
     * @param ctxt context
     * @param context servlet context
     * @param msg message to log
     * @param exn <i>Throwable</i> instance to log
     * @return <i>unit</i>
     * @throws Fail.Exception if <i>exn</i> is not a <i>Throwable</i> instance
     * @throws Fail.Exception if any error occurs
     */
    @Primitive
    public static Value cadmiumservlet_context_log_exn(final CodeRunner ctxt,
                                                       final Value context,
                                                       final Value msg,
                                                       final Value exn)
        throws Fail.Exception {
        final Object e = exn.asBlock().asCustom();
        if (!(e instanceof Throwable)) {
            Fail.invalidArgument("Servlet.Context.log_exn");
        } // end if
        try {
            final ServletContext sc =
                (ServletContext) context.asBlock().asCustom();
            sc.log(msg.asBlock().asString(), (Throwable) e);
            return Value.UNIT;
        } catch (final Exception ce) {
            Cadmium.fail(ctxt, ce);
            return Value.UNIT; // never reached
        } // end try/catch
    } // end method 'cadmiumservlet_context_log_exn(CodeRunner, Value, Value, Value)'

    /**
     * Removes an attribute from context.
     * @param ctxt context
     * @param context servlet context
     * @param name attribute name
     * @return <i>unit</i>
     * @throws Fail.Exception if any error occurs
     */
    @Primitive
    public static Value cadmiumservlet_context_remove_attribute(final CodeRunner ctxt,
                                                                final Value context,
                                                                final Value name)
        throws Fail.Exception {
        try {
            final ServletContext sc =
                (ServletContext) context.asBlock().asCustom();
            sc.removeAttribute(name.asBlock().asString());
            return Value.UNIT;
        } catch (final Exception e) {
            Cadmium.fail(ctxt, e);
            return Value.UNIT; // never reached
        } // end try/catch
    } // end method 'cadmiumservlet_context_remove_attribute(CodeRunner, Value, Value)'

    /**
     * Adds an attribute binding to the context.
     * @param ctxt context
     * @param context servlet context
     * @param name attribute name
     * @param value attribute value
     * @return <i>unit</i>
     * @throws Fail.Exception if any error occurs
     */
    @Primitive
    public static Value cadmiumservlet_context_set_attribute(final CodeRunner ctxt,
                                                             final Value context,
                                                             final Value name,
                                                             final Value value)
        throws Fail.Exception {
        try {
            final ServletContext sc =
                (ServletContext) context.asBlock().asCustom();
            sc.setAttribute(name.asBlock().asString(),
                            value.asBlock().asCustom());
            return Value.UNIT;
        } catch (final Exception e) {
            Cadmium.fail(ctxt, e);
            return Value.UNIT; // never reached
        } // end try/catch
    } // end method 'cadmiumservlet_context_set_attribute(CodeRunner, Value, Value, Value)'

    /**
     * Returns an attribute of the passed session.
     * @param ctxt context
     * @param session session
     * @param name attribute name
     * @return the value of the attribute whose name is passed
     * @throws Fail.Exception if no such attribute exists
     * @throws Fail.Exception if any error occurs
     */
    @Primitive
    public static Value cadmiumservlet_session_get_attribute(final CodeRunner ctxt,
                                                             final Value session,
                                                             final Value name)
        throws Fail.Exception {
        try {
            final HttpSession hs = (HttpSession) session.asBlock().asCustom();
            final Object res = hs.getAttribute(name.asBlock().asString());
            if (res == null) {
                Fail.raiseNotFound();
            } // end if
            return Cadmium.createObject(res);
        } catch (final Fail.Exception fe) {
            throw fe;
        } catch (final Exception e) {
            Cadmium.fail(ctxt, e);
            return Value.UNIT; // never reached
        } // end try/catch
    } // end method 'cadmiumservlet_session_get_attribute(CodeRunner, Value, Value)'

    /**
     * Returns the list of attribute names.
     * @param ctxt context
     * @param session session
     * @return the list of attribute names
     * @throws Fail.Exception if any error occurs
     */
    @Primitive
    public static Value cadmiumservlet_session_get_attribute_names(final CodeRunner ctxt,
                                                                   final Value session)
        throws Fail.Exception {
        try {
            final HttpSession hs = (HttpSession) session.asBlock().asCustom();
            final Enumeration e = hs.getAttributeNames();
            Value res = Value.EMPTY_LIST;
            while (e.hasMoreElements()) {
                final String s = (String) e.nextElement();
                final Block b = Block.createBlock(Block.TAG_CONS,
                                                  Value.createFromBlock(Block.createString(s)),
                                                  res);
                res = Value.createFromBlock(b);
            } // end while
            return res;
        } catch (final Exception e) {
            Cadmium.fail(ctxt, e);
            return Value.UNIT; // never reached
        } // end try/catch
    } // end method 'cadmiumservlet_session_get_attribute_names(CodeRunner, Value)'

    /**
     * Returns the creation time for the passed session.
     * @param ctxt context
     * @param session session
     * @return the creation time for the passed session
     * @throws Fail.Exception if any error occurs
     */
    @Primitive
    public static Value cadmiumservlet_session_get_creation_time(final CodeRunner ctxt,
                                                                 final Value session)
        throws Fail.Exception {
        try {
            final HttpSession hs = (HttpSession) session.asBlock().asCustom();
            final long res = hs.getCreationTime();
            return Helper.createInt64(res);
        } catch (final Exception e) {
            Cadmium.fail(ctxt, e);
            return Value.UNIT; // never reached
        } // end try/catch
    } // end method 'cadmiumservlet_session_get_creation_time(CodeRunner, Value)'

    /**
     * Returns the ID for the passed session.
     * @param ctxt context
     * @param session session
     * @return the ID for the passed session
     * @throws Fail.Exception if any error occurs
     */
    @Primitive
    public static Value cadmiumservlet_session_get_id(final CodeRunner ctxt,
                                                      final Value session)
        throws Fail.Exception {
        try {
            final HttpSession hs = (HttpSession) session.asBlock().asCustom();
            final String res = hs.getId();
            return Value.createFromBlock(Block.createString(res));
        } catch (final Exception e) {
            Cadmium.fail(ctxt, e);
            return Value.UNIT; // never reached
        } // end try/catch
    } // end method 'cadmiumservlet_session_get_id(CodeRunner, Value)'

    /**
     * Returns the last accessed time for the passed session.
     * @param ctxt context
     * @param session session
     * @return the last accessed time for the passed session
     * @throws Fail.Exception if any error occurs
     */
    @Primitive
    public static Value cadmiumservlet_session_get_last_accessed_time(final CodeRunner ctxt,
                                                                      final Value session)
        throws Fail.Exception {
        try {
            final HttpSession hs = (HttpSession) session.asBlock().asCustom();
            final long res = hs.getLastAccessedTime();
            return Helper.createInt64(res);
        } catch (final Exception e) {
            Cadmium.fail(ctxt, e);
            return Value.UNIT; // never reached
        } // end try/catch
    } // end method 'cadmiumservlet_session_get_last_accessed_time(CodeRunner, Value)'

    /**
     * Returns the max inactive interval for the passed session.
     * @param ctxt context
     * @param session session
     * @return the max inactive interval for the passed session
     * @throws Fail.Exception if any error occurs
     */
    @Primitive
    public static Value cadmiumservlet_session_get_max_inactive_interval(final CodeRunner ctxt,
                                                                         final Value session)
        throws Fail.Exception {
        try {
            final HttpSession hs = (HttpSession) session.asBlock().asCustom();
            final int res = hs.getMaxInactiveInterval();
            return Helper.createInt32(res);
        } catch (final Exception e) {
            Cadmium.fail(ctxt, e);
            return Value.UNIT; // never reached
        } // end try/catch
    } // end method 'cadmiumservlet_session_get_max_inactive_interval(CodeRunner, Value)'

    /**
     * Returns the servlet context for the passed session.
     * @param ctxt context
     * @param session session
     * @return the servlet context for the passed session
     * @throws Fail.Exception if any error occurs
     */
    @Primitive
    public static Value cadmiumservlet_session_get_servlet_context(final CodeRunner ctxt,
                                                                   final Value session)
        throws Fail.Exception {
        try {
            final HttpSession hs = (HttpSession) session.asBlock().asCustom();
            return Cadmium.createObject(hs.getServletContext());
        } catch (final Exception e) {
            Cadmium.fail(ctxt, e);
            return Value.UNIT; // never reached
        } // end try/catch
    } // end method 'cadmiumservlet_session_get_servlet_context(CodeRunner, Value)'

    /**
     * Invalidates the passed session.
     * @param ctxt context
     * @param session session
     * @return <i>unit</i>
     * @throws Fail.Exception if the session is already invalidated
     * @throws Fail.Exception if any error occurs
     */
    @Primitive
    public static Value cadmiumservlet_session_invalidate(final CodeRunner ctxt,
                                                          final Value session)
        throws Fail.Exception {
        try {
            final HttpSession hs = (HttpSession) session.asBlock().asCustom();
            hs.invalidate();
            return Value.UNIT;
        } catch (final IllegalStateException ise) {
            Fail.invalidArgument("Servlet.Session.invalidate");
            return Value.UNIT; // never reached
        } catch (final Exception e) {
            Cadmium.fail(ctxt, e);
            return Value.UNIT; // never reached
        } // end try/catch
    } // end method 'cadmiumservlet_session_invalidate(CodeRunner, Value)'

    /**
     * Tests whether the client does not yet know about the session.
     * @param ctxt context
     * @param session session
     * @return <tt>true</tt> if the client does not yet know about the session,
     *         <tt>false</tt> otherwise
     * @throws Fail.Exception if the session is already invalidated
     * @throws Fail.Exception if any error occurs
     */
    @Primitive
    public static Value cadmiumservlet_session_is_new(final CodeRunner ctxt,
                                                      final Value session)
        throws Fail.Exception {
        try {
            final HttpSession hs = (HttpSession) session.asBlock().asCustom();
            return hs.isNew() ? Value.TRUE : Value.FALSE;
        } catch (final IllegalStateException ise) {
            Fail.invalidArgument("Servlet.Session.is_new");
            return Value.UNIT; // never reached
        } catch (final Exception e) {
            Cadmium.fail(ctxt, e);
            return Value.UNIT; // never reached
        } // end try/catch
    } // end method 'cadmiumservlet_session_is_new(CodeRunner, Value)'

    /**
     * Removes an attribute from session.
     * @param ctxt context
     * @param session session
     * @param name attribute name
     * @return <i>unit</i>
     * @throws Fail.Exception if any error occurs
     */
    @Primitive
    public static Value cadmiumservlet_session_remove_attribute(final CodeRunner ctxt,
                                                                final Value session,
                                                                final Value name)
        throws Fail.Exception {
        try {
            final HttpSession hs = (HttpSession) session.asBlock().asCustom();
            hs.removeAttribute(name.asBlock().asString());
            return Value.UNIT;
        } catch (final Exception e) {
            Cadmium.fail(ctxt, e);
            return Value.UNIT; // never reached
        } // end try/catch
    } // end method 'cadmiumservlet_session_remove_attribute(CodeRunner, Value, Value)'

    /**
     * Adds an attribute binding to the session.
     * @param ctxt context
     * @param session session
     * @param name attribute name
     * @param value attribute value
     * @return <i>unit</i>
     * @throws Fail.Exception if any error occurs
     */
    @Primitive
    public static Value cadmiumservlet_session_set_attribute(final CodeRunner ctxt,
                                                             final Value session,
                                                             final Value name,
                                                             final Value value)
        throws Fail.Exception {
        try {
            final HttpSession hs = (HttpSession) session.asBlock().asCustom();
            hs.setAttribute(name.asBlock().asString(),
                            value.asBlock().asCustom());
            return Value.UNIT;
        } catch (final Exception e) {
            Cadmium.fail(ctxt, e);
            return Value.UNIT; // never reached
        } // end try/catch
    } // end method 'cadmiumservlet_session_set_attribute(CodeRunner, Value, Value, Value)'

    /**
     * Sets the max inactive interval for the passed session.
     * @param ctxt context
     * @param session session
     * @param value new max inactive interval
     * @return <i>unit</i>
     * @throws Fail.Exception if any error occurs
     */
    @Primitive
    public static Value cadmiumservlet_session_set_max_inactive_interval(final CodeRunner ctxt,
                                                                         final Value session,
                                                                         final Value value)
        throws Fail.Exception {
        try {
            final HttpSession hs = (HttpSession) session.asBlock().asCustom();
            hs.setMaxInactiveInterval(value.asBlock().asInt32());
            return Value.UNIT;
        } catch (final Exception e) {
            Cadmium.fail(ctxt, e);
            return Value.UNIT; // never reached
        } // end try/catch
    } // end method 'cadmiumservlet_session_set_max_inactive_interval(CodeRunner, Value, Value)'

    /**
     * Returns an attribute of the passed request.
     * @param ctxt context
     * @param request request
     * @param name attribute name
     * @return the value of the attribute whose name is passed
     * @throws Fail.Exception if no such attribute exists
     * @throws Fail.Exception if any error occurs
     */
    @Primitive
    public static Value cadmiumservlet_request_get_attribute(final CodeRunner ctxt,
                                                             final Value request,
                                                             final Value name)
        throws Fail.Exception {
        try {
            final HttpServletRequest hsr =
                (HttpServletRequest) request.asBlock().asCustom();
            final Object res = hsr.getAttribute(name.asBlock().asString());
            if (res == null) {
                Fail.raiseNotFound();
            } // end if
            return Cadmium.createObject(res);
        } catch (final Fail.Exception fe) {
            throw fe;
        } catch (final Exception e) {
            Cadmium.fail(ctxt, e);
            return Value.UNIT; // never reached
        } // end try/catch
    } // end method 'cadmiumservlet_request_get_attribute(CodeRunner, Value, Value)'

    /**
     * Returns the list of attribute names.
     * @param ctxt context
     * @param request request
     * @return the list of attribute names
     * @throws Fail.Exception if any error occurs
     */
    @Primitive
    public static Value cadmiumservlet_request_get_attribute_names(final CodeRunner ctxt,
                                                                   final Value request)
        throws Fail.Exception {
        try {
            final HttpServletRequest hsr =
                (HttpServletRequest) request.asBlock().asCustom();
            final Enumeration e = hsr.getAttributeNames();
            Value res = Value.EMPTY_LIST;
            while (e.hasMoreElements()) {
                final String s = (String) e.nextElement();
                final Block b = Block.createBlock(Block.TAG_CONS,
                                                  Value.createFromBlock(Block.createString(s)),
                                                  res);
                res = Value.createFromBlock(b);
            } // end while
            return res;
        } catch (final Exception e) {
            Cadmium.fail(ctxt, e);
            return Value.UNIT; // never reached
        } // end try/catch
    } // end method 'cadmiumservlet_request_get_attribute_names(CodeRunner, Value)'

    /**
     * Returns the authentication scheme of the passed request.
     * @param ctxt context
     * @param request request
     * @return the authentication scheme of the passed request
     * @throws Fail.Exception if any error occurs
     */
    @Primitive
    public static Value cadmiumservlet_request_get_auth_type(final CodeRunner ctxt,
                                                             final Value request)
        throws Fail.Exception {
        try {
            final HttpServletRequest hsr =
                (HttpServletRequest) request.asBlock().asCustom();
            final String res = hsr.getAuthType();
            return Value.createFromBlock(Block.createString(res != null ? res : ""));
        } catch (final Exception e) {
            Cadmium.fail(ctxt, e);
            return Value.UNIT; // never reached
        } // end try/catch
    } // end method 'cadmiumservlet_request_get_auth_type(CodeRunner, Value)'

    /**
     * Returns the character encoding of the passed request.
     * @param ctxt context
     * @param request request
     * @return the character encoding of the passed request
     * @throws Fail.Exception if any error occurs
     */
    @Primitive
    public static Value cadmiumservlet_request_get_character_encoding(final CodeRunner ctxt,
                                                                      final Value request)
        throws Fail.Exception {
        try {
            final HttpServletRequest hsr =
                (HttpServletRequest) request.asBlock().asCustom();
            final String res = hsr.getCharacterEncoding();
            return Value.createFromBlock(Block.createString(res != null ? res : ""));
        } catch (final Exception e) {
            Cadmium.fail(ctxt, e);
            return Value.UNIT; // never reached
        } // end try/catch
    } // end method 'cadmiumservlet_request_get_character_encoding(CodeRunner, Value)'

    /**
     * Returns the content length of the passed request.
     * @param ctxt context
     * @param request request
     * @return the content length of the passed request
     * @throws Fail.Exception if any error occurs
     */
    @Primitive
    public static Value cadmiumservlet_request_get_content_length(final CodeRunner ctxt,
                                                                  final Value request)
        throws Fail.Exception {
        try {
            final HttpServletRequest hsr =
                (HttpServletRequest) request.asBlock().asCustom();
            final int res = hsr.getContentLength();
            return Helper.createInt32(res);
        } catch (final Exception e) {
            Cadmium.fail(ctxt, e);
            return Value.UNIT; // never reached
        } // end try/catch
    } // end method 'cadmiumservlet_request_get_content_length(CodeRunner, Value)'

    /**
     * Returns the content type of the passed request.
     * @param ctxt context
     * @param request request
     * @return the content type of the passed request
     * @throws Fail.Exception if any error occurs
     */
    @Primitive
    public static Value cadmiumservlet_request_get_content_type(final CodeRunner ctxt,
                                                                final Value request)
        throws Fail.Exception {
        try {
            final HttpServletRequest hsr =
                (HttpServletRequest) request.asBlock().asCustom();
            final String res = hsr.getContentType();
            return Value.createFromBlock(Block.createString(res != null ? res : ""));
        } catch (final Exception e) {
            Cadmium.fail(ctxt, e);
            return Value.UNIT; // never reached
        } // end try/catch
    } // end method 'cadmiumservlet_request_get_content_type(CodeRunner, Value)'

    /**
     * Returns the context path of the passed request.
     * @param ctxt context
     * @param request request
     * @return the context path of the passed request
     * @throws Fail.Exception if any error occurs
     */
    @Primitive
    public static Value cadmiumservlet_request_get_context_path(final CodeRunner ctxt,
                                                                final Value request)
        throws Fail.Exception {
        try {
            final HttpServletRequest hsr =
                (HttpServletRequest) request.asBlock().asCustom();
            final String res = hsr.getContextPath();
            return Value.createFromBlock(Block.createString(res));
        } catch (final Exception e) {
            Cadmium.fail(ctxt, e);
            return Value.UNIT; // never reached
        } // end try/catch
    } // end method 'cadmiumservlet_request_get_context_path(CodeRunner, Value)'

    /**
     * Returns cookies of the passed request.
     * @param ctxt context
     * @param request request
     * @return the cookies of the passed request
     * @throws Fail.Exception if any error occurs
     */
    @Primitive
    public static Value cadmiumservlet_request_get_cookies(final CodeRunner ctxt,
                                                           final Value request)
        throws Fail.Exception {
        try {
            final HttpServletRequest hsr =
                (HttpServletRequest) request.asBlock().asCustom();
            Value res = Value.EMPTY_LIST;
            final Cookie[] cookies = hsr.getCookies();
            if (cookies != null) {
                for (Cookie c : cookies) {
                    final Block b = Block.createBlock(Block.TAG_CONS,
                                                      Cadmium.createObject(c),
                                                      res);
                    res = Value.createFromBlock(b);
                } // end for
            } // end if
            return res;
        } catch (final Exception e) {
            Cadmium.fail(ctxt, e);
            return Value.UNIT; // never reached
        } // end try/catch
    } // end method 'cadmiumservlet_request_get_cookies(CodeRunner, Value)'

    /**
     * Returns the date header for the passed name.
     * @param ctxt context
     * @param request request
     * @param name header name
     * @return the date header for the passed name
     * @throws Fail.Exception if the header is not a date header
     * @throws Fail.Exception if any error occurs
     */
    @Primitive
    public static Value cadmiumservlet_request_get_date_header(final CodeRunner ctxt,
                                                               final Value request,
                                                               final Value name)
        throws Fail.Exception {
        try {
            final HttpServletRequest hsr =
                (HttpServletRequest) request.asBlock().asCustom();
            final long res = hsr.getDateHeader(name.asBlock().asString());
            return Helper.createInt64(res);
        } catch (final IllegalArgumentException iae) {
            Fail.invalidArgument("Servlet.Request.get_date_header");
            return Value.UNIT; // never reached
        } catch (final Exception e) {
            Cadmium.fail(ctxt, e);
            return Value.UNIT; // never reached
        } // end try/catch
    } // end method 'cadmiumservlet_request_get_date_header(CodeRunner, Value, Value)'

    /**
     * Returns the header for the passed name.
     * @param ctxt context
     * @param request request
     * @param name header name
     * @return the header for the passed name
     * @throws Fail.Exception if no such header exist
     * @throws Fail.Exception if any error occurs
     */
    @Primitive
    public static Value cadmiumservlet_request_get_header(final CodeRunner ctxt,
                                                          final Value request,
                                                          final Value name)
        throws Fail.Exception {
        try {
            final HttpServletRequest hsr =
                (HttpServletRequest) request.asBlock().asCustom();
            final String res = hsr.getHeader(name.asBlock().asString());
            if (res == null) {
                Fail.raiseNotFound();
            } // end if
            return Value.createFromBlock(Block.createString(res));
        } catch (final Fail.Exception fe) {
            throw fe;
        } catch (final Exception e) {
            Cadmium.fail(ctxt, e);
            return Value.UNIT; // never reached
        } // end try/catch
    } // end method 'cadmiumservlet_request_get_header(CodeRunner, Value, Value)'

    /**
     * Returns the list of header names for the passed request.
     * @param ctxt context
     * @param request request
     * @return the list of header names for the passed request
     * @throws Fail.Exception if any error occurs
     */
    @Primitive
    public static Value cadmiumservlet_request_get_header_names(final CodeRunner ctxt,
                                                                final Value request)
        throws Fail.Exception {
        try {
            final HttpServletRequest hsr =
                (HttpServletRequest) request.asBlock().asCustom();
            final Enumeration e = hsr.getHeaderNames();
            Value res = Value.EMPTY_LIST;
            if (e != null) {
                while (e.hasMoreElements()) {
                    final String s = (String) e.nextElement();
                    final Block b = Block.createBlock(Block.TAG_CONS,
                                                      Value.createFromBlock(Block.createString(s)),
                                                      res);
                    res = Value.createFromBlock(b);
                } // end while
            } // end if
            return res;
        } catch (final Exception e) {
            Cadmium.fail(ctxt, e);
            return Value.UNIT; // never reached
        } // end try/catch
    } // end method 'cadmiumservlet_request_get_header_names(CodeRunner, Value)'

    /**
     * Returns all headers for the passed name.
     * @param ctxt context
     * @param request request
     * @param name name
     * @return all headers for the passed name
     * @throws Fail.Exception if any error occurs
     */
    @Primitive
    public static Value cadmiumservlet_request_get_headers(final CodeRunner ctxt,
                                                           final Value request,
                                                           final Value name)
        throws Fail.Exception {
        try {
            final HttpServletRequest hsr =
                (HttpServletRequest) request.asBlock().asCustom();
            final Enumeration e = hsr.getHeaders(name.asBlock().asString());
            Value res = Value.EMPTY_LIST;
            if (e != null) {
                while (e.hasMoreElements()) {
                    final String s = (String) e.nextElement();
                    final Block b = Block.createBlock(Block.TAG_CONS,
                                                      Value.createFromBlock(Block.createString(s)),
                                                      res);
                    res = Value.createFromBlock(b);
                } // end while
            } // end if
            return res;
        } catch (final Exception e) {
            Cadmium.fail(ctxt, e);
            return Value.UNIT; // never reached
        } // end try/catch
    } // end method 'cadmiumservlet_request_get_headers(CodeRunner, Value, Value)'

    /**
     * Returns the input stream for the passed request.
     * @param ctxt context
     * @param request request
     * @return the input stream for the passed request
     * @throws Fail.Exception if any error occurs
     */
    @Primitive
    public static Value cadmiumservlet_request_get_input_stream(final CodeRunner ctxt,
                                                                final Value request)
        throws Fail.Exception {
        try {
            final HttpServletRequest hsr =
                (HttpServletRequest) request.asBlock().asCustom();
            final InputStream res = hsr.getInputStream();
            final Channel ch = new Channel(res);
            ctxt.getContext().addChannel(ch);
            final Block b = Block.createCustom(Custom.CHANNEL_SIZE,
                                               Custom.CHANNEL_OPS);
            b.setCustom(ch);
            return Value.createFromBlock(b);
        } catch (final Exception e) {
            Cadmium.fail(ctxt, e);
            return Value.UNIT; // never reached
        } // end try/catch
    } // end method 'cadmiumservlet_request_get_input_stream(CodeRunner, Value)'

    /**
     * Returns the int header for the passed name.
     * @param ctxt context
     * @param request request
     * @param name header name
     * @return the int header for the passed name
     * @throws Fail.Exception if the header is not an int header
     * @throws Fail.Exception if any error occurs
     */
    @Primitive
    public static Value cadmiumservlet_request_get_int_header(final CodeRunner ctxt,
                                                              final Value request,
                                                              final Value name)
        throws Fail.Exception {
        try {
            final HttpServletRequest hsr =
                (HttpServletRequest) request.asBlock().asCustom();
            final int res = hsr.getIntHeader(name.asBlock().asString());
            return Helper.createInt32(res);
        } catch (final NumberFormatException nfe) {
            Fail.invalidArgument("Servlet.Request.get_int_header");
            return Value.UNIT; // never reached
        } catch (final Exception e) {
            Cadmium.fail(ctxt, e);
            return Value.UNIT; // never reached
        } // end try/catch
    } // end method 'cadmiumservlet_request_get_int_header(CodeRunner, Value, Value)'

    /**
     * Returns the local address for the passed request.
     * @param ctxt context
     * @param request request
     * @return the local address for the passed request
     * @throws Fail.Exception if any error occurs
     */
    @Primitive
    public static Value cadmiumservlet_request_get_local_addr(final CodeRunner ctxt,
                                                              final Value request)
        throws Fail.Exception {
        try {
            final HttpServletRequest hsr =
                (HttpServletRequest) request.asBlock().asCustom();
            final String res = hsr.getLocalAddr();
            return Value.createFromBlock(Block.createString(res));
        } catch (final Exception e) {
            Cadmium.fail(ctxt, e);
            return Value.UNIT; // never reached
        } // end try/catch
    } // end method 'cadmiumservlet_request_get_local_addr(CodeRunner, Value)'

    /**
     * Returns the locale for the passed request.
     * @param ctxt context
     * @param request request
     * @return the locale for the passed request
     * @throws Fail.Exception if any error occurs
     */
    @Primitive
    public static Value cadmiumservlet_request_get_locale(final CodeRunner ctxt,
                                                          final Value request)
        throws Fail.Exception {
        try {
            final HttpServletRequest hsr =
                (HttpServletRequest) request.asBlock().asCustom();
            final String res = hsr.getLocale().getLanguage();
            return Value.createFromBlock(Block.createString(res));
        } catch (final Exception e) {
            Cadmium.fail(ctxt, e);
            return Value.UNIT; // never reached
        } // end try/catch
    } // end method 'cadmiumservlet_request_get_locale(CodeRunner, Value)'

    /**
     * Returns the locales for the passed request.
     * @param ctxt context
     * @param request request
     * @return the locales for the passed request
     * @throws Fail.Exception if any error occurs
     */
    @Primitive
    public static Value cadmiumservlet_request_get_locales(final CodeRunner ctxt,
                                                           final Value request)
        throws Fail.Exception {
        try {
            final HttpServletRequest hsr =
                (HttpServletRequest) request.asBlock().asCustom();
            final Enumeration e = hsr.getLocales();
            Value res = Value.EMPTY_LIST;
            while (e.hasMoreElements()) {
                final String s = ((Locale) e.nextElement()).getLanguage();
                final Block b = Block.createBlock(Block.TAG_CONS,
                                                  Value.createFromBlock(Block.createString(s)),
                                                  res);
                res = Value.createFromBlock(b);
            } // end while
            return res;
        } catch (final Exception e) {
            Cadmium.fail(ctxt, e);
            return Value.UNIT; // never reached
        } // end try/catch
    } // end method 'cadmiumservlet_request_get_locales(CodeRunner, Value)'

    /**
     * Returns the local name for the passed request.
     * @param ctxt context
     * @param request request
     * @return the local name for the passed request
     * @throws Fail.Exception if any error occurs
     */
    @Primitive
    public static Value cadmiumservlet_request_get_local_name(final CodeRunner ctxt,
                                                              final Value request)
        throws Fail.Exception {
        try {
            final HttpServletRequest hsr =
                (HttpServletRequest) request.asBlock().asCustom();
            final String res = hsr.getLocalName();
            return Value.createFromBlock(Block.createString(res));
        } catch (final Exception e) {
            Cadmium.fail(ctxt, e);
            return Value.UNIT; // never reached
        } // end try/catch
    } // end method 'cadmiumservlet_request_get_local_name(CodeRunner, Value)'

    /**
     * Returns the local port for the passed request.
     * @param ctxt context
     * @param request request
     * @return the local port for the passed request
     * @throws Fail.Exception if any error occurs
     */
    @Primitive
    public static Value cadmiumservlet_request_get_local_port(final CodeRunner ctxt,
                                                              final Value request)
        throws Fail.Exception {
        try {
            final HttpServletRequest hsr =
                (HttpServletRequest) request.asBlock().asCustom();
            final int res = hsr.getLocalPort();
            return Helper.createInt32(res);
        } catch (final Exception e) {
            Cadmium.fail(ctxt, e);
            return Value.UNIT; // never reached
        } // end try/catch
    } // end method 'cadmiumservlet_request_get_local_port(CodeRunner, Value)'

    /**
     * Returns the method for the passed request.
     * @param ctxt context
     * @param request request
     * @return the method for the passed request
     * @throws Fail.Exception if any error occurs
     */
    @Primitive
    public static Value cadmiumservlet_request_get_method(final CodeRunner ctxt,
                                                          final Value request)
        throws Fail.Exception {
        try {
            final HttpServletRequest hsr =
                (HttpServletRequest) request.asBlock().asCustom();
            final String res = hsr.getMethod();
            return Value.createFromBlock(Block.createString(res));
        } catch (final Exception e) {
            Cadmium.fail(ctxt, e);
            return Value.UNIT; // never reached
        } // end try/catch
    } // end method 'cadmiumservlet_request_get_method(CodeRunner, Value)'

    /**
     * Returns the parameter for the passed name.
     * @param ctxt context
     * @param request request
     * @param name name
     * @return the parameter for the passed name
     * @throws Fail.Exception if no parameter exist for the passed name
     * @throws Fail.Exception if any error occurs
     */
    @Primitive
    public static Value cadmiumservlet_request_get_parameter(final CodeRunner ctxt,
                                                             final Value request,
                                                             final Value name)
        throws Fail.Exception {
        try {
            final HttpServletRequest hsr =
                (HttpServletRequest) request.asBlock().asCustom();
            final String res = hsr.getParameter(name.asBlock().asString());
            if (res == null) {
                Fail.raiseNotFound();
            } // end if
            return Value.createFromBlock(Block.createString(res));
        } catch (final Fail.Exception fe) {
            throw fe;
        } catch (final Exception e) {
            Cadmium.fail(ctxt, e);
            return Value.UNIT; // never reached
        } // end try/catch
    } // end method 'cadmiumservlet_request_get_parameter(CodeRunner, Value, Value)'

    /**
     * Returns the parameter list of the passed request.
     * @param ctxt context
     * @param request request
     * @return the parameter list of the passed request
     * @throws Fail.Exception if any error occurs
     */
    @Primitive
    public static Value cadmiumservlet_request_get_parameter_list(final CodeRunner ctxt,
                                                                  final Value request)
        throws Fail.Exception {
        try {
            final HttpServletRequest hsr =
                (HttpServletRequest) request.asBlock().asCustom();
            Value res = Value.EMPTY_LIST;
            for (Object o : hsr.getParameterMap().entrySet()) {
                final Map.Entry e = (Map.Entry) o;
                final Value k = Value.createFromBlock(Block.createString((String) e.getKey()));
                final String[] values = (String[]) e.getValue();
                Value v = Value.EMPTY_LIST;
                for (String s : values) {
                    final Block b = Block.createBlock(Block.TAG_CONS,
                                                      Value.createFromBlock(Block.createString(s)),
                                                      v);
                    v = Value.createFromBlock(b);
                } // end for
                final Block couple = Block.createBlock(0, k, v);
                final Block b = Block.createBlock(Block.TAG_CONS,
                                                  Value.createFromBlock(couple),
                                                  res);
                res = Value.createFromBlock(b);
            } // end for
            return res;
        } catch (final Exception e) {
            Cadmium.fail(ctxt, e);
            return Value.UNIT; // never reached
        } // end try/catch
    } // end method 'cadmiumservlet_request_get_parameter_list(CodeRunner, Value)'

    /**
     * Returns the parameter names of the passed request.
     * @param ctxt context
     * @param request request
     * @return the parameter names of the passed request
     * @throws Fail.Exception if any error occurs
     */
    @Primitive
    public static Value cadmiumservlet_request_get_parameter_names(final CodeRunner ctxt,
                                                                   final Value request)
        throws Fail.Exception {
        try {
            final HttpServletRequest hsr =
                (HttpServletRequest) request.asBlock().asCustom();
            final Enumeration e = hsr.getParameterNames();
            Value res = Value.EMPTY_LIST;
            while (e.hasMoreElements()) {
                final String s = (String) e.nextElement();
                final Block b = Block.createBlock(Block.TAG_CONS,
                                                  Value.createFromBlock(Block.createString(s)),
                                                  res);
                res = Value.createFromBlock(b);
            } // end while
            return res;
        } catch (final Exception e) {
            Cadmium.fail(ctxt, e);
            return Value.UNIT; // never reached
        } // end try/catch
    } // end method 'cadmiumservlet_request_get_parameter_names(CodeRunner, Value)'

    /**
     * Returns the all parameter values for the passed name.
     * @param ctxt context
     * @param request request
     * @param name name
     * @return the all parameter values for the passed name
     * @throws Fail.Exception if no parameter exist for the passed name
     * @throws Fail.Exception if any error occurs
     */
    @Primitive
    public static Value cadmiumservlet_request_get_parameter_values(final CodeRunner ctxt,
                                                                    final Value request,
                                                                    final Value name)
        throws Fail.Exception {
        try {
            final HttpServletRequest hsr =
                (HttpServletRequest) request.asBlock().asCustom();
            final String[] values =
                hsr.getParameterValues(name.asBlock().asString());
            if (values == null) {
                Fail.raiseNotFound();
            } // end if
            Value res = Value.EMPTY_LIST;
            for (String s : values) {
                final Block b = Block.createBlock(Block.TAG_CONS,
                                                  Value.createFromBlock(Block.createString(s)),
                                                  res);
                res = Value.createFromBlock(b);
            } // end for
            return res;
        } catch (final Fail.Exception fe) {
            throw fe;
        } catch (final Exception e) {
            Cadmium.fail(ctxt, e);
            return Value.UNIT; // never reached
        } // end try/catch
    } // end method 'cadmiumservlet_request_get_parameter_values(CodeRunner, Value, Value)'

    /**
     * Returns the path info for the passed request.
     * @param ctxt context
     * @param request request
     * @return the path info for the passed request
     * @throws Fail.Exception if any error occurs
     */
    @Primitive
    public static Value cadmiumservlet_request_get_path_info(final CodeRunner ctxt,
                                                             final Value request)
        throws Fail.Exception {
        try {
            final HttpServletRequest hsr =
                (HttpServletRequest) request.asBlock().asCustom();
            final String res = hsr.getPathInfo();
            return Value.createFromBlock(Block.createString(res != null ? res : ""));
        } catch (final Exception e) {
            Cadmium.fail(ctxt, e);
            return Value.UNIT; // never reached
        } // end try/catch
    } // end method 'cadmiumservlet_request_get_path_info(CodeRunner, Value)'

    /**
     * Returns the translated path for the passed request.
     * @param ctxt context
     * @param request request
     * @return the translated path for the passed request
     * @throws Fail.Exception if any error occurs
     */
    @Primitive
    public static Value cadmiumservlet_request_get_path_translated(final CodeRunner ctxt,
                                                                   final Value request)
        throws Fail.Exception {
        try {
            final HttpServletRequest hsr =
                (HttpServletRequest) request.asBlock().asCustom();
            final String res = hsr.getPathTranslated();
            return Value.createFromBlock(Block.createString(res != null ? res : ""));
        } catch (final Exception e) {
            Cadmium.fail(ctxt, e);
            return Value.UNIT; // never reached
        } // end try/catch
    } // end method 'cadmiumservlet_request_get_path_translated(CodeRunner, Value)'

    /**
     * Returns the protocol for the passed request.
     * @param ctxt context
     * @param request request
     * @return the protocol for the passed request
     * @throws Fail.Exception if any error occurs
     */
    @Primitive
    public static Value cadmiumservlet_request_get_protocol(final CodeRunner ctxt,
                                                            final Value request)
        throws Fail.Exception {
        try {
            final HttpServletRequest hsr =
                (HttpServletRequest) request.asBlock().asCustom();
            final String res = hsr.getProtocol();
            return Value.createFromBlock(Block.createString(res));
        } catch (final Exception e) {
            Cadmium.fail(ctxt, e);
            return Value.UNIT; // never reached
        } // end try/catch
    } // end method 'cadmiumservlet_request_get_protocol(CodeRunner, Value)'

    /**
     * Returns the query string for the passed request.
     * @param ctxt context
     * @param request request
     * @return the query string for the passed request
     * @throws Fail.Exception if any error occurs
     */
    @Primitive
    public static Value cadmiumservlet_request_get_query_string(final CodeRunner ctxt,
                                                                final Value request)
        throws Fail.Exception {
        try {
            final HttpServletRequest hsr =
                (HttpServletRequest) request.asBlock().asCustom();
            final String res = hsr.getQueryString();
            return Value.createFromBlock(Block.createString(res != null ? res : ""));
        } catch (final Exception e) {
            Cadmium.fail(ctxt, e);
            return Value.UNIT; // never reached
        } // end try/catch
    } // end method 'cadmiumservlet_request_get_query_string(CodeRunner, Value)'

    /**
     * Returns the remote address for the passed request.
     * @param ctxt context
     * @param request request
     * @return the remote address for the passed request
     * @throws Fail.Exception if any error occurs
     */
    @Primitive
    public static Value cadmiumservlet_request_get_remote_addr(final CodeRunner ctxt,
                                                               final Value request)
        throws Fail.Exception {
        try {
            final HttpServletRequest hsr =
                (HttpServletRequest) request.asBlock().asCustom();
            final String res = hsr.getRemoteAddr();
            return Value.createFromBlock(Block.createString(res));
        } catch (final Exception e) {
            Cadmium.fail(ctxt, e);
            return Value.UNIT; // never reached
        } // end try/catch
    } // end method 'cadmiumservlet_request_get_remote_addr(CodeRunner, Value)'

    /**
     * Returns the remote host for the passed request.
     * @param ctxt context
     * @param request request
     * @return the remote host for the passed request
     * @throws Fail.Exception if any error occurs
     */
    @Primitive
    public static Value cadmiumservlet_request_get_remote_host(final CodeRunner ctxt,
                                                               final Value request)
        throws Fail.Exception {
        try {
            final HttpServletRequest hsr =
                (HttpServletRequest) request.asBlock().asCustom();
            final String res = hsr.getRemoteHost();
            return Value.createFromBlock(Block.createString(res));
        } catch (final Exception e) {
            Cadmium.fail(ctxt, e);
            return Value.UNIT; // never reached
        } // end try/catch
    } // end method 'cadmiumservlet_request_get_remote_host(CodeRunner, Value)'

    /**
     * Returns the remote port for the passed request.
     * @param ctxt context
     * @param request request
     * @return the remote port for the passed request
     * @throws Fail.Exception if any error occurs
     */
    @Primitive
    public static Value cadmiumservlet_request_get_remote_port(final CodeRunner ctxt,
                                                               final Value request)
        throws Fail.Exception {
        try {
            final HttpServletRequest hsr =
                (HttpServletRequest) request.asBlock().asCustom();
            final int res = hsr.getRemotePort();
            return Helper.createInt32(res);
        } catch (final Exception e) {
            Cadmium.fail(ctxt, e);
            return Value.UNIT; // never reached
        } // end try/catch
    } // end method 'cadmiumservlet_request_get_remote_port(CodeRunner, Value)'

    /**
     * Returns the remote user for the passed request.
     * @param ctxt context
     * @param request request
     * @return the remote user for the passed request
     * @throws Fail.Exception if any error occurs
     */
    @Primitive
    public static Value cadmiumservlet_request_get_remote_user(final CodeRunner ctxt,
                                                               final Value request)
        throws Fail.Exception {
        try {
            final HttpServletRequest hsr =
                (HttpServletRequest) request.asBlock().asCustom();
            final String res = hsr.getRemoteUser();
            return Value.createFromBlock(Block.createString(res != null ? res : ""));
        } catch (final Exception e) {
            Cadmium.fail(ctxt, e);
            return Value.UNIT; // never reached
        } // end try/catch
    } // end method 'cadmiumservlet_request_get_remote_user(CodeRunner, Value)'

    /**
     * Returns the requested session ID for the passed request.
     * @param ctxt context
     * @param request request
     * @return the requested session ID for the passed request
     * @throws Fail.Exception if any error occurs
     */
    @Primitive
    public static Value cadmiumservlet_request_get_requested_session_id(final CodeRunner ctxt,
                                                                        final Value request)
        throws Fail.Exception {
        try {
            final HttpServletRequest hsr =
                (HttpServletRequest) request.asBlock().asCustom();
            final String res = hsr.getRequestedSessionId();
            return Value.createFromBlock(Block.createString(res != null ? res : ""));
        } catch (final Exception e) {
            Cadmium.fail(ctxt, e);
            return Value.UNIT; // never reached
        } // end try/catch
    } // end method 'cadmiumservlet_request_get_requested_session_id(CodeRunner, Value)'

    /**
     * Returns the request URI for the passed request.
     * @param ctxt context
     * @param request request
     * @return the request URI for the passed request
     * @throws Fail.Exception if any error occurs
     */
    @Primitive
    public static Value cadmiumservlet_request_get_request_uri(final CodeRunner ctxt,
                                                               final Value request)
        throws Fail.Exception {
        try {
            final HttpServletRequest hsr =
                (HttpServletRequest) request.asBlock().asCustom();
            final String res = hsr.getRequestURI();
            return Value.createFromBlock(Block.createString(res));
        } catch (final Exception e) {
            Cadmium.fail(ctxt, e);
            return Value.UNIT; // never reached
        } // end try/catch
    } // end method 'cadmiumservlet_request_get_request_uri(CodeRunner, Value)'

    /**
     * Returns the request URL for the passed request.
     * @param ctxt context
     * @param request request
     * @return the request URL for the passed request
     * @throws Fail.Exception if any error occurs
     */
    @Primitive
    public static Value cadmiumservlet_request_get_request_url(final CodeRunner ctxt,
                                                               final Value request)
        throws Fail.Exception {
        try {
            final HttpServletRequest hsr =
                (HttpServletRequest) request.asBlock().asCustom();
            final String res = hsr.getRequestURL().toString();
            return Value.createFromBlock(Block.createString(res));
        } catch (final Exception e) {
            Cadmium.fail(ctxt, e);
            return Value.UNIT; // never reached
        } // end try/catch
    } // end method 'cadmiumservlet_request_get_request_url(CodeRunner, Value)'

    /**
     * Returns the scheme for the passed request.
     * @param ctxt context
     * @param request request
     * @return the scheme for the passed request
     * @throws Fail.Exception if any error occurs
     */
    @Primitive
    public static Value cadmiumservlet_request_get_scheme(final CodeRunner ctxt,
                                                          final Value request)
        throws Fail.Exception {
        try {
            final HttpServletRequest hsr =
                (HttpServletRequest) request.asBlock().asCustom();
            final String res = hsr.getScheme();
            return Value.createFromBlock(Block.createString(res));
        } catch (final Exception e) {
            Cadmium.fail(ctxt, e);
            return Value.UNIT; // never reached
        } // end try/catch
    } // end method 'cadmiumservlet_request_get_scheme(CodeRunner, Value)'

    /**
     * Returns the server name for the passed request.
     * @param ctxt context
     * @param request request
     * @return the server name for the passed request
     * @throws Fail.Exception if any error occurs
     */
    @Primitive
    public static Value cadmiumservlet_request_get_server_name(final CodeRunner ctxt,
                                                               final Value request)
        throws Fail.Exception {
        try {
            final HttpServletRequest hsr =
                (HttpServletRequest) request.asBlock().asCustom();
            final String res = hsr.getServerName();
            return Value.createFromBlock(Block.createString(res));
        } catch (final Exception e) {
            Cadmium.fail(ctxt, e);
            return Value.UNIT; // never reached
        } // end try/catch
    } // end method 'cadmiumservlet_request_get_server_name(CodeRunner, Value)'

    /**
     * Returns the server port for the passed request.
     * @param ctxt context
     * @param request request
     * @return the server port for the passed request
     * @throws Fail.Exception if any error occurs
     */
    @Primitive
    public static Value cadmiumservlet_request_get_server_port(final CodeRunner ctxt,
                                                               final Value request)
        throws Fail.Exception {
        try {
            final HttpServletRequest hsr =
                (HttpServletRequest) request.asBlock().asCustom();
            final int res = hsr.getServerPort();
            return Helper.createInt32(res);
        } catch (final Exception e) {
            Cadmium.fail(ctxt, e);
            return Value.UNIT; // never reached
        } // end try/catch
    } // end method 'cadmiumservlet_request_get_server_port(CodeRunner, Value)'

    /**
     * Returns the servlet path for the passed request.
     * @param ctxt context
     * @param request request
     * @return the servlet path for the passed request
     * @throws Fail.Exception if any error occurs
     */
    @Primitive
    public static Value cadmiumservlet_request_get_servlet_path(final CodeRunner ctxt,
                                                                final Value request)
        throws Fail.Exception {
        try {
            final HttpServletRequest hsr =
                (HttpServletRequest) request.asBlock().asCustom();
            final String res = hsr.getServletPath();
            return Value.createFromBlock(Block.createString(res));
        } catch (final Exception e) {
            Cadmium.fail(ctxt, e);
            return Value.UNIT; // never reached
        } // end try/catch
    } // end method 'cadmiumservlet_request_get_servlet_path(CodeRunner, Value)'

    /**
     * Returns the session for the passed request.
     * @param ctxt context
     * @param request request
     * @return the session for the passed request
     * @throws Fail.Exception if any error occurs
     */
    @Primitive
    public static Value cadmiumservlet_request_get_session(final CodeRunner ctxt,
                                                           final Value request)
        throws Fail.Exception {
        try {
            final HttpServletRequest hsr =
                (HttpServletRequest) request.asBlock().asCustom();
            return Cadmium.createObject(hsr.getSession());
        } catch (final Exception e) {
            Cadmium.fail(ctxt, e);
            return Value.UNIT; // never reached
        } // end try/catch
    } // end method 'cadmiumservlet_request_get_session(CodeRunner, Value)'

    /**
     * Returns the session for the passed request.
     * @param ctxt context
     * @param request request
     * @param create whether to create a session if there is no current session
     * @return the session for the passed request
     * @throws Fail.Exception if there is no current session and passed boolean
     *                        is <tt>false</tt>
     * @throws Fail.Exception if any error occurs
     */
    @Primitive
    public static Value cadmiumservlet_request_get_session_creating(final CodeRunner ctxt,
                                                                    final Value request,
                                                                    final Value create)
        throws Fail.Exception {
        try {
            final HttpServletRequest hsr =
                (HttpServletRequest) request.asBlock().asCustom();
            final HttpSession res = hsr.getSession(create == Value.TRUE);
            if (res == null) {
                Fail.raiseNotFound();
            } // end if
            return Cadmium.createObject(res);
        } catch (final Fail.Exception fe) {
            throw fe;
        } catch (final Exception e) {
            Cadmium.fail(ctxt, e);
            return Value.UNIT; // never reached
        } // end try/catch
    } // end method 'cadmiumservlet_request_get_session_creating(CodeRunner, Value, Value)'

    /**
     * Tests whether the requested session ID is passed by a cookie.
     * @param ctxt context
     * @param request request
     * @return <tt>true</tt> if the requested session ID is passed by a cookie,
     *         <tt>false</tt> otherwise
     * @throws Fail.Exception if any error occurs
     */
    @Primitive
    public static Value cadmiumservlet_request_is_requested_session_id_from_cookie(final CodeRunner ctxt,
                                                                                   final Value request)
        throws Fail.Exception {
        try {
            final HttpServletRequest hsr =
                (HttpServletRequest) request.asBlock().asCustom();
            final boolean res = hsr.isRequestedSessionIdFromCookie();
            return res ? Value.TRUE : Value.FALSE;
        } catch (final Exception e) {
            Cadmium.fail(ctxt, e);
            return Value.UNIT; // never reached
        } // end try/catch
    } // end method 'cadmiumservlet_request_is_requested_session_id_from_cookie(CodeRunner, Value)'

    /**
     * Tests whether the requested session ID is passed by the URL.
     * @param ctxt context
     * @param request request
     * @return <tt>true</tt> if the requested session ID is passed by the URL,
     *         <tt>false</tt> otherwise
     * @throws Fail.Exception if any error occurs
     */
    @Primitive
    public static Value cadmiumservlet_request_is_requested_session_id_from_url(final CodeRunner ctxt,
                                                                                final Value request)
        throws Fail.Exception {
        try {
            final HttpServletRequest hsr =
                (HttpServletRequest) request.asBlock().asCustom();
            final boolean res = hsr.isRequestedSessionIdFromURL();
            return res ? Value.TRUE : Value.FALSE;
        } catch (final Exception e) {
            Cadmium.fail(ctxt, e);
            return Value.UNIT; // never reached
        } // end try/catch
    } // end method 'cadmiumservlet_request_is_requested_session_id_from_url(CodeRunner, Value)'

    /**
     * Tests whether the requested session ID is still valid.
     * @param ctxt context
     * @param request request
     * @return <tt>true</tt> if the requested session ID is still valid,
     *         <tt>false</tt> otherwise
     * @throws Fail.Exception if any error occurs
     */
    @Primitive
    public static Value cadmiumservlet_request_is_requested_session_id_valid(final CodeRunner ctxt,
                                                                             final Value request)
        throws Fail.Exception {
        try {
            final HttpServletRequest hsr =
                (HttpServletRequest) request.asBlock().asCustom();
            final boolean res = hsr.isRequestedSessionIdValid();
            return res ? Value.TRUE : Value.FALSE;
        } catch (final Exception e) {
            Cadmium.fail(ctxt, e);
            return Value.UNIT; // never reached
        } // end try/catch
    } // end method 'cadmiumservlet_request_is_requested_session_id_valid(CodeRunner, Value)'

    /**
     * Tests whether the request was made using a secure channel.
     * @param ctxt context
     * @param request request
     * @return <tt>true</tt> if the request was made using a secure channel,
     *         <tt>false</tt> otherwise
     * @throws Fail.Exception if any error occurs
     */
    @Primitive
    public static Value cadmiumservlet_request_is_secure(final CodeRunner ctxt,
                                                         final Value request)
        throws Fail.Exception {
        try {
            final HttpServletRequest hsr =
                (HttpServletRequest) request.asBlock().asCustom();
            final boolean res = hsr.isSecure();
            return res ? Value.TRUE : Value.FALSE;
        } catch (final Exception e) {
            Cadmium.fail(ctxt, e);
            return Value.UNIT; // never reached
        } // end try/catch
    } // end method 'cadmiumservlet_request_is_secure(CodeRunner, Value)'

    /**
     * Tests whether the authenticated user in included in the passed role.
     * @param ctxt context
     * @param request request
     * @return <tt>true</tt> if the authenticated user in included in the passed role,
     *         <tt>false</tt> otherwise
     * @throws Fail.Exception if any error occurs
     */
    @Primitive
    public static Value cadmiumservlet_request_is_user_in_role(final CodeRunner ctxt,
                                                               final Value request,
                                                               final Value role)
        throws Fail.Exception {
        try {
            final HttpServletRequest hsr =
                (HttpServletRequest) request.asBlock().asCustom();
            final boolean res = hsr.isUserInRole(role.asBlock().asString());
            return res ? Value.TRUE : Value.FALSE;
        } catch (final Exception e) {
            Cadmium.fail(ctxt, e);
            return Value.UNIT; // never reached
        } // end try/catch
    } // end method 'cadmiumservlet_request_is_user_in_role(CodeRunner, Value, Value)'

    /**
     * Removes an attribute from request.
     * @param ctxt context
     * @param request request
     * @param name attribute name
     * @return <i>unit</i>
     * @throws Fail.Exception if any error occurs
     */
    @Primitive
    public static Value cadmiumservlet_request_remove_attribute(final CodeRunner ctxt,
                                                                final Value request,
                                                                final Value name)
        throws Fail.Exception {
        try {
            final HttpServletRequest hsr =
                (HttpServletRequest) request.asBlock().asCustom();
            hsr.removeAttribute(name.asBlock().asString());
            return Value.UNIT;
        } catch (final Exception e) {
            Cadmium.fail(ctxt, e);
            return Value.UNIT; // never reached
        } // end try/catch
    } // end method 'cadmiumservlet_request_remove_attribute(CodeRunner, Value, Value)'

    /**
     * Adds an attribute binding to the request.
     * @param ctxt context
     * @param request request
     * @param name attribute name
     * @param value attribute value
     * @return <i>unit</i>
     * @throws Fail.Exception if any error occurs
     */
    @Primitive
    public static Value cadmiumservlet_request_set_attribute(final CodeRunner ctxt,
                                                             final Value request,
                                                             final Value name,
                                                             final Value value)
        throws Fail.Exception {
        try {
            final HttpServletRequest hsr =
                (HttpServletRequest) request.asBlock().asCustom();
            hsr.setAttribute(name.asBlock().asString(),
                             value.asBlock().asCustom());
            return Value.UNIT;
        } catch (final Exception e) {
            Cadmium.fail(ctxt, e);
            return Value.UNIT; // never reached
        } // end try/catch
    } // end method 'cadmiumservlet_request_set_attribute(CodeRunner, Value, Value, Value)'

    /**
     * Sets the character encoding for the passed request.
     * @param ctxt context
     * @param request request
     * @param encoding new character encoding
     * @return <i>unit</i>
     * @throws Fail.Exception if any error occurs
     */
    @Primitive
    public static Value cadmiumservlet_request_set_character_encoding(final CodeRunner ctxt,
                                                                      final Value request,
                                                                      final Value encoding)
        throws Fail.Exception {
        try {
            final HttpServletRequest hsr =
                (HttpServletRequest) request.asBlock().asCustom();
            hsr.setCharacterEncoding(encoding.asBlock().asString());
            return Value.UNIT;
        } catch (final Exception e) {
            Cadmium.fail(ctxt, e);
            return Value.UNIT; // never reached
        } // end try/catch
    } // end method 'cadmiumservlet_request_set_character_encoding(CodeRunner, Value, Value)'

    /**
     * Adds the passed cookie to the response.
     * @param ctxt context
     * @param response response
     * @param cookie cookie
     * @return <i>unit</i>
     * @throws Fail.Exception if any error occurs
     */
    @Primitive
    public static Value cadmiumservlet_response_add_cookie(final CodeRunner ctxt,
                                                           final Value response,
                                                           final Value cookie)
        throws Fail.Exception {
        try {
            final HttpServletResponse hsr =
                (HttpServletResponse) response.asBlock().asCustom();
            hsr.addCookie((Cookie) cookie.asBlock().asCustom());
            return Value.UNIT;
        } catch (final Exception e) {
            Cadmium.fail(ctxt, e);
            return Value.UNIT; // never reached
        } // end try/catch
    } // end method 'cadmiumservlet_response_add_cookie(CodeRunner, Value, Value)'

    /**
     * Adds the date header with passed name and value to the response.
     * @param ctxt context
     * @param response response
     * @param name header name
     * @param value header value
     * @return <i>unit</i>
     * @throws Fail.Exception if any error occurs
     */
    @Primitive
    public static Value cadmiumservlet_response_add_date_header(final CodeRunner ctxt,
                                                                final Value response,
                                                                final Value name,
                                                                final Value value)
        throws Fail.Exception {
        try {
            final HttpServletResponse hsr =
                (HttpServletResponse) response.asBlock().asCustom();
            hsr.addDateHeader(name.asBlock().asString(),
                              value.asBlock().asInt64());
            return Value.UNIT;
        } catch (final Exception e) {
            Cadmium.fail(ctxt, e);
            return Value.UNIT; // never reached
        } // end try/catch
    } // end method 'cadmiumservlet_response_add_date_header(CodeRunner, Value, Value, Value)'

    /**
     * Adds the header with passed name and value to the response.
     * @param ctxt context
     * @param response response
     * @param name header name
     * @param value header value
     * @return <i>unit</i>
     * @throws Fail.Exception if any error occurs
     */
    @Primitive
    public static Value cadmiumservlet_response_add_header(final CodeRunner ctxt,
                                                           final Value response,
                                                           final Value name,
                                                           final Value value)
        throws Fail.Exception {
        try {
            final HttpServletResponse hsr =
                (HttpServletResponse) response.asBlock().asCustom();
            hsr.addHeader(name.asBlock().asString(),
                          value.asBlock().asString());
            return Value.UNIT;
        } catch (final Exception e) {
            Cadmium.fail(ctxt, e);
            return Value.UNIT; // never reached
        } // end try/catch
    } // end method 'cadmiumservlet_response_add_header(CodeRunner, Value, Value, Value)'

    /**
     * Adds the int header with passed name and value to the response.
     * @param ctxt context
     * @param response response
     * @param name header name
     * @param value header value
     * @return <i>unit</i>
     * @throws Fail.Exception if any error occurs
     */
    @Primitive
    public static Value cadmiumservlet_response_add_int_header(final CodeRunner ctxt,
                                                               final Value response,
                                                               final Value name,
                                                               final Value value)
        throws Fail.Exception {
        try {
            final HttpServletResponse hsr =
                (HttpServletResponse) response.asBlock().asCustom();
            hsr.addIntHeader(name.asBlock().asString(),
                             value.asBlock().asInt32());
            return Value.UNIT;
        } catch (final Exception e) {
            Cadmium.fail(ctxt, e);
            return Value.UNIT; // never reached
        } // end try/catch
    } // end method 'cadmiumservlet_response_add_int_header(CodeRunner, Value, Value, Value)'

    /**
     * Tests whether there exist a header with the passed name.
     * @param ctxt context
     * @param response response
     * @param name header name
     * @return <tt>true</tt> if there exist a header with the passed name,
     *         <tt>false</tt> otherwise
     * @throws Fail.Exception if any error occurs
     */
    @Primitive
    public static Value cadmiumservlet_response_contains_header(final CodeRunner ctxt,
                                                                final Value response,
                                                                final Value name)
        throws Fail.Exception {
        try {
            final HttpServletResponse hsr =
                (HttpServletResponse) response.asBlock().asCustom();
            final boolean res = hsr.containsHeader(name.asBlock().asString());
            return res ? Value.TRUE : Value.FALSE;
        } catch (final Exception e) {
            Cadmium.fail(ctxt, e);
            return Value.UNIT; // never reached
        } // end try/catch
    } // end method 'cadmiumservlet_response_contains_header(CodeRunner, Value, Value)'

    /**
     * Encodes the passed URL for use in redirection.
     * @param ctxt context
     * @param response response
     * @param url URL to encode
     * @return encoded URL
     * @throws Fail.Exception if any error occurs
     */
    @Primitive
    public static Value cadmiumservlet_response_encode_redirect_url(final CodeRunner ctxt,
                                                                    final Value response,
                                                                    final Value url)
        throws Fail.Exception {
        try {
            final HttpServletResponse hsr =
                (HttpServletResponse) response.asBlock().asCustom();
            final String res = hsr.encodeRedirectURL(url.asBlock().asString());
            return Value.createFromBlock(Block.createString(res));
        } catch (final Exception e) {
            Cadmium.fail(ctxt, e);
            return Value.UNIT; // never reached
        } // end try/catch
    } // end method 'cadmiumservlet_response_encode_redirect_url(CodeRunner, Value, Value)'

    /**
     * Encodes the passed URL.
     * @param ctxt context
     * @param response response
     * @param url URL to encode
     * @return encoded URL
     * @throws Fail.Exception if any error occurs
     */
    @Primitive
    public static Value cadmiumservlet_response_encode_url(final CodeRunner ctxt,
                                                           final Value response,
                                                           final Value url)
        throws Fail.Exception {
        try {
            final HttpServletResponse hsr =
                (HttpServletResponse) response.asBlock().asCustom();
            final String res = hsr.encodeURL(url.asBlock().asString());
            return Value.createFromBlock(Block.createString(res));
        } catch (final Exception e) {
            Cadmium.fail(ctxt, e);
            return Value.UNIT; // never reached
        } // end try/catch
    } // end method 'cadmiumservlet_response_encode_url(CodeRunner, Value, Value)'

    /**
     * Forces the content in the buffer to be written to the client.
     * @param ctxt context
     * @param response response
     * @return <i>unit</i>
     * @throws Fail.Exception if any error occurs
     */
    @Primitive
    public static Value cadmiumservlet_response_flush_buffer(final CodeRunner ctxt,
                                                             final Value response)
        throws Fail.Exception {
        try {
            final HttpServletResponse hsr =
                (HttpServletResponse) response.asBlock().asCustom();
            hsr.flushBuffer();
            return Value.UNIT;
        } catch (final Exception e) {
            Cadmium.fail(ctxt, e);
            return Value.UNIT; // never reached
        } // end try/catch
    } // end method 'cadmiumservlet_response_flush_buffer(CodeRunner, Value)'

    /**
     * Returns the buffer size for the passed response.
     * @param ctxt context
     * @param response response
     * @return the buffer size for the passed response
     * @throws Fail.Exception if any error occurs
     */
    @Primitive
    public static Value cadmiumservlet_response_get_buffer_size(final CodeRunner ctxt,
                                                                final Value response)
        throws Fail.Exception {
        try {
            final HttpServletResponse hsr =
                (HttpServletResponse) response.asBlock().asCustom();
            final int res = hsr.getBufferSize();
            return Helper.createInt32(res);
        } catch (final Exception e) {
            Cadmium.fail(ctxt, e);
            return Value.UNIT; // never reached
        } // end try/catch
    } // end method 'cadmiumservlet_response_get_buffer_size(CodeRunner, Value)'

    /**
     * Returns the character encoding for the passed response.
     * @param ctxt context
     * @param response response
     * @return the character encoding for the passed response
     * @throws Fail.Exception if any error occurs
     */
    @Primitive
    public static Value cadmiumservlet_response_get_character_encoding(final CodeRunner ctxt,
                                                                       final Value response)
        throws Fail.Exception {
        try {
            final HttpServletResponse hsr =
                (HttpServletResponse) response.asBlock().asCustom();
            final String res = hsr.getCharacterEncoding();
            return Value.createFromBlock(Block.createString(res));
        } catch (final Exception e) {
            Cadmium.fail(ctxt, e);
            return Value.UNIT; // never reached
        } // end try/catch
    } // end method 'cadmiumservlet_response_get_character_encoding(CodeRunner, Value)'

    /**
     * Returns the content type for the passed response.
     * @param ctxt context
     * @param response response
     * @return the content type for the passed response
     * @throws Fail.Exception if any error occurs
     */
    @Primitive
    public static Value cadmiumservlet_response_get_content_type(final CodeRunner ctxt,
                                                                 final Value response)
        throws Fail.Exception {
        try {
            final HttpServletResponse hsr =
                (HttpServletResponse) response.asBlock().asCustom();
            final String res = hsr.getContentType();
            return Value.createFromBlock(Block.createString(res != null ? res : ""));
        } catch (final Exception e) {
            Cadmium.fail(ctxt, e);
            return Value.UNIT; // never reached
        } // end try/catch
    } // end method 'cadmiumservlet_response_get_content_type(CodeRunner, Value)'

    /**
     * Returns the locale for the passed response.
     * @param ctxt context
     * @param response response
     * @return the locale for the passed response
     * @throws Fail.Exception if any error occurs
     */
    @Primitive
    public static Value cadmiumservlet_response_get_locale(final CodeRunner ctxt,
                                                           final Value response)
        throws Fail.Exception {
        try {
            final HttpServletResponse hsr =
                (HttpServletResponse) response.asBlock().asCustom();
            final String res = hsr.getLocale().getLanguage();
            return Value.createFromBlock(Block.createString(res));
        } catch (final Exception e) {
            Cadmium.fail(ctxt, e);
            return Value.UNIT; // never reached
        } // end try/catch
    } // end method 'cadmiumservlet_response_get_locale(CodeRunner, Value)'

    /**
     * Returns the output stream for the passed response.
     * @param ctxt context
     * @param response response
     * @return the output stream for the passed response
     * @throws Fail.Exception if any error occurs
     */
    @Primitive
    public static Value cadmiumservlet_response_get_output_stream(final CodeRunner ctxt,
                                                                  final Value response)
        throws Fail.Exception {
        try {
            final HttpServletResponse hsr =
                (HttpServletResponse) response.asBlock().asCustom();
            final OutputStream res = hsr.getOutputStream();
            final Channel ch = new Channel(res);
            ctxt.getContext().addChannel(ch);
            final Block b = Block.createCustom(Custom.CHANNEL_SIZE,
                                               Custom.CHANNEL_OPS);
            b.setCustom(ch);
            return Value.createFromBlock(b);
        } catch (final Exception e) {
            Cadmium.fail(ctxt, e);
            return Value.UNIT; // never reached
        } // end try/catch
    } // end method 'cadmiumservlet_response_get_output_stream(CodeRunner, Value)'

    /**
     * Tests whether the passed response has been committed.
     * @param ctxt context
     * @param response response
     * @return <tt>true</tt> if the passed response has been committed,
     *         <tt>false</tt> otherwise
     * @throws Fail.Exception if any error occurs
     */
    @Primitive
    public static Value cadmiumservlet_response_is_committed(final CodeRunner ctxt,
                                                             final Value response)
        throws Fail.Exception {
        try {
            final HttpServletResponse hsr =
                (HttpServletResponse) response.asBlock().asCustom();
            final boolean res = hsr.isCommitted();
            return res ? Value.TRUE : Value.FALSE;
        } catch (final Exception e) {
            Cadmium.fail(ctxt, e);
            return Value.UNIT; // never reached
        } // end try/catch
    } // end method 'cadmiumservlet_response_is_committed(CodeRunner, Value)'

    /**
     * Clears the buffer as well as the status code and headers.
     * @param ctxt context
     * @param response response
     * @return <i>unit</i>
     * @throws Fail.Exception if the response has already been committed
     * @throws Fail.Exception if any error occurs
     */
    @Primitive
    public static Value cadmiumservlet_response_reset(final CodeRunner ctxt,
                                                      final Value response)
        throws Fail.Exception {
        try {
            final HttpServletResponse hsr =
                (HttpServletResponse) response.asBlock().asCustom();
            hsr.reset();
            return Value.UNIT;
        } catch (final IllegalStateException ise) {
            Fail.invalidArgument("Servlet.Response.reset");
            return Value.UNIT; // never reached
        } catch (final Exception e) {
            Cadmium.fail(ctxt, e);
            return Value.UNIT; // never reached
        } // end try/catch
    } // end method 'cadmiumservlet_response_reset(CodeRunner, Value)'

    /**
     * Clears the buffer.
     * @param ctxt context
     * @param response response
     * @return <i>unit</i>
     * @throws Fail.Exception if the response has already been committed
     * @throws Fail.Exception if any error occurs
     */
    @Primitive
    public static Value cadmiumservlet_response_reset_buffer(final CodeRunner ctxt,
                                                             final Value response)
        throws Fail.Exception {
        try {
            final HttpServletResponse hsr =
                (HttpServletResponse) response.asBlock().asCustom();
            hsr.resetBuffer();
            return Value.UNIT;
        } catch (final IllegalStateException ise) {
            Fail.invalidArgument("Servlet.Response.reset_buffer");
            return Value.UNIT; // never reached
        } catch (final Exception e) {
            Cadmium.fail(ctxt, e);
            return Value.UNIT; // never reached
        } // end try/catch
    } // end method 'cadmiumservlet_response_reset_buffer(CodeRunner, Value)'

    /**
     * Sends an error message.
     * @param ctxt context
     * @param response response
     * @param err error code
     * @return <i>unit</i>
     * @throws Fail.Exception if the response has already been committed
     * @throws Fail.Exception if any error occurs
     */
    @Primitive
    public static Value cadmiumservlet_response_send_error(final CodeRunner ctxt,
                                                           final Value response,
                                                           final Value err)
        throws Fail.Exception {
        try {
            final HttpServletResponse hsr =
                (HttpServletResponse) response.asBlock().asCustom();
            hsr.sendError(err.asBlock().asInt32());
            return Value.UNIT;
        } catch (final IllegalStateException ise) {
            Fail.invalidArgument("Servlet.Response.send_error");
            return Value.UNIT; // never reached
        } catch (final Exception e) {
            Cadmium.fail(ctxt, e);
            return Value.UNIT; // never reached
        } // end try/catch
    } // end method 'cadmiumservlet_response_send_error(CodeRunner, Value, Value)'

    /**
     * Sends an error message.
     * @param ctxt context
     * @param response response
     * @param err error code
     * @param msg error message
     * @return <i>unit</i>
     * @throws Fail.Exception if the response has already been committed
     * @throws Fail.Exception if any error occurs
     */
    @Primitive
    public static Value cadmiumservlet_response_send_error_msg(final CodeRunner ctxt,
                                                               final Value response,
                                                               final Value err,
                                                               final Value msg)
        throws Fail.Exception {
        try {
            final HttpServletResponse hsr =
                (HttpServletResponse) response.asBlock().asCustom();
            hsr.sendError(err.asBlock().asInt32(), msg.asBlock().asString());
            return Value.UNIT;
        } catch (final IllegalStateException ise) {
            Fail.invalidArgument("Servlet.Response.send_error_msg");
            return Value.UNIT; // never reached
        } catch (final Exception e) {
            Cadmium.fail(ctxt, e);
            return Value.UNIT; // never reached
        } // end try/catch
    } // end method 'cadmiumservlet_response_send_error_msg(CodeRunner, Value, Value, Value)'

    /**
     * Sends a redirect response.
     * @param ctxt context
     * @param response response
     * @param loc redirection location
     * @return <i>unit</i>
     * @throws Fail.Exception if the response has already been committed
     * @throws Fail.Exception if any error occurs
     */
    @Primitive
    public static Value cadmiumservlet_response_send_redirect(final CodeRunner ctxt,
                                                              final Value response,
                                                              final Value loc)
        throws Fail.Exception {
        try {
            final HttpServletResponse hsr =
                (HttpServletResponse) response.asBlock().asCustom();
            hsr.sendRedirect(loc.asBlock().asString());
            return Value.UNIT;
        } catch (final IllegalStateException ise) {
            Fail.invalidArgument("Servlet.Response.send_redirect");
            return Value.UNIT; // never reached
        } catch (final Exception e) {
            Cadmium.fail(ctxt, e);
            return Value.UNIT; // never reached
        } // end try/catch
    } // end method 'cadmiumservlet_response_send_redirect(CodeRunner, Value, Value)'

    /**
     * Sets buffer size for the passed response.
     * @param ctxt context
     * @param response response
     * @param sz new buffer size
     * @return <i>unit</i>
     * @throws Fail.Exception if the response has already been committed
     * @throws Fail.Exception if any error occurs
     */
    @Primitive
    public static Value cadmiumservlet_response_set_buffer_size(final CodeRunner ctxt,
                                                                final Value response,
                                                                final Value sz)
        throws Fail.Exception {
        try {
            final HttpServletResponse hsr =
                (HttpServletResponse) response.asBlock().asCustom();
            hsr.setBufferSize(sz.asBlock().asInt32());
            return Value.UNIT;
        } catch (final IllegalStateException ise) {
            Fail.invalidArgument("Servlet.Response.set_buffer_size");
            return Value.UNIT; // never reached
        } catch (final Exception e) {
            Cadmium.fail(ctxt, e);
            return Value.UNIT; // never reached
        } // end try/catch
    } // end method 'cadmiumservlet_response_set_buffer_size(CodeRunner, Value, Value)'

    /**
     * Sets character encoding for the passed response.
     * @param ctxt context
     * @param response response
     * @param encoding new encoding
     * @return <i>unit</i>
     * @throws Fail.Exception if any error occurs
     */
    @Primitive
    public static Value cadmiumservlet_response_set_character_encoding(final CodeRunner ctxt,
                                                                       final Value response,
                                                                       final Value encoding)
        throws Fail.Exception {
        try {
            final HttpServletResponse hsr =
                (HttpServletResponse) response.asBlock().asCustom();
            hsr.setCharacterEncoding(encoding.asBlock().asString());
            return Value.UNIT;
        } catch (final Exception e) {
            Cadmium.fail(ctxt, e);
            return Value.UNIT; // never reached
        } // end try/catch
    } // end method 'cadmiumservlet_response_set_character_encoding(CodeRunner, Value, Value)'

    /**
     * Sets content length for the passed response.
     * @param ctxt context
     * @param response response
     * @param len new length
     * @return <i>unit</i>
     * @throws Fail.Exception if any error occurs
     */
    @Primitive
    public static Value cadmiumservlet_response_set_content_length(final CodeRunner ctxt,
                                                                   final Value response,
                                                                   final Value len)
        throws Fail.Exception {
        try {
            final HttpServletResponse hsr =
                (HttpServletResponse) response.asBlock().asCustom();
            hsr.setContentLength(len.asBlock().asInt32());
            return Value.UNIT;
        } catch (final Exception e) {
            Cadmium.fail(ctxt, e);
            return Value.UNIT; // never reached
        } // end try/catch
    } // end method 'cadmiumservlet_response_set_content_length(CodeRunner, Value, Value)'

    /**
     * Sets content type for the passed response.
     * @param ctxt context
     * @param response response
     * @param type new type
     * @return <i>unit</i>
     * @throws Fail.Exception if any error occurs
     */
    @Primitive
    public static Value cadmiumservlet_response_set_content_type(final CodeRunner ctxt,
                                                                 final Value response,
                                                                 final Value type)
        throws Fail.Exception {
        try {
            final HttpServletResponse hsr =
                (HttpServletResponse) response.asBlock().asCustom();
            hsr.setContentType(type.asBlock().asString());
            return Value.UNIT;
        } catch (final Exception e) {
            Cadmium.fail(ctxt, e);
            return Value.UNIT; // never reached
        } // end try/catch
    } // end method 'cadmiumservlet_response_set_content_type(CodeRunner, Value, Value)'

    /**
     * Sets the date header with passed name to value.
     * @param ctxt context
     * @param response response
     * @param name header name
     * @param value header value
     * @return <i>unit</i>
     * @throws Fail.Exception if any error occurs
     */
    @Primitive
    public static Value cadmiumservlet_response_set_date_header(final CodeRunner ctxt,
                                                                final Value response,
                                                                final Value name,
                                                                final Value value)
        throws Fail.Exception {
        try {
            final HttpServletResponse hsr =
                (HttpServletResponse) response.asBlock().asCustom();
            hsr.setDateHeader(name.asBlock().asString(),
                              value.asBlock().asInt64());
            return Value.UNIT;
        } catch (final Exception e) {
            Cadmium.fail(ctxt, e);
            return Value.UNIT; // never reached
        } // end try/catch
    } // end method 'cadmiumservlet_response_set_date_header(CodeRunner, Value, Value, Value)'

    /**
     * Sets the header with passed name to value.
     * @param ctxt context
     * @param response response
     * @param name header name
     * @param value header value
     * @return <i>unit</i>
     * @throws Fail.Exception if any error occurs
     */
    @Primitive
    public static Value cadmiumservlet_response_set_header(final CodeRunner ctxt,
                                                           final Value response,
                                                           final Value name,
                                                           final Value value)
        throws Fail.Exception {
        try {
            final HttpServletResponse hsr =
                (HttpServletResponse) response.asBlock().asCustom();
            hsr.setHeader(name.asBlock().asString(),
                          value.asBlock().asString());
            return Value.UNIT;
        } catch (final Exception e) {
            Cadmium.fail(ctxt, e);
            return Value.UNIT; // never reached
        } // end try/catch
    } // end method 'cadmiumservlet_response_set_header(CodeRunner, Value, Value, Value)'

    /**
     * Sets the int header with passed name to value.
     * @param ctxt context
     * @param response response
     * @param name header name
     * @param value header value
     * @return <i>unit</i>
     * @throws Fail.Exception if any error occurs
     */
    @Primitive
    public static Value cadmiumservlet_response_set_int_header(final CodeRunner ctxt,
                                                               final Value response,
                                                               final Value name,
                                                               final Value value)
        throws Fail.Exception {
        try {
            final HttpServletResponse hsr =
                (HttpServletResponse) response.asBlock().asCustom();
            hsr.setIntHeader(name.asBlock().asString(),
                             value.asBlock().asInt32());
            return Value.UNIT;
        } catch (final Exception e) {
            Cadmium.fail(ctxt, e);
            return Value.UNIT; // never reached
        } // end try/catch
    } // end method 'cadmiumservlet_response_set_int_header(CodeRunner, Value, Value, Value)'

    /**
     * Sets locale for the passed response.
     * @param ctxt context
     * @param response response
     * @param lang new locale
     * @return <i>unit</i>
     * @throws Fail.Exception if any error occurs
     */
    @Primitive
    public static Value cadmiumservlet_response_set_locale(final CodeRunner ctxt,
                                                           final Value response,
                                                           final Value lang)
        throws Fail.Exception {
        try {
            final HttpServletResponse hsr =
                (HttpServletResponse) response.asBlock().asCustom();
            hsr.setLocale(new Locale(lang.asBlock().asString()));
            return Value.UNIT;
        } catch (final Exception e) {
            Cadmium.fail(ctxt, e);
            return Value.UNIT; // never reached
        } // end try/catch
    } // end method 'cadmiumservlet_response_set_locale(CodeRunner, Value, Value)'

    /**
     * Sets status for the passed response.
     * @param ctxt context
     * @param response response
     * @param value new status
     * @return <i>unit</i>
     * @throws Fail.Exception if any error occurs
     */
    @Primitive
    public static Value cadmiumservlet_response_set_status(final CodeRunner ctxt,
                                                           final Value response,
                                                           final Value value)
        throws Fail.Exception {
        try {
            final HttpServletResponse hsr =
                (HttpServletResponse) response.asBlock().asCustom();
            hsr.setStatus(value.asBlock().asInt32());
            return Value.UNIT;
        } catch (final Exception e) {
            Cadmium.fail(ctxt, e);
            return Value.UNIT; // never reached
        } // end try/catch
    } // end method 'cadmiumservlet_response_set_status(CodeRunner, Value, Value)'

    /**
     * Returns the value of an init parameter.
     * @param ctxt context
     * @param servlet servlet
     * @param name init parameter name
     * @return the value of the init parameter whose name is passed
     * @throws Fail.Exception if no such parameter exists
     * @throws Fail.Exception if any error occurs
     */
    @Primitive
    public static Value cadmiumservlet_servlet_get_init_parameter(final CodeRunner ctxt,
                                                                  final Value servlet,
                                                                  final Value name)
        throws Fail.Exception {
        try {
            final HttpServlet hs = (HttpServlet) servlet.asBlock().asCustom();
            final String res = hs.getInitParameter(name.asBlock().asString());
            if (res == null) {
                Fail.raiseNotFound();
            } // end if
            return Value.createFromBlock(Block.createString(res));
        } catch (final Fail.Exception fe) {
            throw fe;
        } catch (final Exception e) {
            Cadmium.fail(ctxt, e);
            return Value.UNIT; // never reached
        } // end try/catch
    } // end method 'cadmiumservlet_servlet_get_init_parameter(CodeRunner, Value, Value)'

    /**
     * Returns the list of the parameter names.
     * @param ctxt context
     * @param servlet servlet
     * @return the list of the parameter names
     * @throws Fail.Exception if any error occurs
     */
    @Primitive
    public static Value cadmiumservlet_servlet_get_init_parameter_names(final CodeRunner ctxt,
                                                                        final Value servlet)
        throws Fail.Exception {
        try {
            final HttpServlet hs = (HttpServlet) servlet.asBlock().asCustom();
            final Enumeration e = hs.getInitParameterNames();
            Value res = Value.EMPTY_LIST;
            while (e.hasMoreElements()) {
                final String s = (String) e.nextElement();
                final Block b = Block.createBlock(Block.TAG_CONS,
                                                  Value.createFromBlock(Block.createString(s)),
                                                  res);
                res = Value.createFromBlock(b);
            } // end while
            return res;
        } catch (final Exception e) {
            Cadmium.fail(ctxt, e);
            return Value.UNIT; // never reached
        } // end try/catch
    } // end method 'cadmiumservlet_servlet_get_init_parameter_names(CodeRunner, Value)'

    /**
     * Returns the context of the passed servlet.
     * @param ctxt context
     * @param servlet servlet
     * @return the context of the passed servlet
     * @throws Fail.Exception if any error occurs
     */
    @Primitive
    public static Value cadmiumservlet_servlet_get_servlet_context(final CodeRunner ctxt,
                                                                   final Value servlet)
        throws Fail.Exception {
        try {
            final HttpServlet hs = (HttpServlet) servlet.asBlock().asCustom();
            return Cadmium.createObject(hs.getServletContext());
        } catch (final Exception e) {
            Cadmium.fail(ctxt, e);
            return Value.UNIT; // never reached
        } // end try/catch
    } // end method 'cadmiumservlet_servlet_get_servlet_context(CodeRunner, Value)'

    /**
     * Returns the name of the passed servlet.
     * @param ctxt context
     * @param servlet servlet
     * @return the name of the passed servlet
     * @throws Fail.Exception if any error occurs
     */
    @Primitive
    public static Value cadmiumservlet_servlet_get_servlet_name(final CodeRunner ctxt,
                                                                final Value servlet)
        throws Fail.Exception {
        try {
            final HttpServlet hs = (HttpServlet) servlet.asBlock().asCustom();
            final String res = hs.getServletName();
            return Value.createFromBlock(Block.createString(res));
        } catch (final Exception e) {
            Cadmium.fail(ctxt, e);
            return Value.UNIT; // never reached
        } // end try/catch
    } // end method 'cadmiumservlet_servlet_get_servlet_name(CodeRunner, Value)'

    /**
     * Logs a message.
     * @param ctxt context
     * @param servlet servlet
     * @param msg message to log
     * @return <i>unit</i>
     * @throws Fail.Exception if any error occurs
     */
    @Primitive
    public static Value cadmiumservlet_servlet_log(final CodeRunner ctxt,
                                                   final Value servlet,
                                                   final Value msg)
        throws Fail.Exception {
        try {
            final HttpServlet hs = (HttpServlet) servlet.asBlock().asCustom();
            hs.log(msg.asBlock().asString());
            return Value.UNIT;
        } catch (final Exception e) {
            Cadmium.fail(ctxt, e);
            return Value.UNIT; // never reached
        } // end try/catch
    } // end method 'cadmiumservlet_servlet_log(CodeRunner, Value, Value)'

    /**
     * Logs a message.
     * @param ctxt context
     * @param servlet servlet
     * @param msg message to log
     * @param exn <i>Throwable</i> instance to log
     * @return <i>unit</i>
     * @throws Fail.Exception if <i>exn</i> is not a <i>Throwable</i> instance
     * @throws Fail.Exception if any error occurs
     */
    @Primitive
    public static Value cadmiumservlet_servlet_log_exn(final CodeRunner ctxt,
                                                       final Value servlet,
                                                       final Value msg,
                                                       final Value exn)
        throws Fail.Exception {
        final Object e = exn.asBlock().asCustom();
        if (!(e instanceof Throwable)) {
            Fail.invalidArgument("Servlet.Servlet.log_exn");
        } // end if
        try {
            final HttpServlet hs = (HttpServlet) servlet.asBlock().asCustom();
            hs.log(msg.asBlock().asString(), (Throwable) e);
            return Value.UNIT;
        } catch (final Exception ce) {
            Cadmium.fail(ctxt, ce);
            return Value.UNIT; // never reached
        } // end try/catch
    } // end method 'cadmiumservlet_servlet_log_exn(CodeRunner, Value, Value, Value)'

} // end class 'Servlets'
