/*
 * This file is part of Cadmium.
 * Copyright (C) 2007-2010 Xavier Clerc.
 *
 * Cadmium is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * Cadmium is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package fr.x9c.cadmium.primitives.cadmiumswixml;

import java.awt.Component;
import java.awt.Container;
import java.awt.Frame;
import java.awt.event.ActionListener;
import java.io.File;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.net.URL;
import java.util.Iterator;
import java.util.Locale;
import java.util.Map;

import org.swixml.SwingEngine;

import fr.x9c.cadmium.kernel.Block;
import fr.x9c.cadmium.kernel.Channel;
import fr.x9c.cadmium.kernel.CodeRunner;
import fr.x9c.cadmium.kernel.Fail;
import fr.x9c.cadmium.kernel.Primitive;
import fr.x9c.cadmium.kernel.PrimitiveProvider;
import fr.x9c.cadmium.kernel.Value;
import fr.x9c.cadmium.primitives.cadmium.Cadmium;

/**
 * This class provides the primitives allowing use of
 * <a href="http://www.swixml.org">SwiXml</a> from Cadmium.
 *
 * @author <a href="mailto:cadmium@x9c.fr">Xavier Clerc</a>
 * @version 1.1
 * @since 1.1
 */
@PrimitiveProvider
public final class SwiXml {

    /**
     * No instance of this class.
     */
    private SwiXml() {
    } // end empty constructor

    /**
     * Tests whether SwiXml is in debug or release mode.
     * @param ctxt context
     * @param unit ignored
     * @return {@link fr.x9c.cadmium.kernel.Value#TRUE} if SwiXml is in debug
     *         mode, {@link fr.x9c.cadmium.kernel.Value#FALSE} otherwise
     */
    @Primitive
    public static Value cadmiumswixml_is_debug_mode(final CodeRunner ctxt,
                                                    final Value unit) {
        return SwingEngine.DEBUG_MODE ? Value.TRUE : Value.FALSE;
    } // end method 'cadmiumswixml_is_debug_mode(CodeRunner, Value)'

    /**
     * Tests whether the underlying platform is MacOS X.
     * @param ctxt context
     * @param unit ignored
     * @return {@link fr.x9c.cadmium.kernel.Value#TRUE} if the underlying
     *         platform is MacOS X, {@link fr.x9c.cadmium.kernel.Value#FALSE}
     *         otherwise
     * @throws Fail.Exception if SwiXml throws an exception
     */
    @Primitive
    public static Value cadmiumswixml_is_macosx(final CodeRunner ctxt,
                                                final Value unit)
        throws Fail.Exception {
        try {
            return SwingEngine.isMacOSX() ? Value.TRUE : Value.FALSE;
        } catch (final Exception e) {
            Cadmium.fail(ctxt, e);
            return Value.UNIT; // never reached
        } // end try/catch
    } // end method 'cadmiumswixml_is_macosx(CodeRunner, Value)'

    /**
     * Tests whether MacOS X support is enabled.
     * @param ctxt context
     * @param unit ignored
     * @return {@link fr.x9c.cadmium.kernel.Value#TRUE} if MacOS X support is
     *         enabled, {@link fr.x9c.cadmium.kernel.Value#FALSE} otherwise
     * @throws Fail.Exception if SwiXml throws an exception
     */
    @Primitive
    public static Value cadmiumswixml_is_macosx_supported(final CodeRunner ctxt,
                                                          final Value unit)
        throws Fail.Exception {
        try {
            return SwingEngine.isMacOSXSupported() ? Value.TRUE : Value.FALSE;
        } catch (final Exception e) {
            Cadmium.fail(ctxt, e);
            return Value.UNIT; // never reached
        } // end try/catch
    } // end method 'cadmiumswixml_is_macosx_supported(CodeRunner, Value)'

    /**
     * Sets MacOS X support.
     * @param ctxt context
     * @param support whether MacOS X support should be enabled
     * @return <i>unit</i>
     * @throws Fail.Exception if SwiXml throws an exception
     */
    @Primitive
    public static Value cadmiumswixml_set_macosx_support(final CodeRunner ctxt,
                                                         final Value support)
        throws Fail.Exception {
        try {
            SwingEngine.setMacOSXSuport(support == Value.TRUE);
            return Value.UNIT;
        } catch (final Exception e) {
            Cadmium.fail(ctxt, e);
            return Value.UNIT; // never reached
        } // end try/catch
    } // end method 'cadmiumswixml_set_macosx_support(CodeRunner, Value)'

    /**
     * Sets the default locale.
     * @param ctxt context
     * @param loc locale (should be a {@link java.util.Locale} instance)
     * @return <i>unit</i>
     * @throws Fail.Exception if SwiXml throws an exception
     * @throws Fail.Exception if a cast error occurs
     */
    @Primitive
    public static Value cadmiumswixml_set_default_locale(final CodeRunner ctxt,
                                                         final Value loc)
        throws Fail.Exception {
        try {
            SwingEngine.setDefaultLocale((Locale) loc.asBlock().asCustom());
            return Value.UNIT;
        } catch (final Exception e) {
            Cadmium.fail(ctxt, e);
            return Value.UNIT; // never reached
        } // end try/catch
    } // end method 'cadmiumswixml_set_default_locale(CodeRunner, Value)'

    /**
     * Sets the resource bundle name.
     * @param ctxt context
     * @param name resource bundle name
     * @return <i>unit</i>
     * @throws Fail.Exception if SwiXml throws an exception
     */
    @Primitive
    public static Value cadmiumswixml_set_resource_bundle_name(final CodeRunner ctxt,
                                                               final Value name)
        throws Fail.Exception {
        try {
            SwingEngine.setResourceBundleName(name.asBlock().asString());
            return Value.UNIT;
        } catch (final Exception e) {
            Cadmium.fail(ctxt, e);
            return Value.UNIT; // never reached
        } // end try/catch
    } // end method 'cadmiumswixml_set_resource_bundle_name(CodeRunner, Value)'

    /**
     * Sets the global application frame, used as the parent for child dialogs.
     * @param ctxt contet
     * @param appFrame frame to set as the global application frame
     *                 (should be a {@link java.awt.Frame} instance)
     * @return <i>unit</i>
     * @throws Fail.Exception if SwiXml throws an exception
     * @throws Fail.Exception if a cast error occurs
     */
    @Primitive
    public static Value cadmiumswixml_set_app_frame(final CodeRunner ctxt,
                                                    final Value appFrame)
        throws Fail.Exception {
        try {
            SwingEngine.setAppFrame((Frame) appFrame.asBlock().asCustom());
            return Value.UNIT;
        } catch (final Exception e) {
            Cadmium.fail(ctxt, e);
            return Value.UNIT; // never reached
        } // end try/catch
    } // end method 'cadmiumswixml_set_app_frame(CodeRunner, Value)'

    /**
     * Returns the global application frame.
     * @param ctxt context
     * @param unit ignored
     * @return the global application frame
     * @throws Fail.Exception if SwiXml throws an exception
     */
    @Primitive
    public static Value cadmiumswixml_get_app_frame(final CodeRunner ctxt,
                                                    final Value unit)
        throws Fail.Exception {
        try {
            return Cadmium.createObject(SwingEngine.getAppFrame());
        } catch (final Exception e) {
            Cadmium.fail(ctxt, e);
            return Value.UNIT; // never reached
        } // end try/catch
    } // end method 'cadmiumswixml_get_app_frame(CodeRunner, Value)'

    /**
     * Constructs an engine.
     * @param ctxt context
     * @param unit ignored
     * @return new engine
     * @throws Fail.Exception if SwiXml throws an exception
     */
    @Primitive
    public static Value cadmiumswixml_make_engine(final CodeRunner ctxt,
                                                  final Value unit)
        throws Fail.Exception {
        try {
            return Cadmium.createObject(new SwingEngine());
        } catch (final Exception e) {
            Cadmium.fail(ctxt, e);
            return Value.UNIT; // never reached
        } // end try/catch
    } // end method 'cadmiumswixml_make_engine(CodeRunner, Value)'

    /**
     * Constructs an engine from a resource (XML data describing UI).
     * @param ctxt context
     * @param resource resource name
     * @return new engine
     * @throws Fail.Exception if SwiXml throws an exception
     */
    @Primitive
    public static Value cadmiumswixml_make_engine_from_resource(final CodeRunner ctxt,
                                                                final Value resource)
        throws Fail.Exception {
        try {
            return Cadmium.createObject(new SwingEngine(resource.asBlock().asString()));
        } catch (final Exception e) {
            Cadmium.fail(ctxt, e);
            return Value.UNIT; // never reached
        } // end try/catch
    } // end method 'cadmiumswixml_make_engine_from_resource(CodeRunner, Value)'

    /**
     * Constructs an engine from a client object. Information will be
     * read/written from/to instance fields of the client using reflection.
     * @param ctxt context
     * @param client client
     * @return new engine
     * @throws Fail.Exception if SwiXml throws an exception
     */
    @Primitive
    public static Value cadmiumswixml_make_engine_from_client(final CodeRunner ctxt,
                                                              final Value client)
        throws Fail.Exception {
        try {
            return Cadmium.createObject(new SwingEngine(client.asBlock().asCustom()));
        } catch (final Exception e) {
            Cadmium.fail(ctxt, e);
            return Value.UNIT; // never reached
        } // end try/catch
    } // end method 'cadmiumswixml_make_engine_from_client(CodeRunner, Value)'

    /**
     * Returns the client object of an engine.
     * @param ctxt context
     * @param engine engine to get client from
     * @return client of passed engine
     * @throws Fail.Exception if SwiXml throws an exception
     */
    @Primitive
    public static Value cadmiumswixml_get_client(final CodeRunner ctxt,
                                                 final Value engine)
        throws Fail.Exception {
        try {
            return Cadmium.createObject(((SwingEngine) engine.asBlock().asCustom()).getClient());
        } catch (final Exception e) {
            Cadmium.fail(ctxt, e);
            return Value.UNIT; // never reached
        } // end try/catch
    } // end method 'cadmiumswixml_get_client(CodeRunner, Value)'

    /**
     * Sets the classloader of an engine.
     * @param ctxt context
     * @param engine engine to set classloader to
     * @param cl new class loader
     *           (should be a {@link java.lang.ClassLoader} instance)
     * @return <i>unit</i>
     * @throws Fail.Exception if SwiXml throws an exception
     * @throws Fail.Exception if a cast error occurs
     */
    @Primitive
    public static Value cadmiumswixml_set_classloader(final CodeRunner ctxt,
                                                      final Value engine,
                                                      final Value cl)
        throws Fail.Exception {
        try {
            final ClassLoader c =
                (java.lang.ClassLoader) cl.asBlock().asCustom();
            ((SwingEngine) engine.asBlock().asCustom()).setClassLoader(c);
            return Value.UNIT;
        } catch (final Exception e) {
            Cadmium.fail(ctxt, e);
            return Value.UNIT; // never reached
        } // end try/catch
    } // end method 'cadmiumswixml_set_classloader(CodeRunner, Value, Value)'

    /**
     * Returns the classloader of an engine.
     * @param ctxt context
     * @param engine engine to get classloader from
     * @return classloader of passed engine
     * @throws Fail.Exception if SwiXml throws an exception
     */
    @Primitive
    public static Value cadmiumswixml_get_classloader(final CodeRunner ctxt,
                                                      final Value engine)
        throws Fail.Exception {
        try {
            return Cadmium.createObject(((SwingEngine) engine.asBlock().asCustom()).getClassLoader());
        } catch (final Exception e) {
            Cadmium.fail(ctxt, e);
            return Value.UNIT; // never reached
        } // end try/catch
    } // end method 'cadmiumswixml_get_classloader(CodeRunner, Value)'

    /**
     * Sets the locale of an engine.
     * @param ctxt context
     * @param engine engine to set locale to
     * @param l new locale (should be a {@link java.util.Locale} instance)
     * @return <i>unit</i>
     * @throws Fail.Exception if SwiXml throws an exception
     * @throws Fail.Exception if a cast error occurs
     */
    @Primitive
    public static Value cadmiumswixml_set_locale(final CodeRunner ctxt,
                                                 final Value engine,
                                                 final Value l)
        throws Fail.Exception {
        try {
            ((SwingEngine) engine.asBlock().asCustom()).setLocale((Locale) l.asBlock().asCustom());
            return Value.UNIT;
        } catch (final Exception e) {
            Cadmium.fail(ctxt, e);
            return Value.UNIT; // never reached
        } // end try/catch
    } // end method 'cadmiumswixml_set_locale(CodeRunner, Value, Value)'

    /**
     * Sets the resource bundle of an engine.
     * @param ctxt context
     * @param engine engine to set resource bundle to
     * @param rb resource bundle name
     * @return <i>unit</i>
     * @throws Fail.Exception if SwiXml throws an exception
     */
    @Primitive
    public static Value cadmiumswixml_set_resource_bundle(final CodeRunner ctxt,
                                                          final Value engine,
                                                          final Value rb)
        throws Fail.Exception {
        try {
            ((SwingEngine) engine.asBlock().asCustom()).setResourceBundle(rb.asBlock().asString());
            return Value.UNIT;
        } catch (final Exception e) {
            Cadmium.fail(ctxt, e);
            return Value.UNIT; // never reached
        } // end try/catch
    } // end method 'cadmiumswixml_set_resource_bundle(CodeRunner, Value, Value)'

    /**
     * Renders a XML file into an UI.
     * @param ctxt context
     * @param engine engine to use for rendering
     * @param file file to render
     * @return the root component of the rendered UI
     * @throws Fail.Exception if SwiXml throws an exception
     */
    @Primitive
    public static Value cadmiumswixml_render_from_file(final CodeRunner ctxt,
                                                       final Value engine,
                                                       final Value file)
        throws Fail.Exception {
        try {
            final File f = ctxt.getContext().getRealFile(file);
            return Cadmium.createObject(((SwingEngine) engine.asBlock().asCustom()).render(f));
        } catch (final Exception e) {
            Cadmium.fail(ctxt, e);
            return Value.UNIT; // never reached
        } // end try/catch
    } // end method 'cadmiumswixml_render_from_file(CodeRunner, Value, Value)'

    /**
     * Renders a XML file into an UI.
     * @param ctxt context
     * @param engine engine to use for rendering
     * @param url url of file to render
     * @return the root component of the rendered UI
     * @throws Fail.Exception if SwiXml throws an exception
     * @throws Fail.Exception if the passed string is not a valid URL
     */
    @Primitive
    public static Value cadmiumswixml_render_from_url(final CodeRunner ctxt,
                                                      final Value engine,
                                                      final Value url)
        throws Fail.Exception {
        try {
            final URL u = new URL(url.asBlock().asString());
            return Cadmium.createObject(((SwingEngine) engine.asBlock().asCustom()).render(u));
        } catch (final Exception e) {
            Cadmium.fail(ctxt, e);
            return Value.UNIT; // never reached
        } // end try/catch
    } // end method 'cadmiumswixml_render_from_url(CodeRunner, Value, Value)'

    /**
     * Renders a XML file into an UI.
     * @param ctxt context
     * @param engine engine to use for rendering
     * @param resource resource file to render
     * @return the root component of the rendered UI
     * @throws Fail.Exception if SwiXml throws an exception
     */
    @Primitive
    public static Value cadmiumswixml_render_from_resource(final CodeRunner ctxt,
                                                           final Value engine,
                                                           final Value resource)
        throws Fail.Exception {
        try {
            final String r = resource.asBlock().asString();
            return Cadmium.createObject(((SwingEngine) engine.asBlock().asCustom()).render(r));
        } catch (final Exception e) {
            Cadmium.fail(ctxt, e);
            return Value.UNIT; // never reached
        } // end try/catch
    } // end method 'cadmiumswixml_render_from_resource(CodeRunner, Value, Value)'

    /**
     * Renders a XML file into an UI.
     * @param ctxt context
     * @param engine engine to use for rendering
     * @param channel channel to read XML data from
     * @return the root component of the rendered UI
     * @throws Fail.Exception if SwiXml throws an exception
     */
    @Primitive
    public static Value cadmiumswixml_render_from_channel(final CodeRunner ctxt,
                                                          final Value engine,
                                                          final Value channel)
        throws Fail.Exception {
        try {
            final InputStream in =
                ((Channel) channel.asBlock().asCustom()).asInputStream();
            final InputStreamReader reader = new InputStreamReader(in);
            return Cadmium.createObject(((SwingEngine) engine.asBlock().asCustom()).render(reader));
        } catch (final Exception e) {
            Cadmium.fail(ctxt, e);
            return Value.UNIT; // never reached
        } // end try/catch
    } // end method 'cadmiumswixml_render_from_channel(CodeRunner, Value, Value)'

    /**
     * Renders a XML file into UI elements and inserts them into a container.
     * @param ctxt context
     * @param engine engine to use for rendering
     * @param url url of file to render
     * @param container where rendered UI elements should be inserted
     *                  (should be a {@link java.awt.Container} instance)
     * @return <i>unit</i>
     * @throws Fail.Exception if SwiXml throws an exception
     * @throws Fail.Exception if a cast error occurs
     * @throws Fail.Exception if the passed string is not a valid URL
     */
    @Primitive
    public static Value cadmiumswixml_insert_from_url(final CodeRunner ctxt,
                                                      final Value engine,
                                                      final Value url,
                                                      final Value container)
        throws Fail.Exception {
        try {
            final URL u = new URL(url.asBlock().asString());
            final Container cnt = (Container) container.asBlock().asCustom();
            ((SwingEngine) engine.asBlock().asCustom()).insert(u, cnt);
            return Value.UNIT;
        } catch (final Exception e) {
            Cadmium.fail(ctxt, e);
            return Value.UNIT; // never reached
        } // end try/catch
    } // end method 'cadmiumswixml_insert_from_url(CodeRunner, Value, Value, Value)'

    /**
     * Renders a XML file into UI elements and inserts them into a container.
     * @param ctxt context
     * @param engine engine to use for rendering
     * @param resource resource file to render
     * @param container where rendered UI elements should be inserted
     *                  (should be a {@link java.awt.Container} instance)
     * @return <i>unit</i>
     * @throws Fail.Exception if SwiXml throws an exception
     * @throws Fail.Exception if a cast error occurs
     */
    @Primitive
    public static Value cadmiumswixml_insert_from_resource(final CodeRunner ctxt,
                                                           final Value engine,
                                                           final Value resource,
                                                           final Value container)
        throws Fail.Exception {
        try {
            final String res = resource.asBlock().asString();
            final Container cnt = (Container) container.asBlock().asCustom();
            ((SwingEngine) engine.asBlock().asCustom()).insert(res, cnt);
            return Value.UNIT;
        } catch (final Exception e) {
            Cadmium.fail(ctxt, e);
            return Value.UNIT; // never reached
        } // end try/catch
    } // end method 'cadmiumswixml_insert_from_resource(CodeRunner, Value, Value, Value)'

    /**
     * Renders a XML file into UI elements and inserts them into a container.
     * @param ctxt context
     * @param engine engine to use for rendering
     * @param channel channel to read XML data from
     * @param container where rendered UI elements should be inserted
     *                  (should be a {@link java.awt.Container} instance)
     * @return <i>unit</i>
     * @throws Fail.Exception if SwiXml throws an exception
     * @throws Fail.Exception if a cast error occurs
     */
    @Primitive
    public static Value cadmiumswixml_insert_from_channel(final CodeRunner ctxt,
                                                          final Value engine,
                                                          final Value channel,
                                                          final Value container)
        throws Fail.Exception {
        try {
            final InputStream in =
                ((Channel) channel.asBlock().asCustom()).asInputStream();
            final Container cnt = (Container) container.asBlock().asCustom();
            ((SwingEngine) engine.asBlock().asCustom()).insert(new InputStreamReader(in), cnt);
            return Value.UNIT;
        } catch (final Exception e) {
            Cadmium.fail(ctxt, e);
            return Value.UNIT; // never reached
        } // end try/catch
    } // end method 'cadmiumswixml_insert_from_channel(CodeRunner, Value, Value, Value)'

    /**
     * Returns the root component of the rendered UI of an engine.
     * @param ctxt context
     * @param engine engine to get root component from
     * @return the root component of the rendered UI of the passed engine
     *         ({@link java.awt.Container} instance)
     * @throws Fail.Exception if SwiXml throws an exception
     */
    @Primitive
    public static Value cadmiumswixml_get_root_component(final CodeRunner ctxt,
                                                         final Value engine)
        throws Fail.Exception {
        try {
            return Cadmium.createObject(((SwingEngine) engine.asBlock().asCustom()).getRootComponent());
        } catch (final Exception e) {
            Cadmium.fail(ctxt, e);
            return Value.UNIT; // never reached
        } // end try/catch
    } // end method 'cadmiumswixml_get_root_component(CodeRunner, Value)'

    /**
     * Returns the identifier/component map of an engine.
     * @param ctxt context
     * @param engine engine to get map from
     * @return a list of <i>(id, component)</i> couples
     * @throws Fail.Exception if SwiXml throws an exception
     */
    @Primitive
    public static Value cadmiumswixml_get_id_map(final CodeRunner ctxt,
                                                 final Value engine)
        throws Fail.Exception {
        try {
            Value res = Value.EMPTY_LIST;
            final Map<?, ?> m =
                ((SwingEngine) engine.asBlock().asCustom()).getIdMap();
            for (Map.Entry<?, ?> e : m.entrySet()) {
                final Block cpl =
                    Block.createBlock(0,
                                      Value.createFromBlock(Block.createString((String) e.getKey())),
                                      Cadmium.createObject(e.getValue()));
                final Block b = Block.createBlock(Block.TAG_CONS,
                                                  Value.createFromBlock(cpl),
                                                  res);
                res = Value.createFromBlock(b);
            } // end for
            return res;
        } catch (final Exception e) {
            Cadmium.fail(ctxt, e);
            return Value.UNIT; // never reached
        } // end try/catch
    } // end method 'cadmiumswixml_get_id_map(CodeRunner, Value)'

    /**
     * Clears the identifier/component map of an engine.
     * @param ctxt context
     * @param engine engine whose map should be cleared
     * @return <i>unit</i>
     * @throws Fail.Exception if SwiXml throws an exception
     */
    @Primitive
    public static Value cadmiumswixml_clear_id_map(final CodeRunner ctxt,
                                                   final Value engine)
        throws Fail.Exception {
        try {
            ((SwingEngine) engine.asBlock().asCustom()).getIdMap().clear();
            return Value.UNIT;
        } catch (final Exception e) {
            Cadmium.fail(ctxt, e);
            return Value.UNIT; // never reached
        } // end try/catch
    } // end method 'cadmiumswixml_clear_id_map(CodeRunner, Value)'

    /**
     * Returns the list of all components of an engine.
     * @param ctxt context
     * @param engine engine to get component list from
     * @return the list of all components of the passed engine
     * @throws Fail.Exception if SwiXml throws an exception
     */
    @Primitive
    public static Value cadmiumswixml_get_all_components(final CodeRunner ctxt,
                                                         final Value engine)
        throws Fail.Exception {
        try {
            Value res = Value.EMPTY_LIST;
            final Iterator it =
                ((SwingEngine) engine.asBlock().asCustom()).getAllComponentItertor();
            while (it.hasNext()) {
                final Block b =
                    Block.createBlock(Block.TAG_CONS,
                                      Cadmium.createObject(it.next()),
                                      res);
                res = Value.createFromBlock(b);
            } // end while
            return res;
        } catch (final Exception e) {
            Cadmium.fail(ctxt, e);
            return Value.UNIT; // never reached
        } // end try/catch
    } // end method 'cadmiumswixml_get_all_components(CodeRunner, Value)'

    /**
     * Returns the list of all descendants of a component of an engine.
     * @param ctxt context
     * @param engine engine to get component list from
     * @param component component to get descendants from
     *                  (should be a {@link java.awt.Component} instance)
     * @return the list of all descendants of the passed component of the
     *         passed engine
     * @throws Fail.Exception if SwiXml throws an exception
     * @throws Fail.Exception if a cast error occurs
     */
    @Primitive
    public static Value cadmiumswixml_get_descendants(final CodeRunner ctxt,
                                                      final Value engine,
                                                      final Value component)
        throws Fail.Exception {
        try {
            Value res = Value.EMPTY_LIST;
            final Component c = (Component) component.asBlock().asCustom();
            final Iterator it =
                ((SwingEngine) engine.asBlock().asCustom()).getDescendants(c);
            while (it.hasNext()) {
                final Block b =
                    Block.createBlock(Block.TAG_CONS,
                                      Cadmium.createObject(it.next()),
                                      res);
                res = Value.createFromBlock(b);
            } // end while
            return res;
        } catch (final Exception e) {
            Cadmium.fail(ctxt, e);
            return Value.UNIT; // never reached
        } // end try/catch
    } // end method 'cadmiumswixml_get_descendants(CodeRunner, Value, Value)'

    /**
     * Looks up for a component.
     * @param ctxt context
     * @param engine engine where search should be performed
     * @param id component identifier
     * @return the component with passed id
     * @throws Fail.Exception <i>Not_found</i> if such a component does not
     *                        exist
     * @throws Fail.Exception if SwiXml throws an exception
     */
    @Primitive
    public static Value cadmiumswixml_find(final CodeRunner ctxt,
                                           final Value engine,
                                           final Value id)
        throws Fail.Exception {
        final Component res;
        try {
            res = ((SwingEngine) engine.asBlock().asCustom()).find(id.asBlock().asString());
        } catch (final Exception e) {
            Cadmium.fail(ctxt, e);
            return Value.UNIT; // never reached
        } // end try/catch
        if (res != null) {
            return Cadmium.createObject(res);
        } else {
            Fail.raiseNotFound();
            return Value.UNIT; // never reached
        } // end if/else
    } // end method 'cadmiumswixml_find(CodeRunner, Value, Value)'

    /**
     * Forgets a component from an engine.
     * @param ctxt context
     * @param engine engine that should forget component
     * @param id identifier of component to forget
     * @return <i>unit</i>
     * @throws Fail.Exception if SwiXml throws an exception
     */
    @Primitive
    public static Value cadmiumswixml_forget(final CodeRunner ctxt,
                                             final Value engine,
                                             final Value id)
        throws Fail.Exception {
        try {
            ((SwingEngine) engine.asBlock().asCustom()).forget(id.asBlock().asString());
            return Value.UNIT;
        } catch (final Exception e) {
            Cadmium.fail(ctxt, e);
            return Value.UNIT; // never reached
        } // end try/catch
    } // end method 'cadmiumswixml_forget(CodeRunner, Value, Value)'

    /**
     * Forgets all undisplayable components of an engine.
     * @param ctxt context
     * @param engine engine to clean
     * @return the number of forgotten components
     * @throws Fail.Exception if SwiXml throws an exception
     */
    @Primitive
    public static Value cadmiumswixml_cleanup(final CodeRunner ctxt,
                                              final Value engine)
        throws Fail.Exception {
        try {
            final int res =
                ((SwingEngine) engine.asBlock().asCustom()).cleanup();
            return Value.createFromLong(Math.min(res, Value.MAX_LONG));
        } catch (final Exception e) {
            Cadmium.fail(ctxt, e);
            return Value.UNIT; // never reached
        } // end try/catch
    } // end method 'cadmiumswixml_cleanup(CodeRunner, Value)'

    /**
     * Sets an action listener for a component in an engine.
     * @param ctxt context
     * @param engine parent engine of component
     * @param component component to register an action listener with
     *                  (should be a {@link java.awt.Component} instance)
     * @param listener listener to register with component
     *                 (should be a {@link java.awt.event.ActionListener} instance)
     * @return {@link fr.x9c.cadmium.kernel.Value#TRUE} if the listener has
     *         been successfully registered with the component,
     *         {@link fr.x9c.cadmium.kernel.Value#FALSE} otherwise
     * @throws Fail.Exception if SwiXml throws an exception
     * @throws Fail.Exception if a cast error occurs
     */
    @Primitive
    public static Value cadmiumswixml_set_action_listener(final CodeRunner ctxt,
                                                          final Value engine,
                                                          final Value component,
                                                          final Value listener)
        throws Fail.Exception {
        try {
            final Component c = (Component) component.asBlock().asCustom();
            final ActionListener al = (ActionListener) listener.asBlock().asCustom();
            return (((SwingEngine) engine.asBlock().asCustom()).setActionListener(c, al)
                    ? Value.TRUE
                    : Value.FALSE);
        } catch (final Exception e) {
            Cadmium.fail(ctxt, e);
            return Value.UNIT; // never reached
        } // end try/catch
    } // end method 'cadmiumswixml_set_action_listener(CodeRunner, Value, Value, Value)'

} // end class 'SwiXml'
