/*
 * This file is part of Cadmium.
 * Copyright (C) 2007-2010 Xavier Clerc.
 *
 * Cadmium is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * Cadmium is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package fr.x9c.cadmium.primitives.labltk;

import fr.x9c.cadmium.kernel.Block;
import fr.x9c.cadmium.kernel.CadmiumException;
import fr.x9c.cadmium.kernel.CodeRunner;
import fr.x9c.cadmium.kernel.Fail;
import fr.x9c.cadmium.kernel.FalseExit;
import fr.x9c.cadmium.kernel.Fatal;
import fr.x9c.cadmium.kernel.Value;

import tcl.lang.Command;
import tcl.lang.Interp;
import tcl.lang.TclException;
import tcl.lang.TclInteger;
import tcl.lang.TclObject;

/**
 * This class implements the Tcl command used to invoke Caml callbacks.
 *
 * @author <a href="mailto:cadmium@x9c.fr">Xavier Clerc</a>
 * @version 1.3
 * @since 1.3
 */
public final class CamlCallbackCommand implements Command {

    /** Code runner used to execute callbacks. */
    private final CodeRunner codeRunner;

    /** Slot used to retrieve handler. */
    private final TKSlot slot;

    /**
     * Builds a Tcl command used to invoke Caml callbacks.
     * @param cr used to execute callbacks - should not be <tt>null</tt>
     * @param s used to retrieve handler - should not be <tt>null</tt>
     */
    public CamlCallbackCommand(final CodeRunner cr, final TKSlot s) {
        assert cr != null : "null cr";
        assert s != null : "null s";
        this.codeRunner= cr;
        this.slot = s;
    } // end constructor(CodeRunner, TKSlot)

    /**
     * Actually execute the callback command.
     * @param interp interpreter to execute command in
     * @param objv arguments to callback command
     * @throws TclException if any error occurs in callback
     */
    public void cmdProc(final Interp interp, final TclObject[] objv)
        throws TclException {
        interp.resetResult();
        if (objv.length >= 2) {
            final int id = TclInteger.get(interp, objv[1]);
            Value args = Value.EMPTY_LIST;
            for (int i = objv.length - 2 - 1; i >= 0; i--) {
                final Block s = Block.createString(objv[i + 2].toString());
                final Block tmp = Block.createBlock(Block.TAG_CONS,
                                                    Value.createFromBlock(s),
                                                    args);
                args = Value.createFromBlock(tmp);
            } // end for
            try {
                this.codeRunner.callback(this.slot.getHandlerCode(),
                                         Value.createFromLong(id),
                                         args);
            } catch (final Fail.Exception fe) {
                throw new TclException(interp, fe.getMessage());
            } catch (final Fatal.Exception fe) {
                throw new TclException(interp, fe.getMessage());
            } catch (final FalseExit fe) {
                throw new TclException(interp, fe.getMessage());
            } catch (final CadmiumException ce) {
                throw new TclException(interp, ce.getMessage());
            } // end try/catch
        } else {
            throw new TclException(interp, "CamlCallbackCommand.cmdProc(-) received an invalid number of parameters");
        } // end if/else
    } // end method 'cmdProc(Interp, TclObject[])'

} // end class 'CamlCallbackCommand'
