/*
 * This file is part of Cadmium.
 * Copyright (C) 2007-2010 Xavier Clerc.
 *
 * Cadmium is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * Cadmium is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package fr.x9c.cadmium.primitives.labltk;

import fr.x9c.cadmium.kernel.CadmiumException;
import fr.x9c.cadmium.kernel.CodeRunner;
import fr.x9c.cadmium.kernel.Fail;
import fr.x9c.cadmium.kernel.FalseExit;
import fr.x9c.cadmium.kernel.Fatal;
import fr.x9c.cadmium.kernel.Value;

import tcl.lang.Interp;
import tcl.lang.TCL;
import tcl.lang.TclException;
import tcl.lang.VarTrace;

/**
 * This class implements the handler of a variable tracer.
 *
 * @author <a href="mailto:cadmium@x9c.fr">Xavier Clerc</a>
 * @version 1.3
 * @since 1.3
 */
public final class CamlVarTrace implements VarTrace {

    /** Code runner used to execute callback. */
    private final CodeRunner codeRunner;

    /** Slot used to retrieve handler. */
    private final TKSlot slot;

    /** Callback identifier. */
    private final int id;

    /**
     * Builds a Tcl variable tracer.
     * @param cr used to execute callback - should not be <tt>null</tt>
     * @param s used to retrieve handler - should not be <tt>null</tt>
     * @param i callback identifier
     */
    public CamlVarTrace(final CodeRunner cr, final TKSlot s, final int i) {
        assert cr != null : "null cr";
        assert s != null : "null s";
        this.codeRunner= cr;
        this.slot = s;
        this.id = i;
    } // end constructor(CodeRunner, TKSlot, int)

    /**
     * Method called when a traced variable is modified.
     * @param interp interpreter
     * @param name1 first part of the variable name
     * @param name2 second part of the variable name
     * @param flags kind of trace
     * @throws TclException if the callback throws an exception
     */
    public void traceProc(final Interp interp,
                          final String name1,
                          final String name2,
                          final int flags)
        throws TclException {
        interp.untraceVar(name1,
                          name2,
                          this,
                          TCL.GLOBAL_ONLY | TCL.TRACE_WRITES | TCL.TRACE_UNSETS);
        try {
            this.codeRunner.callback(this.slot.getHandlerCode(),
                                     Value.createFromLong(this.id),
                                     Value.EMPTY_LIST);
        } catch (final Fail.Exception fe) {
            throw new TclException(interp, fe.getMessage());
        } catch (final Fatal.Exception fe) {
            throw new TclException(interp, fe.getMessage());
        } catch (final FalseExit fe) {
            throw new TclException(interp, fe.getMessage());
        } catch (final CadmiumException ce) {
            throw new TclException(interp, ce.getMessage());
        } // end try/catch
    } // end method 'traceProc(Interp, String, String, int)'

} // end class 'CamlVarTrace'
