/*
 * This file is part of Cadmium.
 * Copyright (C) 2007-2010 Xavier Clerc.
 *
 * Cadmium is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * Cadmium is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package fr.x9c.cadmium.primitives.labltk;

import java.util.ArrayList;
import java.util.List;

import fr.x9c.cadmium.kernel.Block;
import fr.x9c.cadmium.kernel.CodeRunner;
import fr.x9c.cadmium.kernel.Fail;
import fr.x9c.cadmium.kernel.Primitive;
import fr.x9c.cadmium.kernel.PrimitiveProvider;
import fr.x9c.cadmium.kernel.Value;

import tcl.lang.Command;
import tcl.lang.Interp;
import tcl.lang.TclException;
import tcl.lang.TclObject;
import tcl.lang.TclList;
import tcl.lang.TclString;

/**
 * Implements all primitives from 'cltkEval.c'.
 *
 * @author <a href="mailto:cadmium@x9c.fr">Xavier Clerc</a>
 * @version 1.3
 * @since 1.3
 */
@PrimitiveProvider
public final class CltkEval {

    /**
     * No instance of this class.
     */
    private CltkEval() {
    } // end empty constructor

    /**
     * Evaluates a Tcl command.
     * @param ctxt context
     * @param str command to evaluate
     * @return the result of the evaluation, as a string
     * @throws Fail.Exception if any error occurs during evaluation
     */
    @Primitive
    public static Value camltk_tcl_eval(final CodeRunner ctxt,
                                        final Value str)
        throws Fail.Exception {
        final TKSlot slot = TKSlot.checkInit(ctxt);
        final Interp interp = slot.getInterpreter();
        interp.resetResult();
        try {
            Misc.eval(interp, str.asBlock().asString());
        } catch (final TclException te) {
            slot.tkError(te);
        } // end try/catch
        final Block res = Block.createString(interp.getResult().toString());
        return Value.createFromBlock(res);
    } // end method 'camltk_tcl_eval(CodeRunner, Value)'

    /**
     * Builds a list of arguments.
     * @param slot context
     * @param argv list to add arguments to
     * @param v <i>tkArgs</i> value to add to list
     * @throws TclException if an error occurs while appending a Tcl list
     * @throws Fail.Exception if the passed value is invalid
     */
    static void fillArgs(final TKSlot slot, final List<TclObject> argv, final Value v)
        throws TclException, Fail.Exception {
        final Block b = v.asBlock();
        switch (b.getTag()) {
        case 0:
            final String s = b.get(0).asBlock().asString();
            argv.add(TclString.newInstance(s));
            break;
        case 1:
            Value l = b.get(0);
            while (l.isBlock()) {
                final Block tmp = l.asBlock();
                fillArgs(slot, argv, tmp.get(0));
                l = tmp.get(1);
            } // end while
            break;
        case 2:
            final List<TclObject> tmpArgv = new ArrayList<TclObject>();
            fillArgs(slot, tmpArgv, b.get(0));
            final TclObject elem = TclList.newInstance();
            for (TclObject o : tmpArgv) {
                TclList.append(slot.getInterpreter(), elem, o);
            } // end for
            argv.add(elem);
            break;
        default:
            slot.tkError("fillArgs: illegal tag");
        } // end switch
    } // end method 'fillArgs(TKSlot, List<TclObject>, Value)'

    /**
     * Evaluates a Tcl command from an array of arguments.
     * @param ctxt context
     * @param v arguments
     * @return the result of the evaluation, as a string
     * @throws Fail.Exception if any error occurs during evaluation
     */
    @Primitive
    public static Value camltk_tcl_direct_eval(final CodeRunner ctxt,
                                               final Value v)
        throws Fail.Exception {
        final TKSlot slot = TKSlot.checkInit(ctxt);
        final Interp interp = slot.getInterpreter();
        final Block b = v.asBlock();
        final int wosize = b.getWoSize();
        final List<TclObject> argv = new ArrayList<TclObject>();
        try {
            for (int i = 0; i < wosize; i++) {
                fillArgs(slot, argv, b.get(i));
            } // end for
        } catch (final TclException te) {
            slot.tkError("fillArgs error");
        } // end try/catch
        final TclObject[] args = argv.toArray(new TclObject[argv.size()]);
        interp.resetResult();
        final Command cmd = interp.getCommand(args[0].toString());
        if (cmd != null) {
            try {
                cmd.cmdProc(interp, args);
            } catch (final TclException te) {
                slot.tkError(te);
            } // end try/catch
        } else {
            final Command unkCmd = interp.getCommand("unknown");
            if (unkCmd != null) {
                final int len = args.length;
                final TclObject[] newArgs = new TclObject[len + 1];
                System.arraycopy(args, 0, newArgs, 1, len);
                newArgs[0] = TclString.newInstance("unknown");
                try {
                    unkCmd.cmdProc(interp, newArgs);
                } catch (final TclException te) {
                    slot.tkError(te);
                } // end try/catch
            } else {
                final String msg = String.format("Unknown command \"%s\"", args[0]);
                interp.setResult(TclString.newInstance(msg));
            } // end if/else
        } // end if/else
        final Block res = Block.createString(interp.getResult().toString());
        return Value.createFromBlock(res);
    } // end method 'camltk_tcl_direct_eval(CodeRunner, Value)'

} // end class 'CltkEval'
