/*
 * This file is part of Cadmium.
 * Copyright (C) 2007-2010 Xavier Clerc.
 *
 * Cadmium is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * Cadmium is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package fr.x9c.cadmium.primitives.labltk;

import java.io.InputStream;
import java.io.IOException;

import fr.x9c.cadmium.kernel.CadmiumException;
import fr.x9c.cadmium.kernel.CodeRunner;
import fr.x9c.cadmium.kernel.FalseExit;
import fr.x9c.cadmium.kernel.Fail;
import fr.x9c.cadmium.kernel.Fatal;
import fr.x9c.cadmium.kernel.Value;

/**
 * This class implements a thread that waits for input on a stream.
 *
 * @author <a href="mailto:cadmium@x9c.fr">Xavier Clerc</a>
 * @version 1.3
 * @since 1.3
 */
public final class InputWaiterThread extends Thread {

    /** Code runner used to execute callback. */
    private final CodeRunner codeRunner;

    /** Slot used to retrieve handler. */
    private final TKSlot slot;

    /** Callback identifier. */
    private final int id;

    /** Stream to wait on. */
    private final InputStream stream;

    /**
     * Builds a thread waiting for input on a stream. <br/>
     * The thread is not started.
     * @param cr used to execute callback - should not be <tt>null</tt>
     * @param s used to retrieve handler - should not be <tt>null</tt>
     * @param i callback identifier 
     * @param st stream to wait on
     */
    public InputWaiterThread(final CodeRunner cr,
                             final TKSlot s,
                             final int i,
                             final InputStream st) {
        assert cr != null : "null cr";
        assert s != null : "null s";
        assert st != null : "null st";
        this.codeRunner = cr;
        this.slot = s;
        this.id = i;
        this.stream = st;
    } // end constructor(CodeRunner, TKSlot, int, InputStream)

    /**
     * Waits for input on a stream, regularly polling the stream. <br/>
     * The callback is called when data is available on the stream. <br/>
     * The thread should be interrupted to cancel the wait.
     */
    public void run() {
        try {
            while (true) {
                synchronized (this) {
                    wait(50);
                } // end synchronized
                int available = 0;
                try {
                    available = this.stream.available();
                } catch (final IOException ioe) {
                    // nothing to do
                } // end try/catch
                if (available > 0) {
                    try {
                        this.codeRunner.callback(this.slot.getHandlerCode(),
                                                 Value.createFromLong(this.id),
                                                 Value.EMPTY_LIST);
                    } catch (final CadmiumException fe) {
                        // nothing to do
                    } catch (final FalseExit fe) {
                        // nothing to do
                    } catch (final Fatal.Exception fe) {
                        // nothing to do
                    } catch (final Fail.Exception fe) {
                        // nothing to do
                    } // end try/catch
                } // end if
            } // end while
        } catch (final InterruptedException ie) {
            // nothing to do
        } // end try/catch
    } // end method 'run()'

} // end class 'InputWaiterThread'
