/*
 * This file is part of Cadmium.
 * Copyright (C) 2007-2010 Xavier Clerc.
 *
 * Cadmium is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * Cadmium is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package fr.x9c.cadmium.primitives.labltk;

import tcl.lang.Interp;
import tcl.lang.TCL;
import tcl.lang.TclEvent;
import tcl.lang.TclException;

/**
 * This class provides utility methods related to the 'labltk' library.
 *
 * @author <a href="mailto:cadmium@x9c.fr">Xavier Clerc</a>
 * @version 1.3
 * @since 1.3
 */
final class Misc {

    /**
     * No instance of this class.
     */
    private Misc() {
    } // end empty constructor

    /**
     * Evaluates a Tcl command, respecting thread constraints.
     * @param interp interpreter used for evaluation - should not be <tt>null</tt>
     * @param cmd command to evaluate - should not be <tt>null</tt>
     * @throws TclException if any error occurs during evaluation
     */
    static void eval(final Interp interp, final String cmd)
        throws TclException {
        assert interp != null : "null interp";
        assert cmd != null : "null cmd";
        final TclException[] tmp = new TclException[] { null };
        final TclEvent event = new TclEvent() {
                public int processEvent(final int flags) {
                    try {
                        interp.eval(cmd);
                    } catch (final TclException te) {
                        tmp[0] = te;
                    } // end try/catch
                    return 1;
                } // end method 'processEvent(int)'
            };
        interp.getNotifier().queueEvent(event, TCL.QUEUE_TAIL);
        event.sync();
        throwIfRelevant(tmp[0]);
    } // end method 'eval(Interp, String)'

    /**
     * Evaluates a list of Tcl commands from a file, respecting thread constraints.
     * @param interp interpreter used for evaluation - should not be <tt>null</tt>
     * @param file name of file to load commands from - should not be <tt>null</tt>
     * @throws TclException if any error occurs during evaluation
     */
    static void evalFile(final Interp interp, final String file)
        throws TclException {
        assert interp != null : "null interp";
        assert file != null : "null file";
        final TclException[] tmp = new TclException[] { null };
        final TclEvent event = new TclEvent() {
                public int processEvent(final int flags) {
                    try {
                        interp.evalFile(file);
                    } catch (final TclException te) {
                        tmp[0] = te;
                    } // end try/catch
                    return 1;
                } // end method 'processEvent(int)'
            };
        interp.getNotifier().queueEvent(event, TCL.QUEUE_TAIL);
        event.sync();
        throwIfRelevant(tmp[0]);
    } // end method 'evalFile(Interp, String)'

    /**
     * Evaluates a list of Tcl commands from a resource, respecting thread constraints.
     * @param interp interpreter used for evaluation - should not be <tt>null</tt>
     * @param res name of resource to load commands from - should not be <tt>null</tt>
     * @throws TclException if any error occurs during evaluation
     */
    static void evalResource(final Interp interp, final String res)
        throws TclException {
        assert interp != null : "null interp";
        assert res != null : "null res";
        final TclException[] tmp = new TclException[] { null };
        final TclEvent event = new TclEvent() {
                public int processEvent(final int flags) {
                    try {
                        interp.evalResource(res);
                    } catch (final TclException te) {
                        tmp[0] = te;
                    } // end try/catch
                    return 1;
                } // end method 'processEvent(int)'
            };
        interp.getNotifier().queueEvent(event, TCL.QUEUE_TAIL);
        event.sync();
        throwIfRelevant(tmp[0]);
    } // end method 'evalResource(Interp, String)'

    /**
     * Disposes (that is <i>uninitializes</i>) an interpreter.
     * @param interp interpreter to dispose - should not be <tt>null</tt>
     */
    static void disposeInterp(final Interp interp) {
        assert interp != null : "null interp";
        final TclEvent event = new TclEvent() {
                public int processEvent(final int flags) {
                    interp.dispose();
                    return 1;
                } // end method 'processEvent(int)'
            };
        interp.getNotifier().queueEvent(event, TCL.QUEUE_TAIL);
        event.sync();
    } // end method 'disposeInterp(Interp)'

    /**
     * Throws the passed exception if it is considered <i>relevant</i>. <br/>
     * <b>This is an awful hack to circumvent some clearly identified
     * incompatibilities between original Tcl/Tk and Jacl/Swank.</b>
     * @param te exception to throw if relevant
     * @throws TclException if relevant
     */
    private static void throwIfRelevant(final TclException te)
        throws TclException {
        final boolean relevant =
            (te != null)
            && !te.getMessage().startsWith("invalid binding type");
        if (relevant) {
            throw te;
        } // end if
    } // end method 'throwIfRelevant(TclException)'

} // end class 'Misc'
