/*
 * This file is part of Cadmium.
 * Copyright (C) 2007-2010 Xavier Clerc.
 *
 * Cadmium is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * Cadmium is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package fr.x9c.cadmium.primitives.stdlib;

import fr.x9c.cadmium.kernel.Block;
import fr.x9c.cadmium.kernel.CodeRunner;
import fr.x9c.cadmium.kernel.Fail;
import fr.x9c.cadmium.kernel.Primitive;
import fr.x9c.cadmium.kernel.PrimitiveProvider;
import fr.x9c.cadmium.kernel.Value;

/**
 * Implements all primitives from 'finalise.c'.
 *
 * @author <a href="mailto:cadmium@x9c.fr">Xavier Clerc</a>
 * @version 1.0
 * @since 1.0
 */
@PrimitiveProvider
public final class Finalise {

    /**
     * No instance of this class.
     */
    private Finalise() {
    } // end empty constructor

    /**
     * Registers a finalizer for a given value.
     * @param ctxt context
     * @param f finalizer
     * @param v value
     * @return <i>unit</i>
     * @throws Fail.Exception if value is not a block
     */
    @Primitive
    public static Value caml_final_register(final CodeRunner ctxt,
                                            final Value f,
                                            final Value v)
        throws Fail.Exception {
        if (!v.isBlock()) {
            Fail.invalidArgument("Gc.finalise");
        } // end if
        Block b = v.asBlock();
        if (b.getTag() == Block.INFIX_TAG) {
            b = b.getParent();
        } // end if
        b.addFinalizer(f, ctxt);
        return Value.UNIT;
    } // end method 'caml_final_register(CodeRunner, Value, Value)'

    /**
     * Release the finalization lock.
     * @param ctxt context
     * @param unit ignored
     * @return <i>unit</i>
     */
    @Primitive
    public static Value caml_final_release(final CodeRunner ctxt,
                                           final Value unit) {
        ctxt.getContext().releaseFinalizeLock();
        return Value.UNIT;
    } // end method 'caml_final_release(CodeRunner, Value)'

} // end class 'Finalise'
