/*
 * This file is part of Cadmium.
 * Copyright (C) 2007-2010 Xavier Clerc.
 *
 * Cadmium is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * Cadmium is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package fr.x9c.cadmium.primitives.unix;

import java.io.IOException;
import java.io.InterruptedIOException;
import java.net.InetAddress;
import java.net.InetSocketAddress;

import fr.x9c.cadmium.kernel.Block;
import fr.x9c.cadmium.kernel.Channel;
import fr.x9c.cadmium.kernel.CodeRunner;
import fr.x9c.cadmium.kernel.Fail;
import fr.x9c.cadmium.kernel.FalseExit;
import fr.x9c.cadmium.kernel.Primitive;
import fr.x9c.cadmium.kernel.PrimitiveProvider;
import fr.x9c.cadmium.kernel.Value;

/**
 * This class provides implementation for 'unix_bind' primitive.
 *
 * @author <a href="mailto:cadmium@x9c.fr">Xavier Clerc</a>
 * @version 1.0
 * @since 1.0
 */
@PrimitiveProvider
public final class Bind {

    /**
     * No instance of this class.
     */
    private Bind() {
    } // end empty constructor

    /**
     * Binds a socket to an address and port.
     * @param ctxt context
     * @param socket socket to bind
     * @param addr <i>inet_addr, port</i> to bind to
     * @return <i>unit</i>
     * @throws Fail.Exception if bind fails
     */
    @Primitive
    public static Value unix_bind(final CodeRunner ctxt,
                                  final Value socket,
                                  final Value addr)
        throws Fail.Exception, FalseExit {
        final Channel ch = ctxt.getContext().getChannel(socket.asLong());
        if (ch == null) {
            Unix.fail(ctxt, "bind", Unix.INVALID_DESCRIPTOR_MSG);
            return Value.UNIT; // never reached
        } // end if
        final Block addrBlock = addr.asBlock();
        try {
            ch.socketBind(new InetSocketAddress(InetAddress.getByAddress(addrBlock.get(0).asBlock().getBytes()),
                                                addrBlock.get(1).asLong()));
        } catch (final InterruptedIOException iioe) {
            final FalseExit fe =
                FalseExit.createFromContext(ctxt.getContext());
            fe.fillInStackTrace();
            throw fe;
        } catch (final IOException ioe) {
            Unix.fail(ctxt, "bind", ioe);
        } // end try/catch
        return Value.UNIT;
    } // end method 'unix_bind(CodeRunner, Value, Value)'

} // end class 'Bind'
