/*
 * This file is part of Cadmium.
 * Copyright (C) 2007-2010 Xavier Clerc.
 *
 * Cadmium is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * Cadmium is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package fr.x9c.cadmium.support.servlet;

import java.io.IOException;

import javax.servlet.ServletException;
import javax.servlet.http.HttpServlet;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;

import fr.x9c.cadmium.kernel.Context;
import fr.x9c.cadmium.kernel.Value;
import fr.x9c.cadmium.primitives.cadmium.Cadmium;

/**
 * This class implements a servlet whose behaviour is defined by callbacks
 * registered on the OCaml side. Actual callback manipulation is delegated
 * to children classes in such a way that this class is the basis of both
 * bytecode and native servlet runners.
 *
 * @author <a href="mailto:cadmium@x9c.fr">Xavier Clerc</a>
 * @version 1.1
 * @since 1.1
 */
public abstract class Servlet extends HttpServlet {

    /**
     * Empty constructor.
     */
    public Servlet() {
        super();
    } // end empty constructor

    /**
     * {@inheritDoc}
     */
    @Override
    public final void init() throws ServletException {
        callback("Servlet.init", Cadmium.createObject(this));
    } // end method 'init()'

    /**
     * {@inheritDoc}
     */
    @Override
    public final void destroy() {
        callbackNoException("Servlet.destroy", Cadmium.createObject(this));
    } // end method 'destroy()'

    /**
     * {@inheritDoc}
     */
    @Override
    public final String getServletInfo() {
        final Value cb = getCallback("Servlet.info");
        return cb != null ? cb.asBlock().asString() : null;
    } // end method 'getServletInfo()'

    /**
     * {@inheritDoc}
     */
    @Override
    protected final void doDelete(final HttpServletRequest req,
                                  final HttpServletResponse resp)
        throws IOException, ServletException {
        if (callbackOption("Servlet.do_delete",
                           Cadmium.createObject(this),
                           Cadmium.createObject(req),
                           Cadmium.createObject(resp))) {
            super.doDelete(req, resp);
        } // end if
    } // end method 'doDelete(HttpServletRequest, HttpServletResponse)'

    /**
     * {@inheritDoc}
     */
    @Override
    protected final void doGet(final HttpServletRequest req,
                               final HttpServletResponse resp)
        throws IOException, ServletException {
        if (callbackOption("Servlet.do_get",
                           Cadmium.createObject(this),
                           Cadmium.createObject(req),
                           Cadmium.createObject(resp))) {
            super.doGet(req, resp);
        } // end if
    } // end method 'doGet(HttpServletRequest, HttpServletResponse)'

    /**
     * {@inheritDoc}
     */
    @Override
    protected final void doHead(final HttpServletRequest req,
                                final HttpServletResponse resp)
        throws IOException, ServletException {
        if (callbackOption("Servlet.do_head",
                           Cadmium.createObject(this),
                           Cadmium.createObject(req),
                           Cadmium.createObject(resp))) {
            super.doHead(req, resp);
        } // end if
    } // end method 'doHead(HttpServletRequest, HttpServletResponse)'

    /**
     * {@inheritDoc}
     */
    @Override
    protected final void doOptions(final HttpServletRequest req,
                                   final HttpServletResponse resp)
        throws IOException, ServletException {
        if (callbackOption("Servlet.do_options",
                           Cadmium.createObject(this),
                           Cadmium.createObject(req),
                           Cadmium.createObject(resp))) {
            super.doOptions(req, resp);
        } // end if
    } // end method 'doOptions(HttpServletRequest, HttpServletResponse)'

    /**
     * {@inheritDoc}
     */
    @Override
    protected final void doPost(final HttpServletRequest req,
                                final HttpServletResponse resp)
        throws IOException, ServletException {
        if (callbackOption("Servlet.do_post",
                           Cadmium.createObject(this),
                           Cadmium.createObject(req),
                           Cadmium.createObject(resp))) {
            super.doPost(req, resp);
        } // end if
    } // end method 'doPost(HttpServletRequest, HttpServletResponse)'

    /**
     * {@inheritDoc}
     */
    @Override
    protected final void doPut(final HttpServletRequest req,
                               final HttpServletResponse resp)
        throws IOException, ServletException {
        if (callbackOption("Servlet.do_put",
                           Cadmium.createObject(this),
                           Cadmium.createObject(req),
                           Cadmium.createObject(resp))) {
            super.doPut(req, resp);
        } // end if
    } // end method 'doPut(HttpServletRequest, HttpServletResponse)'

    /**
     * {@inheritDoc}
     */
    @Override
    protected final void doTrace(final HttpServletRequest req,
                                 final HttpServletResponse resp)
        throws IOException, ServletException {
        if (callbackOption("Servlet.do_trace",
                           Cadmium.createObject(this),
                           Cadmium.createObject(req),
                           Cadmium.createObject(resp))) {
            super.doTrace(req, resp);
        } // end if
    } // end method 'doTrace(HttpServletRequest, HttpServletResponse)'

    /**
     * {@inheritDoc}
     */
    @Override
    protected final long getLastModified(final HttpServletRequest req) {
        if (getCallback("Servlet.get_last_modified") != null) {
            try {
                final Value res =
                    callback("Servlet.get_last_modified",
                             Cadmium.createObject(this),
                             Cadmium.createObject(req));
                return res.asBlock().asInt64();
            } catch (final Exception e) {
                return -1L;
            } // end try/catch
        } else {
            return super.getLastModified(req);
        } // end if/else
    } // end method 'getLastModified(HttpServletRequest)'

    /**
     * Runs a callback, with passed parameters.
     * @param cb callback to run - should not be <tt>null</tt>
     * @param args callback parameters
     * @return callback result
     * @throws ServletException if any error is raised during callback
     *                          execution
     */
    protected abstract Value callback(final String cb, final Value... args)
        throws ServletException;

    /**
     * Runs a callback, with passed parameters. <br/>
     * <b>Exceptions are silently ignored and return value is discarded.</b>
     * @param cb callback to run - should not be <tt>null</tt>
     * @param args callback parameters
     */
    protected abstract void callbackNoException(final String cb,
                                                final Value... args);

    /**
     * Runs a (possibly missing) callback, with passed parameters.
     * @param cb callback to run - should not be <tt>null</tt>
     * @param args callback parameters
     * @return <tt>false</tt> if the callback was found and executed,
     *         <tt>true</tt> if the callback do not exist
     * @throws ServletException if any error is raised during callback
     *                          execution
     */
    protected abstract boolean callbackOption(final String cb,
                                              final Value... args)
        throws ServletException;

    /**
     * Returns the value of a given callback.
     * @param cb callback to get calue for - should not be <tt>null</tt>
     * @return callback registered with passed name
     */
    protected abstract Value getCallback(final String cb);

    /**
     * Returns the context.
     * @return the context
     */
    protected abstract Context getContext();

} // end class 'Servlet'
