(*
 * This file is part of Cadmium.
 * Copyright (C) 2007-2010 Xavier Clerc.
 *
 * Cadmium is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * Cadmium is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *)

(** Access to Java elements. *)


(** {6 Types} *)

type java_object
(** The abstract type of Java object instances.
   This custom type supports both hashing and serialization but not comparison
   If an attempt is made to serialize an object that do not implement
   {i java.io.Serializable}, a [Sys_error] is raised. *)

type java_class = java_object
(** The abstract type of Java class. *)

type java_constructor = java_object
(** The abstract type of Java constructors. *)

type java_field = java_object
(** The abstract type of Java fields. *)

type java_method = java_object
(** The abstract type of Java methods. *)

type 'c java_type =
  | Boolean                   (** boolean Java type *)
  | Byte                      (** byte Java type (8-bit signed integer) *)
  | Char                      (** character Java type (16-bit unsigned integer) *)
  | Double                    (** double Java type (double-precision floating-point number) *)
  | Float                     (** float Java type (single-precision floating-point number) *)
  | Int                       (** int Java type (32-bit signed integer) *)
  | Long                      (** long Java type (64-bit signed integer) *)
  | Short                     (** short Java type (16-bit signed integer) *)
  | String                    (** string of Java characters (immutable) *)
  | Class of 'c               (** any (other) Java type *)
  | Void                      (** void/no return type *)
  | Array of 'c java_type     (** array, parameter is elements type *)
(** The type of Java types (primitives, arrays and bare objects). *)

type 'o java_value =
  | Boolean_val of bool       (** boolean Java value *)
  | Byte_val of int           (** byte Java value (8-bit signed integer) *)
  | Char_val of int           (** character Java value (16-bit unsigned integer) *)
  | Double_val of float       (** double Java value (double-precision floating-point number) *)
  | Float_val of float        (** float Java value (single-precision floating-point number) *)
  | Int_val of int32          (** int Java value (32-bit signed integer) *)
  | Long_val of int64         (** long Java value (64-bit signed integer) *)
  | Short_val of int          (** short Java value (16-bit signed integer) *)
  | String_val of string      (** string of Java characters (immutable) *)
  | Object_val of 'o          (** any (other) Java value *)
(** The type of Java values (primitives and bare objects). *)

type java_modifier =
  | Abstract                  (** type / method modifier *)
  | Final                     (** type / method / field modifier *)
  | Interface                 (** type modifier *)
  | Native                    (** method modifier *)
  | Private                   (** visibility modifier *)
  | Protected                 (** visibility modifier *)
  | Public                    (** visibility modifier *)
  | Static                    (** for class-wide definitions (as opposed to instance-wide definitions) *)
  | StrictFP                  (** for strict floating-point operations *)
  | Synchronized              (** synchronization modifier *)
  | Transcient                (** field (serialization) modifier *)
  | Volatile                  (** field modifier *)
(** The type of Java modifiers (for classes, fields, methods, {i etc.}). *)


(** {6 Base functions} *)

val is_available : unit -> bool
(** Tests whether Cadmium primitives are available. *)

val is_null : java_object -> bool
(** Tests whether a Java instance is null. *)

val get_null : unit -> java_object
(** Returns a null instance. *)

val instance_of : java_object -> java_class -> bool
(** Tests whether a Java instance is of a given type. *)

val cast : java_class -> java_object -> java_object
(** Casts a Java instance to a given class. *)

val java_string_of_string : string -> java_object
(** Creates a Java string from a string. *)

val string_of_java_string : java_object -> string
(** Creates a string from a Java string. *)

val java_char_of_char : char -> int
(** Converts a [char] into a Java char.*)

val char_of_java_char : int -> char
(** Converts a Java char into a [char].*)

val equal : java_object -> java_object -> bool
(** Applies Java operator {i ==} to two instances. *)

val not_equal : java_object -> java_object -> bool
(** Applies Java operator {i !=} to two instances. *)


(** {6 java.lang and java.lang.reflect} *)

(** Each of the [Object], [Class], [Constructor], [Field], [Method], and [Array]
   modules are "translations" of the {i java.lang.Object}, {i java.lang.Class},
   {i java.lang.reflect.Constructor}, {i java.lang.reflect.Field},
   {i java.lang.reflect.Method}, and {i java.lang.reflect.Array} Java classes.

   Translation is done as follows:
- instance methods with {i n} parameters are translated to functions with
  {i n + 1} parameters, the first one being the object the method is called on;
- static methods keep the same number of parameters across translation;
- constructors are translated to [make_xxx] functions;
- {i java.lang.Class}, {i java.lang.Constructor}, {i java.lang.Field},
   {i java.lang.Method} are translated to [java_class], [java_constructor],
   [java_field], [java_method];
- other object instances are translated to [java_object];
- mapping of parameter and return types is done according to [java_value] type;
- void return type as well as absence of parameter are translated to [unit];
- the methods can throw the same exceptions as their Java conterparts,
  always encapsulated in a [Java_exception] exception;
- each module provides two functions [to_obj] and [as_xxx] that {i cast} to and
   from [java_object].

   Method names are ocaml-ified, by replacing case changes by underlines
   ({i e.g.} hashCode is translated to hash_code). Obviously, some names are
   postfixed to handle name clashes. *)

module Object :
  sig
    val make : unit -> java_object
    val clone : java_object -> java_object
    val equals : java_object -> java_object -> bool
    val get_class : java_object -> java_class
    val hash_code : java_object -> int32
    val notify : java_object -> unit
    val notify_all : java_object -> unit
    val to_string : java_object -> string
    val wait : java_object -> unit
    val wait_time : java_object -> int64 -> unit
    val wait_time' : java_object -> int64 -> int32 -> unit
  end

module Class :
  sig
    val cast : java_class -> java_object -> java_object
    val for_name : string -> java_class
    val get_canonical_name : java_class -> string
    val get_constructor : java_class -> java_class java_type list -> java_constructor
    val get_constructors : java_class -> java_constructor list
    val get_declared_constructor : java_class -> java_class java_type list -> java_constructor
    val get_declared_constructors : java_class -> java_constructor list
    val get_field : java_class -> string -> java_field
    val get_fields : java_class -> java_field list
    val get_declared_field : java_class -> string -> java_field
    val get_declared_fields : java_class -> java_field list
    val get_method : java_class -> string -> java_class java_type list -> java_method
    val get_methods : java_class -> java_method list
    val get_declared_method : java_class -> string -> java_class java_type list -> java_method
    val get_declared_methods : java_class -> java_method list
    val get_interfaces : java_class -> java_class list
    val get_modifiers : java_class -> java_modifier list
    val get_name : java_class -> string
    val get_package : java_class -> string
    val get_simple_name : java_class -> string
    val get_superclass : java_class -> java_class option
    val is_array : java_class -> bool
    val is_interface : java_class -> bool
    val is_primitive : java_class -> bool
    val new_instance : java_class -> java_object
    val to_object : java_class -> java_object
    val as_class : java_object -> java_class
  end

module Constructor :
  sig
    val get_declaring_class : java_constructor -> java_class
    val get_exception_types : java_constructor -> java_class list
    val get_modifiers : java_constructor -> java_modifier list
    val get_parameter_types : java_constructor -> java_class java_type list
    val new_instance : java_constructor -> java_object java_value list -> java_object
    val to_object : java_constructor -> java_object
    val as_constructor : java_object -> java_constructor
  end

module Field :
  sig
    val get_declaring_class : java_field -> java_class
    val get_modifiers : java_field -> java_modifier list
    val get_name : java_field -> string
    val get : java_field -> java_object -> java_object
    val get_boolean : java_field -> java_object -> bool
    val get_byte : java_field -> java_object -> int
    val get_char : java_field -> java_object -> int
    val get_double : java_field -> java_object -> float
    val get_float : java_field -> java_object -> float
    val get_int : java_field -> java_object -> int32
    val get_long : java_field -> java_object -> int64
    val get_short : java_field -> java_object -> int
    val set : java_field -> java_object -> java_object -> unit
    val set_boolean : java_field -> java_object -> bool -> unit
    val set_byte : java_field -> java_object -> int -> unit
    val set_char : java_field -> java_object -> int -> unit
    val set_double : java_field -> java_object -> float -> unit
    val set_float : java_field -> java_object -> float -> unit
    val set_int : java_field -> java_object -> int32 -> unit
    val set_long : java_field -> java_object -> int64 -> unit
    val set_short : java_field -> java_object -> int -> unit
    val to_object : java_field -> java_object
    val as_field : java_object -> java_field
  end

module Method :
  sig
    val get_declaring_class : java_method -> java_class
    val get_exception_types : java_method -> java_class list
    val get_modifiers : java_method -> java_modifier list
    val get_name : java_method -> string
    val get_parameter_types : java_method -> java_class java_type list
    val get_return_type : java_method -> java_class java_type
    val invoke : java_method -> java_object -> java_object java_value list -> java_object java_value
    val to_object : java_method -> java_object
    val as_method : java_object -> java_method
  end

module Array :
  sig
    val new_instance : java_class java_type -> int -> java_object
    val new_instance_dims : java_class java_type -> int list -> java_object
    val get_length : java_object -> int
    val get : java_object -> int -> java_object
    val get_boolean : java_object -> int -> bool
    val get_byte : java_object -> int -> int
    val get_char : java_object -> int -> int
    val get_double : java_object -> int -> float
    val get_float : java_object -> int -> float
    val get_int : java_object -> int -> int32
    val get_long : java_object -> int -> int64
    val get_short : java_object -> int -> int
    val set : java_object -> int -> java_object -> unit
    val set_boolean : java_object -> int -> bool -> unit
    val set_byte : java_object -> int -> int -> unit
    val set_char : java_object -> int -> int -> unit
    val set_double : java_object -> int -> float -> unit
    val set_float : java_object -> int -> float -> unit
    val set_int : java_object -> int -> int32 -> unit
    val set_long : java_object -> int -> int64 -> unit
    val set_short : java_object -> int -> int -> unit
  end


(** {6 Conversions between OCaml and Java arrays} *)

(** This module provides function converting OCaml arrays both from and to
   their Java counterparts. When converting from Java to OCaml, it is the
   responsibility of the programmer to provide a Java instance of correct
   type. *)

module ArrayUtils :
  sig
    val bool_array_to_java : bool array -> java_object
    val byte_array_to_java : int array -> java_object
    val char_array_to_java : char array -> java_object
    val double_array_to_java : float array -> java_object
    val float_array_to_java : float array -> java_object
    val int_array_to_java : int32 array -> java_object
    val long_array_to_java : int64 array -> java_object
    val short_array_to_java : int array -> java_object
    val string_array_to_java : string array -> java_object
    val bool_array_from_java : java_object -> bool array
    val byte_array_from_java : java_object -> int array
    val char_array_from_java : java_object -> char array
    val double_array_from_java : java_object -> float array
    val float_array_from_java : java_object -> float array
    val int_array_from_java : java_object -> int32 array
    val long_array_from_java : java_object -> int64 array
    val short_array_from_java : java_object -> int array
    val string_array_from_java : java_object -> string array
  end


(** {6 Exceptions} *)

exception Java_exception of (string * string * java_object)
(** Encapsulates Java exceptions, arguments are:
   exception class, exception message and exception instance *)

val print_stack_trace : out_channel -> java_object -> unit
(** Prints the stack trace of a Java exception onto an output channel. *)

val throw_exception : java_object -> unit
(** [throw_exception e] throws [e] if it is an exception, raises
   [Invalid_argument] otherwise *)


(** {6 Access to Graphics elements} *)

val get_frame : unit -> java_object
(** Returns the frame used by the 'Graphics' module (can be {i null} if no
   frame is used by the module). Raises [Invalid_argument] if the 'Graphics'
   module has not been initialized. *)

val get_canvas : unit -> java_object
(** Returns the canvas used by the 'Graphics' module.
   Raises [Invalid_argument] if the 'Graphics' module has not been initialized. *)

val get_buffer : unit -> java_object
(** Returns the buffer used by the 'Graphics' module.
   Raises [Invalid_argument] if the 'Graphics' module has not been initialized. *)


(** {6 Callbacks / Proxies} *)

val create_interface : (string * (java_object java_type list) * java_object java_type) list -> java_class
(** [create_interface meths] returns the Java class corresponding to the newly
   compiled interface that contains the methods from [l]. [l] contains method
   definitions given by a name, a list of parameter types and a return type.
   The fully qualified name of the interface is auto-generated. *)

val create_named_interface : (string * (java_object java_type list) * java_object java_type) list -> string -> java_class
(** [create_named_interface meths name] returns the Java class corresponding
   to the newly compiled interface whose fully qualified name is [name] that
   contains the methods from [l]. [l] contains method definitions given by a
   name, a list of parameter types and a return type. *)

val register : java_class list -> (java_method * (java_object java_value list -> java_object java_value)) list -> java_object
(** [register itf l] returns a Java instance that implements [itf] interfaces
   and calls ocaml functions according to [l] when corresponding methods are
   called over returned instance: each [(m, f)] element of [l] means that each
   call to method [m] will result in a call to function [f], the return value
   of the call to [m] being the result of [f]. If a call on returned instance
   does not correspond to any function (according to [l]), {i null} will be
   returned. *)


(** {6 Synchronization} *)

val synchronized : java_object -> (unit -> unit) -> unit
(** [synchronized o f] executes [f ()] in a synchronized block over [o]. *)


(** {6 Applet} *)

type applet_parameter = string * string * string
(** Represents an applet parameter defined by its name, type, and description. *)

type applet_handler = {
    info : string; (** Data to be returned by {i java.applet.Applet.getAppletInfo()}. *)
    parameters : applet_parameter array; (** Data to be returned by {i java.applet.Applet.getParameterInfo()}. *)
    init : java_object -> unit; (** Callback for {i java.applet.Applet.init()}, parameter being applet instance. *)
    start : java_object -> unit; (** Callback for {i java.applet.Applet.start()}, parameter being applet instance. *)
    run : java_object -> unit; (** Called by the applet thread launched right after the end of {i start}. *)
    stop : java_object -> unit; (** Callback for {i java.applet.Applet.stop()}, parameter being applet instance. *)
    destroy : java_object -> unit; (** Callback for {i java.applet.Applet.destroy()}, parameter being applet instance. *)
  }
(** Applet handler defining applet behaviour. *)

val register_applet : applet_handler -> unit
(** Registers the passed handler defining applet behaviour,
   replacing any previously registered handler. *)

val get_applet_parameter : java_object -> string -> string
(** [get_applet_parameter applet name] returns the value of the parameter named
   [name] for the applet [applet]. Raises [Not_found] if the parameter is not
   set. *)


(** {6 Runtime} *)

(** The following functions allow access to elements of the Cadmium runtime.
   A {i code runner} is a Cadmium thread running OCaml code either natively
   or through interpretation. A {i context} represents the state of an OCaml
   program and is shared between the various threads of this program. Finally,
   an {i interpreter} is an object that runs OCaml bytecode. By using such an
   interpreter it is possible for a native Cadmium program to run OCaml
   bytecode.

   For more information about {i code runner}, {i context}, and {i interpreter}
   as well as related objects, one is advised to refer to the Cadmium
   javadoc. *)

type interpreter_parameters = {
    file : string; (** argv.(0) value, NOT bytecode file *)
    arguments : string array; (** other arguments to program *)
    init_stack_size : int; (** initial stack size (in elements) *)
    max_stack_size : int; (** maximum stack size (in elements) *)
    backtrace : bool; (** whether backtrace is activated *)
    exit_stops_jvm : bool; (** whether exit stops the whole JVM *)
    providers : string array; (** additional primitive providers as fully-qualified Java class names *)
    compile_dispatchers : bool; (** whether dispatchers should be compiled *)
    use_awt : bool; (** whether the [Graphics] frame should be AWT *)
    bare_canvas : bool; (** whether the [Graphics] component should be a bare canvas *)
    use_javax_sound : bool; (** whether javax.sound should be used for beeps *)
    use_jdbm : bool; (** whether jdbm should be used for dbm *)
    os : string; (** OS identifier (suggested values: {i Unix}, {i Cygwin}, {i Win32}, {i MacOS}, {i Cadmium}) *)
    unix_emulation : bool (** whether unix primitives should be emulated *)
  }
(** Interpreter parameters. *)

type interpreter
(** Type for interpreter instances ({i fr.x9c.cadmium.kernel.Interpreter}). *)

val is_native : unit -> bool
(** Tests whether the current code runner is native. *)

val get_coderunner : unit -> java_object
(** Returns the current code runner ({i fr.x9c.cadmium.kernel.CodeRunner}
   instance). *)

val get_context : unit -> java_object
(** Returns the current context ({i fr.x9c.cadmium.kernel.Context}
   instance). *)

val get_parameters : unit -> interpreter_parameters
(** Returns the current parameters. *)

val make_interpreter : interpreter_parameters -> string -> string -> in_channel -> out_channel -> out_channel -> interpreter
(** [make_interpreter params dir file stdin stdout stderr] returns an
   interpreter executing bytecode from [file], using [dir] as initial working
   directory, [stdin] as standard input, [stdout] as standard output, [stderr]
   as standard error, and [params] as parameters. Raises [Java_exception] if
   interpreter creation fails. *)

val interpreter_execute : interpreter -> 'a
(** Executes an interpreter, returning the result of execution. The returned
   value should be an [int] indicating the exit code of the program.
   Raises [Java_exception] if execution fails. *)

val interpreter_execute_callback : interpreter -> string -> 'a list -> 'b
(** [interpreter_execute_callback i c p] executes callback [c] in interpreter
   [i] with parameters [p]. As there is no need that all parameters to a
   callback have the same type, one is advised to use [Obj.magic] circumvent
   the problem. Raises [Java_exception] if execution fails. *)

val interpreter_get_context : interpreter -> java_object
(** Returns the context of an interpreter ({i fr.x9c.cadmium.kernel.Context}
   instance). *)

val interpreter_get_parameters : interpreter -> interpreter_parameters
(** Returns the parameters of an interpreter. *)

val java_object_of_interpreter : interpreter -> java_object
(** Returns the interpreter instance as a Java object
   ({i fr.x9c.cadmium.kernel.Interpreter} instance). *)


(** {6 Scripts} *)

val get_binding : string -> 'a
(** Returns the value associated with the passed identifier.
   If the value is a Java primitive, primitive wrapper, or a string, it is
   automatically mapped to the corresponding OCaml value. Otherwise, it is
   presented as a java_object instance.
   Raises [Not_found] if such a value does not exist. *)
