(*
 * This file is part of Cadmium.
 * Copyright (C) 2007-2010 Xavier Clerc.
 *
 * Cadmium is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * Cadmium is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *)

(** Access to databases through JDBC. *)

(** The following modules are "translations" of the Java classes having the
    same name in the package {i java.sql} package. Detailed information about
    the various functions can hence be found in the Javadoc of this package.
    One should only notice that SQL {i null} values are represented by {i option}
    types. *)

(** {6 Types} *)

type jdbc_type =
  | Array (** SQL type {i ARRAY}. *)
  | Big_int (** SQL type {i BIGINT}. *)
  | Binary (** SQL type {i BINARY}. *)
  | Bit (** SQL type {i BIT}. *)
  | Blob (** SQL type {i BLOB}. *)
  | Boolean (** SQL type {i BOOLEAN}. *)
  | Char (** SQL type {i CHAR}. *)
  | Clob (** SQL type {i CLOB}. *)
  | Datalink (** SQL type {i DATALINK}. *)
  | Date (** SQL type {i DATE}. *)
  | Decimal (** SQL type {i DECIMAL}. *)
  | Distinct (** SQL type {i DISTINCT}. *)
  | Double (** SQL type {i DOUBLE}. *)
  | Float (** SQL type {i FLOAT}. *)
  | Integer (** SQL type {i INTEGER}. *)
  | Java_object (** SQL type {i JAVA_OBJECT}. *)
  | Long_nvarchar (** SQL type {i LONGNVARCHAR}. *)
  | Long_varbinary (** SQL type {i LONGVARBINARY}. *)
  | Long_varchar (** SQL type {i LONGVARCHAR}. *)
  | Nchar (** SQL type {i NCHAR}. *)
  | Nclob (** SQL type {i NCLOB}. *)
  | Numeric (** SQL type {i NUMERIC}. *)
  | Nvarchar (** SQL type {i NVARCHAR}. *)
  | Other (** SQL type for database-specific types. *)
  | Real (** SQL type {i REAL}. *)
  | Ref (** SQL type {i REF}. *)
  | Rowid (** SQL type {i ROWID}. *)
  | Small_int (** SQL type {i SMALLINT}. *)
  | SQL_XML (** SQL type {i XML}. *)
  | Struct (** SQL type {i STRUCT}. *)
  | Time (** SQL type {i TIME}. *)
  | Timestamp (** SQL type {i TIMESTAMP}. *)
  | Tiny_int (** SQL type {i TINYINT}. *)
  | Varbinary (** SQL type {i VARBINARY}. *)
  | Varchar (** SQL type {i VARCHAR}. *)
(** This type defines the various JDBC types. *)

type client_info_status =
  | Reason_unknown (** Indicates that the modification failed for an unknown reason. *)
  | Reason_unknown_property (** Indicates that the modification failed because the property is unknown. *)
  | Reason_value_invalid (** Indicates that the modification failed because the property value is invalid. *)
  | Reason_value_truncated (** Indicates that the modification failed because the property value is too large. *)
(** This type defines the kinds of failure related to client info status. *)

type rowid_lifetime =
  | Rowid_unsupported (** Indicates that row identifiers are unsupported. *)
  | Rowid_valid_forever (** Indicates that row identifiers are valid for ever. *)
  | Rowid_valid_other (** Indicates that the lifetime of row identifiers is unknown. *)
  | Rowid_valid_session (** Indicates that row identifiers are valid for the current session. *)
  | Rowid_valid_transaction (** Indicates that row identifiers are valid for the current transaction. *)
(** This type defines the lifetime of row identifiers. *)

type transaction_isolation =
  | Transaction_none (** Indicates that transactions are not supported. *)
  | Transaction_read_committed (** Indicates that reads of dirty values are not allowed. *)
  | Transaction_read_uncommitted (** Indicates that reads of dirty values are allowed. *)
  | Transaction_repeatable_read (** Indicates that both reads of dirty values and repeatable reads are not allowed. *)
  | Transaction_serializable (** Indicates that reads of dirty values, repeatable reads, and phantom reads are not allowed. *)
(** This type defines the transaction isolation levels of result sets. *)

type result_set_concurrency =
  | Concurrency_read_only (** Indicates that a result set is read-only. *)
  | Concurrency_updatable (** Indicates that a result set is updatable. *)
(** This type defines the concurrency modes of result sets. *)

type result_set_fetch_direction =
  | Fetch_forward (** Indicates that the result set is iterated forward. *)
  | Fetch_reverse (** Indicates that the result set is iterated backward. *)
  | Fetch_unknown (** Indicates that the direction of the result set iterator is unknown. *)
(** This type defines the fetch directions of result sets. *)

type result_set_holdability =
  | Close_cursors_at_commit (** Indicates that cursors are closed at commit. *)
  | Hold_cursors_over_commit (** Indicates that cursors remain untouched at commit. *)
(** This type defines the holdability of result sets across commits. *)

type result_set_type =
  | Type_forward_only (** Indicates that the cursor of a result set can only move forward. *)
  | Type_scroll_insensitive (** Indicates that the cursor of a result set is scrollable and insensitive to data changes. *)
  | Type_scroll_sensitive (** Indicates that the cursor of a result set is scrollable and sensitive to data changes. *)
(** This type defines the types of result sets. *)

type result_scope =
  | Scope_temporary (** Indicates that the result is very temporary. *)
  | Scope_transaction (** Indicates that the result is valid for the current transaction. *)
  | Scope_session (** Indicates that the result is valid for the current session. *)
(** This type defines the scope of a result. *)

type statement_action =
  | Close_all_results (** Indicates that result sets should be closed. *)
  | Close_current_result (** Indicates that the current result set should be closed. *)
  | Keep_current_result (** Indicates that result sets should remain untouched. *)
(** This type defines the action to perform when moving to the next result of a batch execution. *)

type statement_autogen_keys =
  | No_generated_keys (** Indicates that auto-generated values should not be retrievable. *)
  | Return_generated_keys (** Indicates that auto-generated values should be retrievable. *)
(** This type defines whether auto-generated values should be retrievable. *)

type statement_result =
  | Execute_failed (** Indicates that the execution failed. *)
  | Success_no_info (** Indicates that the execution succeeded, without further information. *)
  | Success of int32 (** Indicates that the execution succeeded, the parameter being the number of rows affected by the execution. *)
(** This type defines the result of a statement. *)

type parameter_mode =
  | Mode_in (** Indicates that the parameter is an input parameter. *)
  | Mode_out (** Indicates that the parameter is an ouput parameter. *)
  | Mode_inout (** Indicates that the parameter is both an input and output parameter. *)
  | Mode_unknown (** Indicates that the mode of the parameter is unknown. *)
(** This type defines the mode of parameters. *)

type nullable =
  | No_nulls (** Indicates that the considered element cannot be null. *)
  | Nullable (** Indicates that the considered element can be null. *)
  | Nullable_unknown (** Indicates that the metadata cannot say whether the considered element can be null. *)
(** This type defines whether an element (column or parameter) may be null. *)

type sql_state_type =
  | State_xopen (** XOpen state type. *)
  | State_sql (** SQL state type. *)
  | State_sql99 (** SQL99 state type. *)
(** This type defines the kinds of state types. *)

module Savepoint :
  sig
    type t
        (** Type of save points. *)

    val get_savepoint_id : t -> int32
        (** Returns the identifier of the passed savepoint. *)

    val get_savepoint_name : t -> string
        (** Returns the name of the passed savepoint. *)
  end
(** Save points (used to mark point for rollback). *)

type sql_warning = { reason : string; state : string }
(** This type represents a SQL warnings issued by a database. *)


(** {6 Data types} *)

module Array :
  sig
    type t
        (** Type of JDBC arrays. *)

    val free : t -> unit
        (** Frees the passed array. *)

    val get_array : t -> Cadmium.java_object
        (** Converts the passed array into a bare Java object. *)

    val get_array_slice : t -> int64 -> int32 -> Cadmium.java_object
        (** [get_array_slice a ofs sz] returns a slice array as a base Java Object.
            The slice start at offset [ofs] and has a size of [sz]. *)

    val get_base_type : t -> jdbc_type
        (** Returns the type of the elements in the array. *)

    val get_base_type_name : t -> string
        (** Returns the type name of the elements in the array. *)
  end
(** Arrays (as managed by JDBC). *)

module Blob :
  sig
    type t
        (** Type of blobs. *)

    val free : t -> unit
        (** Frees the passed blob. *)

    val get_binary_stream : t -> in_channel
        (** Returns a stream allowing to read the blob data. *)

    val get_binary_stream_pos : t -> int64 -> int64 -> in_channel
        (** Returns a stream allowing to read the blob data.
            The first passed integer is the offset to start reading from.
            The second passed integer is the length of the data. *)

    val get_bytes : t -> int64 -> int32 -> string
        (** [get_bytes b ofs sz] returns a part of the blob data.
            The returned bytes are those from offset [ofs] to offset [ofs + sz - 1]. *)

    val length : t -> int64
        (** Returns the length of the passed blob. *)

    val position_blob : t -> t -> int64 -> int64
        (** [position_blob b b' ofs] returns the byte position in [b] at which
            the pattern designated by [b'] begins.
            Search starts at offset [ofs]. *)

    val position_bytes : t -> string -> int64 -> int64
        (** [position_bytes b b' ofs] returns the byte position in [b] at which
            the pattern designated by [b'] begins.
            Search starts at offset [ofs]. *)

    val set_binary_stream : t -> int64 -> out_channel
        (** Returns a stream allowing to write the blob data.
            The passed integer is the offset to start writing to. *)

    val set_bytes : t -> int64 -> string -> int32
        (** [set_bytes b ofs v] writes the array of bytes [v] to the blob [b],
            starting at offset [ofs]. The number of bytes actually written is
            returned. *)

    val set_bytes_ofs : t -> int64 -> string -> int -> int -> int32
        (** [set_bytes_ofs b ofs v ofs' sz] writes a part of the array of bytes
            [v] to the blob [b]. [sz] bytes are to be written, starting at
            offset [ofs'] in [v] and [ofs] in [b]. The number of bytes actually
            written is returned. *)

    val truncate : t -> int64 -> unit
        (** Truncates the passed blob at the passed offset. *)
  end
(** Blobs (Binary Large Object) .*)

module Clob :
  sig
    type t
        (** Type of clobs. *)

    val free : t -> unit
        (** Frees the passed clob. *)

    val get_ascii_stream : t -> in_channel
        (** Returns a stream allowing to read the clob data. *)

    val get_sub_string : t -> int64 -> int32 -> string
        (** [get_sub_string c ofs sz] returns a part of the clob data.
            The returned bytes are those from offset [ofs] to offset [ofs + sz - 1]. *)

    val length : t -> int64
        (** Returns the length of the passed clob. *)

    val position_clob : t -> t -> int64 -> int64
        (** [position_clob c c' ofs] returns the byte position in [c] at which
            the pattern designated by [c'] begins.
            Search starts at offset [ofs]. *)

    val position_string : t -> string -> int64 -> int64
        (** [position_string c c' ofs] returns the byte position in [c] at which
            the pattern designated by [c'] begins.
            Search starts at offset [ofs]. *)

    val set_ascii_stream : t -> int64 -> out_channel
        (** Returns a stream allowing to write the clob data.
            The passed integer is the offset to start writing to. *)

    val set_string : t -> int64 -> string -> int32
        (** [set_string c ofs v] writes the string [v] to the clob [c],
            starting at offset [ofs]. The number of characters actually written
            is returned. *)

    val set_string_ofs : t -> int64 -> string -> int -> int -> int32
        (** [set_string_ofs c ofs v ofs' sz] writes a part of the string
            [v] to the clob [c]. [sz] characters are to be written, starting at
            offset [ofs'] in [v] and [ofs] in [c]. The number of characters
            actually written is returned. *)

    val truncate : t -> int64 -> unit
        (** Truncates the passed clob at the passed offset. *)
  end
(** Clobs (Character Large Object) .*)

module Date :
  sig
    type t
        (** Type of dates. *)

    val make : unit -> t
        (** Returns the current date. *)

    val make_of_int64 : int64 -> t
        (** Returns a date using the passed integer as the number of milliseconds since the epoch. *)

    val make_of_string : string -> t
        (** Returns a date built from the passed string. *)

    val after : t -> t -> bool
        (** [after d1 d2] tests whether [d1] is strictly later than [d2]. *)

    val before : t -> t -> bool
        (** [before d1 d2] tests whether [d1] is strictly earlier than [d2]. *)

    val compare_to : t -> t -> int
        (** [compare_to d1 d2] compares [d1] and [d2] returning:
            - -1 if [d1] is earlier than [d2];
            - 0 if [d1] is equal to [d2];
            - 1 if [d1] is later than [d2]. *)

    val equals : t -> t -> bool
        (** Tests whether the passed dates are equal. *)

    val get_day_of_month : t -> int
        (** Returns the day of the month for the passed date. *)

    val get_day_of_week : t -> int
        (** Returns the day of the week for the passed date. *)

    val get_day_of_week_in_month : t -> int
        (** Returns the day of the week in month for the passed date. *)

    val get_day_of_year : t -> int
        (** Returns the day of the year for the passed date. *)

    val get_hour_of_day : t -> int
        (** Returns the hour of the day for the passed date. *)

    val get_millisecond : t -> int
        (** Returns the millisecond of the second for the passed date. *)

    val get_minute : t -> int
        (** Returns the minute of the hour for the passed date. *)

    val get_month : t -> int
        (** Returns the month of the year for the passed date. *)

    val get_second : t -> int
        (** Returns the second of the minute for the passed date. *)

    val get_time : t -> int64
        (** Returns the number of milliseconds since the epoch. *)

    val get_week_of_month : t -> int
        (** Returns the week of the month for the passed date. *)

    val get_week_of_year : t -> int
        (** Returns the week of the year for the passed date. *)

    val get_year : t -> int
        (** Returns the year for the passed date. *)

    val set_day_of_month : t -> int -> unit
        (** Sets the day of the month for the passed date to the given value. *)

    val set_day_of_week : t -> int -> unit
        (** Sets the day of the week for the passed date to the given value. *)

    val set_day_of_week_in_month : t -> int -> unit
        (** Sets the day of the week in month for the passed date to the given value. *)

    val set_day_of_year : t -> int -> unit
        (** Sets the day of the year for the passed date to the given value. *)

    val set_hour_of_day : t -> int -> unit
        (** Sets the hour of the day for the passed date to the given value. *)

    val set_millisecond : t -> int -> unit
        (** Sets the millisecond of the second for the passed date to the given value. *)

    val set_minute : t -> int -> unit
        (** Sets the minute of the hour for the passed date to the given value. *)

    val set_month : t -> int -> unit
        (** Sets the month of the year for the passed date to the given value. *)

    val set_second : t -> int -> unit
        (** Sets the second of the minute for the passed date to the given value. *)

    val set_time : t -> int64 -> unit
        (** Sets the number of milliseconds since the epoch for the passed date to the given value. *)

    val set_week_of_month : t -> int -> unit
        (** Sets the week of the month for the passed date to the given value. *)

    val set_week_of_year : t -> int -> unit
        (** Sets the week of the year for the passed date to the given value. *)

    val set_year : t -> int -> unit
        (** Sets the year for the passed date to the given value. *)

    val to_string : t -> string
        (** Converts the passed date into a string. *)
  end
(** Dates. *)

module NClob :
  sig
    type t
        (** Type of nclobs. *)

    val free : t -> unit
        (** Frees the passed nclob. *)

    val get_ascii_stream : t -> in_channel
        (** Returns a stream allowing to read the nclob data. *)

    val get_sub_string : t -> int64 -> int32 -> string
        (** [get_sub_string n ofs sz] returns a part of the nclob data.
            The returned bytes are those from offset [ofs] to offset [ofs + sz - 1]. *)

    val length : t -> int64
        (** Returns the length of the passed nclob. *)

    val position_clob : t -> t -> int64 -> int64
        (** [position_clob n n' ofs] returns the byte position in [n] at which
            the pattern designated by [n'] begins.
            Search starts at offset [ofs]. *)

    val position_string : t -> string -> int64 -> int64
        (** [position_string n n' ofs] returns the byte position in [n] at which
            the pattern designated by [n'] begins.
            Search starts at offset [ofs]. *)

    val set_ascii_stream : t -> int64 -> out_channel
        (** Returns a stream allowing to write the nclob data.
            The passed integer is the offset to start writing to. *)

    val set_string : t -> int64 -> string -> int32
        (** [set_string n ofs v] writes the string [v] to the nclob [n],
            starting at offset [ofs]. The number of characters actually written
            is returned. *)

    val set_string_ofs : t -> int64 -> string -> int -> int -> int32
        (** [set_string_ofs c ofs v ofs' sz] writes a part of the string
            [v] to the nclob [c]. [sz] characters are to be written, starting at
            offset [ofs'] in [v] and [ofs] in [n]. The number of characters
            actually written is returned. *)

    val truncate : t -> int64 -> unit
        (** Truncates the passed nclob at the passed offset. *)
  end
(** NClobs (Character Large Object using National Character Set) .*)

module Ref :
  sig
    type t
        (** Type of references to structured SQL types. *)

    val get_base_type_name : t -> string
        (** Returns the SQL name of the structured type referenced by the
            passed reference. *)

    val get_object : t -> Cadmium.java_object
        (** Returns the structured type instance referenced by the passed
            reference. *)

    val set_object : t -> Cadmium.java_object -> unit
        (** Sets the structured type value that the passed reference points to. *)
  end
(** References to structured SQL types. *)

module RowId :
  sig
    type t
        (** Type of row identifiers. *)

    val equals : t -> t -> bool
        (** Tests whether the passed row identifiers are equal. *)

    val get_bytes : t -> string
        (** Returns the actual data for the passed row identifier (as a byte array). *)

    val hash_code : t -> int32
        (** Returns the hash code of the passed row identifier. *)

    val to_string : t -> string
        (** Converts the passed row identifier into a string. *)
  end
(** Row identifiers. *)

module SQLXML :
  sig
    type t
        (** Type of XML values. *)

    val free : t -> unit
        (** Frees the passed XML value. *)

    val get_binary_stream : t -> in_channel
        (** Returns a stream allowing to read the XML data. *)

    val get_string : t -> string
        (** Returns the XML value as a string. *)

    val set_binary_stream : t -> out_channel
        (** Returns a stream allowing to write the XML data. *)

    val set_string : t -> string -> unit
        (** Changes the XML data to the passed string value .*)
  end
(** XML value as stored in a SQL database. *)

module Struct :
  sig
    type t
        (** Type of SQL structured types. *)

    val get_attributes : t -> Cadmium.java_object list
        (** Returns the ordered values of the attributes of the passed type. *)

    val get_sql_type_name : t -> string
        (** Returns the name of the passed type. *)
  end
(** Structured SQL types. *)

module Time :
  sig
    type t
        (** Type of times. *)

    val make : unit -> t
        (** Returns the current time. *)

    val make_of_int64 : int64 -> t
        (** Returns a time using the passed integer as the number of milliseconds since the epoch. *)

    val make_of_string : string -> t
        (** Returns a time built from the passed string. *)

    val after : t -> t -> bool
        (** [after t1 t2] tests whether [t1] is strictly later than [t2]. *)

    val before : t -> t -> bool
        (** [before t1 t2] tests whether [t1] is strictly earlier than [t2]. *)

    val compare_to : t -> t -> int
        (** [compare_to t1 t2] compares [t1] and [t2] returning:
            - -1 if [t1] is earlier than [t2];
            - 0 if [t1] is equal to [t2];
            - 1 if [t1] is later than [t2]. *)

    val equals : t -> t -> bool
        (** Tests whether the passed times are equal. *)

    val get_day_of_month : t -> int
        (** Returns the day of the month for the passed time. *)

    val get_day_of_week : t -> int
        (** Returns the day of the week for the passed time. *)

    val get_day_of_week_in_month : t -> int
        (** Returns the day of the week in month for the passed time. *)

    val get_day_of_year : t -> int
        (** Returns the day of the year for the passed time. *)

    val get_hour_of_day : t -> int
        (** Returns the hour of the day for the passed time. *)

    val get_millisecond : t -> int
        (** Returns the millisecond of the second for the passed time. *)

    val get_minute : t -> int
        (** Returns the minute of the hour for the passed time. *)

    val get_month : t -> int
        (** Returns the month of the year for the passed time. *)

    val get_second : t -> int
        (** Returns the second of the minute for the passed time. *)

    val get_time : t -> int64
        (** Returns the number of milliseconds since the epoch. *)

    val get_week_of_month : t -> int
        (** Returns the week of the month for the passed time. *)

    val get_week_of_year : t -> int
        (** Returns the week of the year for the passed time. *)

    val get_year : t -> int
        (** Returns the year for the passed time. *)

    val set_day_of_month : t -> int -> unit
        (** Sets the day of the month for the passed time to the given value. *)

    val set_day_of_week : t -> int -> unit
        (** Sets the day of the week for the passed time to the given value. *)

    val set_day_of_week_in_month : t -> int -> unit
        (** Sets the day of the week in month for the passed time to the given value. *)

    val set_day_of_year : t -> int -> unit
        (** Sets the day of the year for the passed time to the given value. *)

    val set_hour_of_day : t -> int -> unit
        (** Sets the hour of the day for the passed time to the given value. *)

    val set_millisecond : t -> int -> unit
        (** Sets the millisecond of the second for the passed time to the given value. *)

    val set_minute : t -> int -> unit
        (** Sets the minute of the hour for the passed time to the given value. *)

    val set_month : t -> int -> unit
        (** Sets the month of the year for the passed time to the given value. *)

    val set_second : t -> int -> unit
        (** Sets the second of the minute for the passed time to the given value. *)

    val set_time : t -> int64 -> unit
        (** Sets the number of milliseconds since the epoch for the passed time to the given value. *)

    val set_week_of_month : t -> int -> unit
        (** Sets the week of the month for the passed time to the given value. *)

    val set_week_of_year : t -> int -> unit
        (** Sets the week of the year for the passed time to the given value. *)

    val set_year : t -> int -> unit
        (** Sets the year for the passed time to the given value. *)

    val to_string : t -> string
        (** Converts the passed time into a string. *)
  end
(** Times. *)

module Timestamp :
  sig
    type t
        (** Type of timestamps. *)

    val make : unit -> t
        (** Returns the current timestamp. *)

    val make_of_int64 : int64 -> t
        (** Returns a timestamp using the passed integer as the number of milliseconds since the epoch. *)

    val make_of_string : string -> t
        (** Returns a timestamp built from the passed string. *)

    val after : t -> t -> bool
        (** [after t1 t2] tests whether [t1] is strictly later than [t2]. *)

    val before : t -> t -> bool
        (** [before t1 t2] tests whether [t1] is strictly earlier than [t2]. *)

    val compare_to : t -> t -> int
        (** [compare_to t1 t2] compares [t1] and [t2] returning:
            - -1 if [t1] is earlier than [t2];
            - 0 if [t1] is equal to [t2];
            - 1 if [t1] is later than [t2]. *)

    val equals : t -> t -> bool
        (** Tests whether the passed timestamps are equal. *)

    val get_day_of_month : t -> int
        (** Returns the day of the month for the passed timestamp. *)

    val get_day_of_week : t -> int
        (** Returns the day of the week for the passed timestamp. *)

    val get_day_of_week_in_month : t -> int
        (** Returns the day of the week in month for the passed timestamp. *)

    val get_day_of_year : t -> int
        (** Returns the day of the year for the passed timestamp. *)

    val get_hour_of_day : t -> int
        (** Returns the hour of the day for the passed timestamp. *)

    val get_millisecond : t -> int
        (** Returns the millisecond of the second for the passed timestamp. *)

    val get_minute : t -> int
        (** Returns the minute of the hour for the passed timestamp. *)

    val get_month : t -> int
        (** Returns the month of the year for the passed timestamp. *)

    val get_nanos : t -> int32
        (** Returns the nanosecond of the millisecond for the passed timestamp. *)

    val get_second : t -> int
        (** Returns the second of the minute for the passed timestamp. *)

    val get_time : t -> int64
        (** Returns the number of milliseconds since the epoch. *)

    val get_week_of_month : t -> int
        (** Returns the week of the month for the passed timestamp. *)

    val get_week_of_year : t -> int
        (** Returns the week of the year for the passed timestamp. *)

    val get_year : t -> int
        (** Returns the year for the passed timestamp. *)

    val set_day_of_month : t -> int -> unit
        (** Sets the day of the month for the passed timestamp to the given value. *)

    val set_day_of_week : t -> int -> unit
        (** Sets the day of the week for the passed timestamp to the given value. *)

    val set_day_of_week_in_month : t -> int -> unit
        (** Sets the day of the week in month for the passed timestamp to the given value. *)

    val set_day_of_year : t -> int -> unit
        (** Sets the day of the year for the passed timestamp to the given value. *)

    val set_hour_of_day : t -> int -> unit
        (** Sets the hour of the day for the passed timestamp to the given value. *)

    val set_millisecond : t -> int -> unit
        (** Sets the millisecond of the second for the passed timestamp to the given value. *)

    val set_minute : t -> int -> unit
        (** Sets the minute of the hour for the passed timestamp to the given value. *)

    val set_month : t -> int -> unit
        (** Sets the month of the year for the passed timestamp to the given value. *)

    val set_nanos : t -> int32 -> unit
        (** Sets the nanosecond of the millisecond for the passed timestamp to the given value. *)

    val set_second : t -> int -> unit
        (** Sets the second of the minute for the passed timestamp to the given value. *)

    val set_time : t -> int64 -> unit
        (** Sets the number of milliseconds since the epoch for the passed timestamp to the given value. *)

    val set_week_of_month : t -> int -> unit
        (** Sets the week of the month for the passed timestamp to the given value. *)

    val set_week_of_year : t -> int -> unit
        (** Sets the week of the year for the passed timestamp to the given value. *)

    val set_year : t -> int -> unit
        (** Sets the year for the passed timestamp to the given value. *)

    val to_string : t -> string
        (** Converts the passed timestamp into a string. *)
  end
(** Timestamps. *)


(** {6 Result sets} *)

module ResultSetMetaData :
  sig
    type t
        (** Type of metadata for result sets. *)

    val get_catalog_name : t -> int -> string
        (** Returns the catalog name for the column whose index is passed. *)

    val get_column_class_name : t -> int -> string
        (** Returns the class name for the column whose index is passed. *)

    val get_column_count : t -> int
        (** Returns the number of columns. *)

    val get_column_display_size : t -> int -> int
        (** Returns the display size for the column whose index is passed. *)

    val get_column_label : t -> int -> string
        (** Returns the label for the column whose index is passed. *)

    val get_column_name : t -> int -> string
        (** Returns the name for the column whose index is passed. *)

    val get_column_type : t -> int -> jdbc_type
        (** Returns the type for the column whose index is passed. *)

    val get_column_type_name : t -> int -> string
        (** Returns the type name for the column whose index is passed. *)

    val get_precision : t -> int -> int
        (** Returns the precision for the column whose index is passed. *)

    val get_scale : t -> int -> int
        (** Returns the scale for the column whose index is passed. *)

    val get_schema_name : t -> int -> string
        (** Returns the schema name for the column whose index is passed. *)

    val get_table_name : t -> int -> string
        (** Returns the table name for the column whose index is passed. *)

    val is_auto_increment : t -> int -> bool
        (** Tests whether the column whose index is passed is auto increment. *)

    val is_case_sensitive : t -> int -> bool
        (** Tests whether the column whose index is passed is case sensitive. *)

    val is_currency : t -> int -> bool
        (** Tests whether the column whose index is passed contains currency values. *)

    val is_definitely_writable : t -> int -> bool
        (** Tests whether the column whose index is passed is definitely writable. *)

    val is_nullable : t -> int -> nullable
        (** Tests whether the column whose index is passed is nullable. *)

    val is_read_only : t -> int -> bool
        (** Tests whether the column whose index is passed is read-only. *)

    val is_searchable : t -> int -> bool
        (** Tests whether the column whose index is passed is searchable. *)

    val is_signed : t -> int -> bool
        (** Tests whether the column whose index is passed is signed. *)

    val is_writable : t -> int -> bool
        (** Tests whether the column whose index is passed is writable. *)
  end
(** Metadata for result sets. *)

module ResultSet :
  sig
    type t
        (** Type of result sets. *)

    val absolute : t -> int32 -> bool
        (** [absolute rs row] moves the cursor of the passed result set to the row
            numbered [row]. *)

    val after_last : t -> unit
        (** Moves the cursor to the end of the passed result set. *)

    val before_first : t -> unit
        (** Moves the cursor to the start of the passed result set. *)

    val cancel_row_updates : t -> unit
        (** Cancels the updates made to the current row. *)

    val clear_warnings : t -> unit
        (** Clears the list of warnings currently reported by the passed result set. *)

    val close : t -> unit
        (** Closes the passed result set. *)

    val delete_row : t -> unit
        (** Deletes the current row. *)

    val find_column : t -> string -> int
        (** [find_column rs label] returns the index of the column with label
            [label] for the passed result set. *)

    val first : t -> bool
        (** Moves the cursor to the first row of the passed result set.
            Returns a boolean indicating whether such a row actually exists. *)

    val get_array_idx : t -> int -> Array.t option
        (** Returns the column at the given index as an array value. *)

    val get_array_name : t -> string -> Array.t option
        (** Returns the column with the given name as an array value. *)

    val get_ascii_stream_idx : t -> int -> in_channel option
        (** Returns the column at the given index as an ascii stream. *)

    val get_ascii_stream_name : t -> string -> in_channel option
        (** Returns the column with the given name as an ascii stream. *)

    val get_big_decimal_idx : t -> int -> CadmiumMath.BigDecimal.t option
        (** Returns the column at the given index as a big decimal value. *)

    val get_big_decimal_name : t -> string -> CadmiumMath.BigDecimal.t option
        (** Returns the column with the given name as a big decimal value. *)

    val get_binary_stream_idx : t -> int -> in_channel option
        (** Returns the column at the given index as a binary stream. *)

    val get_binary_stream_name : t -> string -> in_channel option
        (** Returns the column with the given name as a binary stream. *)

    val get_blob_idx : t -> int -> Blob.t option
        (** Returns the column at the given index as a blob value. *)

    val get_blob_name : t -> string -> Blob.t option
        (** Returns the column with the given name as a blob value. *)

    val get_boolean_idx : t -> int -> bool option
        (** Returns the column at the given index as a boolean value. *)

    val get_boolean_name : t -> string -> bool option
        (** Returns the column with the given name as a boolean value. *)

    val get_byte_idx : t -> int -> int option
        (** Returns the column at the given index as a byte value. *)

    val get_byte_name : t -> string -> int option
        (** Returns the column with the given name as a byte value. *)

    val get_bytes_idx : t -> int -> string option
        (** Returns the column at the given index as an array of bytes. *)

    val get_bytes_name : t -> string -> string option
        (** Returns the column with the given name as an array of bytes. *)

    val get_clob_idx : t -> int -> Clob.t option
        (** Returns the column at the given index as a clob value. *)

    val get_clob_name : t -> string -> Clob.t option
        (** Returns the column with the given name as a clob value. *)

    val get_concurrency : t -> result_set_concurrency
        (** Returns the concurrency of the passed result set. *)

    val get_cursor_name : t -> string
        (** Returns the cursor name of the passed result set. *)

    val get_date_idx : t -> int -> Date.t option
        (** Returns the column at the given index as a date value. *)

    val get_date_name : t -> string -> Date.t option
        (** Returns the column with the given name as a date value. *)

    val get_double_idx : t -> int -> float option
        (** Returns the column at the given index as a double value. *)

    val get_double_name : t -> string -> float option
        (** Returns the column with the given name as a double value. *)

    val get_fetch_direction : t -> result_set_fetch_direction
        (** Returns the fetch direction for the passed result set. *)

    val get_fetch_size : t -> int32
        (** Returns the fetch size for the passed result set. *)

    val get_float_idx : t -> int -> float option
        (** Returns the column at the given index as a float value. *)

    val get_float_name : t -> string -> float option
        (** Returns the column with the given name as a float value. *)

    val get_holdability : t -> result_set_holdability
        (** Returns the holdability for the passed result set. *)

    val get_int_idx : t -> int -> int32 option
        (** Returns the column at the given index as a int value. *)

    val get_int_name : t -> string -> int32 option
        (** Returns the column with the given name as a int value. *)

    val get_long_idx : t -> int -> int64 option
        (** Returns the column at the given index as a long value. *)

    val get_long_name : t -> string -> int64 option
        (** Returns the column with the given name as a long value. *)

    val get_meta_data : t -> ResultSetMetaData.t
        (** Returns the metadata for the passed result set. *)

    val get_nclob_idx : t -> int -> NClob.t option
        (** Returns the column at the given index as a nclob value. *)

    val get_nclob_name : t -> string -> NClob.t option
        (** Returns the column with the given name as a nclob value. *)

    val get_nstring_idx : t -> int -> string option
        (** Returns the column at the given index as a nstring value. *)

    val get_nstring_name : t -> string -> string option
        (** Returns the column with the given name as a nstring value. *)

    val get_object_idx : t -> int -> Cadmium.java_object option
        (** Returns the column at the given index as an object value. *)

    val get_object_name : t -> string -> Cadmium.java_object option
        (** Returns the column with the given name as an object value. *)

    val get_ref_idx : t -> int -> Ref.t option
        (** Returns the column at the given index as a reference value. *)

    val get_ref_name : t -> string -> Ref.t option
        (** Returns the column with the given name as a reference value. *)

    val get_row : t -> int32
        (** Returns the current row number for the passed result set. *)

    val get_row_id_idx : t -> int -> RowId.t option
        (** Returns the column at the given index as a row identifier value. *)

    val get_row_id_name : t -> string -> RowId.t option
        (** Returns the column with the given name as a row identifier value. *)

    val get_short_idx : t -> int -> int option
        (** Returns the column at the given index as a short value. *)

    val get_short_name : t -> string -> int option
        (** Returns the column with the given name as a short value. *)

    val get_sqlxml_idx : t -> int -> SQLXML.t option
        (** Returns the column at the given index as a SQL XML value. *)

    val get_sqlxml_name : t -> string -> SQLXML.t option
        (** Returns the column with the given name as a SQL XML value. *)

    val get_string_idx : t -> int -> string option
        (** Returns the column at the given index as a string value. *)

    val get_string_name : t -> string -> string option
        (** Returns the column with the given name as a string value. *)

    val get_time_idx : t -> int -> Time.t option
        (** Returns the column at the given index as a time value. *)

    val get_time_name : t -> string -> Time.t option
        (** Returns the column with the given name as a time value. *)

    val get_timestamp_idx : t -> int -> Timestamp.t option
        (** Returns the column at the given index as a timestamp value. *)

    val get_timestamp_name : t -> string -> Timestamp.t option
        (** Returns the column with the given name as a timestamp value. *)

    val get_type : t -> result_set_type
        (** Returns the type of the passed result set. *)

    val get_url_idx : t -> int -> string option
        (** Returns the column at the given index as a URL value. *)

    val get_url_name : t -> string -> string option
        (** Returns the column with the given name as a URL value. *)

    val get_warnings : t -> sql_warning list
        (** Returns the list of warnings currently reported by the passed result set. *)

    val insert_row : t -> unit
        (** Inserts a row into the passed result set. *)

    val is_after_last : t -> bool
        (** Tests whether the cursor of the passed result set is after the last row. *)

    val is_before_first : t -> bool
        (** Tests whether the cursor of the passed result set is before the first row. *)

    val is_closed : t -> bool
        (** Tests whether the passed result set is closed. *)

    val is_first : t -> bool
        (** Tests whether the cursor of the passed result set is on the first row. *)

    val is_last : t -> bool
        (** Tests whether the cursor of the passed result set is on the last row. *)

    val last : t -> bool
        (** Moves the cursor to the last row of the passed result set.
            Returns a boolean indicating whether such a row actually exists. *)

    val move_to_current_row : t -> unit
        (** Moves the cursor of the passed result set to the current row. *)

    val move_to_insert_row : t -> unit
        (** Moves the cursor of the passed result set to the insert row. *)

    val next : t -> bool
        (** Moves the cursor of the passed result set to the next row.
            Returns a boolean indicating whether such a row actually exists. *)

    val previous : t -> bool
        (** Moves the cursor of the passed result set to the previous row.
            Returns a boolean indicating whether such a row actually exists. *)

    val refresh_row : t -> unit
        (** Refreshes the content of the current row from the database. *)

    val relative : t -> int32 -> bool
        (** [relative rs rows] moves the cursor of the passed result set by a
            relative number of rows given by [rows].
            Returns a boolean indicating whether such a row actually exists. *)

    val row_deleted : t -> bool
        (** Tests whether a row has been deleted. *)

    val row_inserted : t -> bool
        (** Tests whether a row has been inserted. *)

    val row_updated : t -> bool
        (** Tests whether a row has been updated. *)

    val set_fetch_direction : t -> result_set_fetch_direction -> unit
        (** Sets the fetch direction for the passed result set. *)

    val set_fetch_size : t -> int32 -> unit
        (** Sets the fetch size for the passed result set. *)

    val update_array_idx : t -> int -> Array.t option -> unit
        (** Sets the column at the given index to the passed array value. *)

    val update_array_name : t -> string -> Array.t option -> unit
        (** Sets the column with the given name to the passed array value. *)

    val update_ascii_stream_idx : t -> int -> in_channel option -> unit
        (** Sets the column at the given index to the value read from the passed stream. *)

    val update_ascii_stream_name : t -> string -> in_channel option -> unit
        (** Sets the column with the given name to the value read from the passed stream. *)

    val update_ascii_stream_length_idx : t -> int -> in_channel option -> int32 -> unit
        (** Sets the column at the given index to the value read from the passed stream.
            The passed int32 indicates how many bytes should be read from the stream. *)

    val update_ascii_stream_length_name : t -> string -> in_channel option -> int32 -> unit
        (** Sets the column with the given name to the value read from the passed stream.
            The passed int32 indicates how many bytes should be read from the stream. *)

    val update_ascii_stream_length64_idx : t -> int -> in_channel option -> int64 -> unit
        (** Sets the column at the given index to the value read from the passed stream.
            The passed int64 indicates how many bytes should be read from the stream. *)

    val update_ascii_stream_length64_name : t -> string -> in_channel option -> int64 -> unit
        (** Sets the column with the given name to the value read from the passed stream.
            The passed int64 indicates how many bytes should be read from the stream. *)

    val update_big_decimal_idx : t -> int -> CadmiumMath.BigDecimal.t option -> unit
        (** Sets the column at the given index to the passed big decimal value. *)

    val update_big_decimal_name : t -> string -> CadmiumMath.BigDecimal.t option -> unit
        (** Sets the column with the given name to the passed big decimal value. *)

    val update_binary_stream_idx : t -> int -> in_channel option -> unit
        (** Sets the column at the given index to the value read from the passed stream. *)

    val update_binary_stream_name : t -> string -> in_channel option -> unit
        (** Sets the column with the given name to the value read from the passed stream. *)

    val update_binary_stream_length_idx : t -> int -> in_channel option -> int32 -> unit
        (** Sets the column at the given index to the value read from the passed stream.
            The passed int32 indicates how many bytes should be read from the stream. *)

    val update_binary_stream_length_name : t -> string -> in_channel option -> int32 -> unit
        (** Sets the column with the given name to the value read from the passed stream.
            The passed int32 indicates how many bytes should be read from the stream. *)

    val update_binary_stream_length64_idx : t -> int -> in_channel option -> int64 -> unit
        (** Sets the column at the given index to the value read from the passed stream.
            The passed int64 indicates how many bytes should be read from the stream. *)

    val update_binary_stream_length64_name : t -> string -> in_channel option -> int64 -> unit
        (** Sets the column with the given name to the value read from the passed stream.
            The passed int64 indicates how many bytes should be read from the stream. *)

    val update_blob_idx : t -> int -> Blob.t option -> unit
        (** Sets the column at the given index to the passed blob value. *)

    val update_blob_name : t -> string -> Blob.t option -> unit
        (** Sets the column with the given name to the passed blob value. *)

    val update_boolean_idx : t -> int -> bool option -> unit
        (** Sets the column at the given index to the passed boolean value. *)

    val update_boolean_name : t -> string -> bool option -> unit
        (** Sets the column with the given name to the passed boolean value. *)

    val update_byte_idx : t -> int -> int option -> unit
        (** Sets the column at the given index to the passed byte value. *)

    val update_byte_name : t -> string -> int option -> unit
        (** Sets the column with the given name to the passed byte value. *)

    val update_bytes_idx : t -> int -> string option -> unit
        (** Sets the column at the given index to the passed byte array value. *)

    val update_bytes_name : t -> string -> string option -> unit
        (** Sets the column with the given name to the passed byte array value. *)

    val update_clob_idx : t -> int -> Clob.t option -> unit
        (** Sets the column at the given index to the passed clob value. *)

    val update_clob_name : t -> string -> Clob.t option -> unit
        (** Sets the column with the given name to the passed clob value. *)

    val update_date_idx : t -> int -> Date.t option -> unit
        (** Sets the column at the given index to the passed date value. *)

    val update_date_name : t -> string -> Date.t option -> unit
        (** Sets the column with the given name to the passed date value. *)

    val update_double_idx : t -> int -> float option -> unit
        (** Sets the column at the given index to the passed double value. *)

    val update_double_name : t -> string -> float option -> unit
        (** Sets the column with the given name to the passed double value. *)

    val update_float_idx : t -> int -> float option -> unit
        (** Sets the column at the given index to the passed float value. *)

    val update_float_name : t -> string -> float option -> unit
        (** Sets the column with the given name to the passed float value. *)

    val update_int_idx : t -> int -> int32 option -> unit
        (** Sets the column at the given index to the passed int value. *)

    val update_int_name : t -> string -> int32 option -> unit
        (** Sets the column with the given name to the passed int value. *)

    val update_long_idx : t -> int -> int64 option -> unit
        (** Sets the column at the given index to the passed long value. *)

    val update_long_name : t -> string -> int64 option -> unit
        (** Sets the column with the given name to the passed long value. *)

    val update_nclob_idx : t -> int -> NClob.t option -> unit
        (** Sets the column at the given index to the passed nclob value. *)

    val update_nclob_name : t -> string -> NClob.t option -> unit
        (** Sets the column with the given name to the passed nclob value. *)

    val update_nstring_idx : t -> int -> string option -> unit
        (** Sets the column at the given index to the passed nstring value. *)

    val update_nstring_name : t -> string -> string option -> unit
        (** Sets the column with the given name to the passed nstring value. *)

    val update_object_idx : t -> int -> Cadmium.java_object option -> unit
        (** Sets the column at the given index to the passed object value. *)

    val update_object_name : t -> string -> Cadmium.java_object option -> unit
        (** Sets the column with the given name to the passed object value. *)

    val update_object_scale_idx : t -> int -> Cadmium.java_object option -> int -> unit
        (** Sets the column at the given index to the passed object value. *)

    val update_object_scale_name : t -> string -> Cadmium.java_object option -> int -> unit
        (** Sets the column with the given name to the passed object value. *)

    val update_ref_idx : t -> int -> Ref.t option -> unit
        (** Sets the column at the given index to the passed reference value. *)

    val update_ref_name : t -> string -> Ref.t option -> unit
        (** Sets the column with the given name to the passed reference value. *)

    val update_row : t -> unit
        (** Updates the database with the content of the current row. *)

    val update_row_id_idx : t -> int -> RowId.t option -> unit
        (** Sets the column at the given index to the passed row identifier value. *)

    val update_row_id_name : t -> string -> RowId.t option -> unit
        (** Sets the column with the given name to the passed row identifier value. *)

    val update_short_idx : t -> int -> int option -> unit
        (** Sets the column at the given index to the passed short value. *)

    val update_short_name : t -> string -> int option -> unit
        (** Sets the column with the given name to the passed short value. *)

    val update_sqlxml_idx : t -> int -> SQLXML.t option -> unit
        (** Sets the column at the given index to the passed SQL XML value. *)

    val update_sqlxml_name : t -> string -> SQLXML.t option -> unit
        (** Sets the column with the given name to the passed SQL XML value. *)

    val update_string_idx : t -> int -> string option -> unit
        (** Sets the column at the given index to the passed string value. *)

    val update_string_name : t -> string -> string option -> unit
        (** Sets the column with the given name to the passed string value. *)

    val update_time_idx : t -> int -> Time.t option -> unit
        (** Sets the column at the given index to the passed time value. *)

    val update_time_name : t -> string -> Time.t option -> unit
        (** Sets the column with the given name to the passed time value. *)

    val update_timestamp_idx : t -> int -> Timestamp.t option -> unit
        (** Sets the column at the given index to the passed timestamp value. *)

    val update_timestamp_name : t -> string -> Timestamp.t option -> unit
        (** Sets the column with the given name to the passed timestamp value. *)
  end
(** Result sets. *)


(** {6 Statements} *)

module ParameterMetaData :
  sig
    type t
        (** Type of metadata for parameters. *)

    val get_parameter_class_name : t -> int -> string
        (** Returns the class name for the parameter whose index is passed. *)

    val get_parameter_count : t -> int
        (** Returns the number of parameters. *)

    val get_parameter_mode : t -> int -> parameter_mode
        (** Returns the mode for the parameter whose index is passed. *)

    val get_parameter_type : t -> int -> jdbc_type
        (** Returns the type for the parameter whose index is passed. *)

    val get_parameter_type_name : t -> int -> string
        (** Returns the type name for the parameter whose index is passed. *)

    val get_precision : t -> int -> int
        (** Returns the precision for the parameter whose index is passed. *)

    val get_scale : t -> int -> int
        (** Returns the scale for the parameter whose index is passed. *)

    val is_nullable : t -> int -> nullable
        (** Tests whether the parameter whose index is passed is nullable. *)

    val is_signed : t -> int -> bool
        (** Tests whether the parameter whose index is passed is signed. *)
  end
(** Metadata for parameters. *)

module Statement :
  sig
    type t
        (** Type of statements. *)

    val add_batch : t -> string -> unit
        (** [add_batch s sql] adds the SQL command [sql] to the list of commands of
            statement [s]. *)

    val cancel : t -> unit
        (** Cancels the passed statement. *)

    val clear_batch : t -> unit
        (** Clears the list of commands for the passed statement. *)

    val clear_warnings : t -> unit
        (** Clears the list of warnings currently reported by the passed statement. *)

    val close : t -> unit
        (** Closes the passed statement. *)

    val execute : t -> string -> bool
        (** [execute s sql] executes the SQL command [sql] and returns whether the
            execution produced a result set. *)

    val execute_autogen : t -> string -> statement_autogen_keys -> bool
        (** [execute_autogen s sql sak] executes the SQL command [sql] and returns
            whether the execution produced a result set. [sak] indicates whether
            auto-generated values should be retrievable. *)

    val execute_idx : t -> string -> int list -> bool
        (** [execute_idx s sql indexes] executes the SQL command [sql] and returns
            whether the execution produced a result set. [indexes] indicates which
            auto-generated values should be retrievable. *)

    val execute_name : t -> string -> string list -> bool
        (** [execute_name s sql names] executes the SQL command [sql] and returns
            whether the execution produced a result set. [names] indicates which
            auto-generated values should be retrievable. *)

    val execute_batch : t -> statement_result list
        (** Executes the list of commands of the statement and returns a list of
            update counts. *)

    val execute_query : t -> string -> ResultSet.t
        (** [execute_query s sql] executes the SQL command [sql] and returns its
            result set. *)

    val execute_update : t -> string -> int32
        (** [execute_update s sql] executes the SQL command [sql] and returns its
            update count. *)

    val execute_update_autogen : t -> string -> statement_autogen_keys -> int32
        (** [execute_update_autogen s sql sak] executes the SQL command [sql] and
            returns its update count. [sak] indicates whether auto-generated values
            should be retrievable. *)

    val execute_update_idx : t -> string -> int list -> int32
        (** [execute_update_idx s sql indexes] executes the SQL command [sql] and
            returns its update count. [indexes] indicates which auto-generated values
            should be retrievable. *)

    val execute_update_name : t -> string -> string list -> int32
        (** [execute_update_name s sql names] executes the SQL command [sql] and
            returns its update count. [names] indicates which auto-generated values
            should be retrievable. *)

    val get_fetch_direction : t -> result_set_fetch_direction
        (** Returns the fetch direction of result sets returned by the passed statement. *)

    val get_fetch_size : t -> int32
        (** Returns the fetch size of result sets returned by the passed statement. *)

    val get_generated_keys : t -> ResultSet.t
        (** Returns a result set containing the values that were auto-generated by
            the last update command. *)

    val get_max_field_size : t -> int32
        (** Returns the maximum column size of result sets returned by the passed statement. *)

    val get_max_rows : t -> int32
        (** Returns the maximum number of rows of result sets returned by the passed statement. *)

    val get_more_results : t -> bool
        (** Move to the next result of the passed statement.
            Returns a boolean indicating whether the next result is a result set. *)

    val get_more_results_current : t -> statement_action -> bool
        (** Move to the next result of the passed statement.
            Returns a boolean indicating whether the next result is a result set.
            The second parameter indicates what should happen to the current result. *)

    val get_query_timeout : t -> int32
        (** Returns the timeout of the passed statement. *)

    val get_result_set : t -> ResultSet.t
        (** Returns the current result as a result set. *)

    val get_result_set_concurrency : t -> result_set_concurrency
        (** Returns the concurrency of result sets returned by the passed statement. *)

    val get_result_set_holdability : t -> result_set_holdability
        (** Returns the holdability of result sets returned by the passed statement. *)

    val get_result_set_type : t -> result_set_type
        (** Returns the type of result sets returned by the passed statement. *)

    val get_update_count : t -> int32
        (** Returns the current result as an update count. *)

    val get_warnings : t -> sql_warning list
        (** Returns the list of warnings currently reported by the passed statement. *)

    val is_closed : t -> bool
        (** Tests whether the passed statement is closed. *)

    val is_poolable : t -> bool
        (** Tests whether the passed statement is poolable. *)

    val set_cursor_name : t -> string -> unit
        (** Sets the name of the cursor that will be used in further executions. *)

    val set_escape_processing : t -> bool -> unit
        (** Sets escape processing on or off. *)

    val set_fetch_direction : t -> result_set_fetch_direction -> unit
        (** Sets the fetch direction of result sets returned by the passed statement. *)

    val set_fetch_size : t -> int32 -> unit
        (** Sets the fetch size of result sets returned by the passed statement. *)

    val set_max_field_size : t -> int32 -> unit
        (** Sets the maximum column size of result sets returned by the passed statement. *)

    val set_max_rows : t -> int32 -> unit
        (** Sets the maximum number of rows of result sets returned by the passed statement. *)

    val set_poolable : t -> bool -> unit
        (** Sets whether the statement is poolable or not. *)

    val set_query_timeout : t -> int32 -> unit
        (** Sets the timeout of the passed statement. *)
  end
(** Simple statements. *)

module PreparedStatement :
  sig
    type t
        (** Type of prepared statements. *)

    val as_statement : t -> Statement.t
        (** {i Casts} the passed prepared statement into a bare statement. *)

    val add_batch : t -> unit
        (** Adds a set of parameters to the passed prepared statement. *)

    val clear_parameters : t -> unit
        (** Clears the current parameter values. *)

    val execute : t -> bool
        (** Execute the prepared statement and returns whether the execution produced
            a result set. *)

    val execute_query : t -> ResultSet.t
        (** Execute the prepared statement and returns its result set. *)

    val execute_update : t -> int32
        (** Execute the prepared statement and returns its update count. *)

    val get_meta_data : t -> ResultSetMetaData.t
        (** Returns the metadata about result sets for the passed prepared statement. *)

    val get_parameter_meta_data : t -> ParameterMetaData.t
        (** Returns the metadata about parameters for the passed prepared statement. *)

    val set_array_idx : t -> int -> Array.t option -> unit
        (** Sets the parameter at the given index to the passed array value. *)

    val set_ascii_stream_idx : t -> int -> in_channel option -> unit
        (** Sets the parameter at the given index to the value read from the passed stream. *)

    val set_ascii_stream_length_idx : t -> int -> in_channel option -> int32 -> unit
        (** Sets the parameter at the given index to the value read from the passed stream.
            The passed int32 indicates how many bytes should be read from the stream. *)

    val set_ascii_stream_length64_idx : t -> int -> in_channel option -> int64 -> unit
        (** Sets the parameter at the given index to the value read from the passed stream.
            The passed int64 indicates how many bytes should be read from the stream. *)

    val set_big_decimal_idx : t -> int -> CadmiumMath.BigDecimal.t option -> unit
        (** Sets the parameter at the given index to the passed big decimal value. *)

    val set_binary_stream_idx : t -> int -> in_channel option -> unit
        (** Sets the parameter at the given index to the value read from the passed stream. *)

    val set_binary_stream_length_idx : t -> int -> in_channel option -> int32 -> unit
        (** Sets the parameter at the given index to the value read from the passed stream.
            The passed int32 indicates how many bytes should be read from the stream. *)

    val set_binary_stream_length64_idx : t -> int -> in_channel option -> int64 -> unit
        (** Sets the parameter at the given index to the value read from the passed stream.
            The passed int64 indicates how many bytes should be read from the stream. *)

    val set_blob_idx : t -> int -> Blob.t option -> unit
        (** Sets the parameter at the given index to the passed blob value. *)

    val set_boolean_idx : t -> int -> bool option -> unit
        (** Sets the parameter at the given index to the passed boolean value. *)

    val set_byte_idx : t -> int -> int option -> unit
        (** Sets the parameter at the given index to the passed byte value. *)

    val set_bytes_idx : t -> int -> string option -> unit
        (** Sets the parameter at the given index to the passed byte array value. *)

    val set_clob_idx : t -> int -> Clob.t option -> unit
        (** Sets the parameter at the given index to the passed clob value. *)

    val set_date_idx : t -> int -> Date.t option -> unit
        (** Sets the parameter at the given index to the passed date value. *)

    val set_double_idx : t -> int -> float option -> unit
        (** Sets the parameter at the given index to the passed double value. *)

    val set_float_idx : t -> int -> float option -> unit
        (** Sets the parameter at the given index to the passed float value. *)

    val set_int_idx : t -> int -> int32 option -> unit
        (** Sets the parameter at the given index to the passed int value. *)

    val set_long_idx : t -> int -> int64 option -> unit
        (** Sets the parameter at the given index to the passed long value. *)

    val set_nclob_idx : t -> int -> NClob.t option -> unit
        (** Sets the parameter at the given index to the passed nclob value. *)

    val set_nstring_idx : t -> int -> string option -> unit
        (** Sets the parameter at the given index to the passed nstring value. *)

    val set_object_idx : t -> int -> Cadmium.java_object option -> unit
        (** Sets the parameter at the given index to the passed object value. *)

    val set_object_type_idx : t -> int -> Cadmium.java_object option -> jdbc_type -> unit
        (** Sets the parameter at the given index to the passed object value. *)

    val set_object_type_scale_idx : t -> int -> Cadmium.java_object option -> jdbc_type -> int -> unit
        (** Sets the parameter at the given index to the passed object value. *)

    val set_ref_idx : t -> int -> Ref.t option -> unit
        (** Sets the parameter at the given index to the passed reference value. *)

    val set_row_id_idx : t -> int -> RowId.t option -> unit
        (** Sets the parameter at the given index to the passed row identifier value. *)

    val set_short_idx : t -> int -> int option -> unit
        (** Sets the parameter at the given index to the passed short value. *)

    val set_sqlxml_idx : t -> int -> SQLXML.t option -> unit
        (** Sets the parameter at the given index to the passed SQL XML value. *)

    val set_string_idx : t -> int -> string option -> unit
        (** Sets the parameter at the given index to the passed string value. *)

    val set_time_idx : t -> int -> Time.t option -> unit
        (** Sets the parameter at the given index to the passed time value. *)

    val set_timestamp_idx : t -> int -> Timestamp.t option -> unit
        (** Sets the parameter at the given index to the passed timestamp value. *)

    val set_url_idx : t -> int -> string option -> unit
        (** Sets the parameter at the given index to the passed URL value. *)
  end
(** Prepared statements. *)

module CallableStatement :
  sig
    type t
        (** Type of callable statements. *)

    val as_statement : t -> Statement.t
        (** {i Casts} the passed callable statement into a bare statement. *)

    val as_prepared_statement : t -> PreparedStatement.t
        (** {i Casts} the passed callable statement into a prepared statement. *)

    val get_array_idx : t -> int -> Array.t option
        (** Returns the parameter at the given index as an array value. *)

    val get_array_name : t -> string -> Array.t option
        (** Returns the parameter with the given name as an array value. *)

    val get_big_decimal_idx : t -> int -> CadmiumMath.BigDecimal.t option
        (** Returns the parameter at the given index as a big decimal value. *)

    val get_big_decimal_name : t -> string -> CadmiumMath.BigDecimal.t option
        (** Returns the parameter with the given name as a big decimal value. *)

    val get_blob_idx : t -> int -> Blob.t option
        (** Returns the parameter at the given index as a blob value. *)

    val get_blob_name : t -> string -> Blob.t option
        (** Returns the parameter with the given name as a blob value. *)

    val get_boolean_idx : t -> int -> bool option
        (** Returns the parameter at the given index as a boolean value. *)

    val get_boolean_name : t -> string -> bool option
        (** Returns the parameter with the given name as a boolean value. *)

    val get_byte_idx : t -> int -> int option
        (** Returns the parameter at the given index as a byte value. *)

    val get_byte_name : t -> string -> int option
        (** Returns the parameter with the given name as a byte value. *)

    val get_bytes_idx : t -> int -> string option
        (** Returns the parameter at the given index as an array of bytes. *)

    val get_bytes_name : t -> string -> string option
        (** Returns the parameter with the given name as an array of bytes. *)

    val get_clob_idx : t -> int -> Clob.t option
        (** Returns the parameter at the given index as a clob value. *)

    val get_clob_name : t -> string -> Clob.t option
        (** Returns the parameter with the given name as a clob value. *)

    val get_date_idx : t -> int -> Date.t option
        (** Returns the parameter at the given index as a date value. *)

    val get_date_name : t -> string -> Date.t option
        (** Returns the parameter with the given name as a date value. *)

    val get_double_idx : t -> int -> float option
        (** Returns the parameter at the given index as a double value. *)

    val get_double_name : t -> string -> float option
        (** Returns the parameter with the given name as a double value. *)

    val get_float_idx : t -> int -> float option
        (** Returns the parameter at the given index as a float value. *)

    val get_float_name : t -> string -> float option
        (** Returns the parameter with the given name as a float value. *)

    val get_int_idx : t -> int -> int32 option
        (** Returns the parameter at the given index as a int value. *)

    val get_int_name : t -> string -> int32 option
        (** Returns the parameter with the given name as a int value. *)

    val get_long_idx : t -> int -> int64 option
        (** Returns the parameter at the given index as a long value. *)

    val get_long_name : t -> string -> int64 option
        (** Returns the parameter with the given name as a long value. *)

    val get_nclob_idx : t -> int -> NClob.t option
        (** Returns the parameter at the given index as a nclob value. *)

    val get_nclob_name : t -> string -> NClob.t option
        (** Returns the parameter with the given name as a nclob value. *)

    val get_nstring_idx : t -> int -> string option
        (** Returns the parameter at the given index as a nstring value. *)

    val get_nstring_name : t -> string -> string option
        (** Returns the parameter with the given name as a nstring value. *)

    val get_object_idx : t -> int -> Cadmium.java_object option
        (** Returns the parameter at the given index as an object value. *)

    val get_object_name : t -> string -> Cadmium.java_object option
        (** Returns the parameter with the given name as an object value. *)

    val get_ref_idx : t -> int -> Ref.t option
        (** Returns the parameter at the given index as a reference value. *)

    val get_ref_name : t -> string -> Ref.t option
        (** Returns the parameter with the given name as a reference value. *)

    val get_row_id_idx : t -> int -> RowId.t option
        (** Returns the parameter at the given index as a row identifier value. *)

    val get_row_id_name : t -> string -> RowId.t option
        (** Returns the parameter with the given name as a row identifier value. *)

    val get_short_idx : t -> int -> int option
        (** Returns the parameter at the given index as a short value. *)

    val get_short_name : t -> string -> int option
        (** Returns the parameter with the given name as a short value. *)

    val get_sqlxml_idx : t -> int -> SQLXML.t option
        (** Returns the parameter at the given index as a SQL XML value. *)

    val get_sqlxml_name : t -> string -> SQLXML.t option
        (** Returns the parameter with the given name as a SQL XML value. *)

    val get_string_idx : t -> int -> string option
        (** Returns the parameter at the given index as a string value. *)

    val get_string_name : t -> string -> string option
        (** Returns the parameter with the given name as a string value. *)

    val get_time_idx : t -> int -> Time.t option
        (** Returns the parameter at the given index as a time value. *)

    val get_time_name : t -> string -> Time.t option
        (** Returns the parameter with the given name as a time value. *)

    val get_timestamp_idx : t -> int -> Timestamp.t option
        (** Returns the parameter at the given index as a timestamp value. *)

    val get_timestamp_name : t -> string -> Timestamp.t option
        (** Returns the parameter with the given name as a timestamp value. *)

    val get_url_idx : t -> int -> string option
        (** Returns the parameter at the given index as a URL value. *)

    val get_url_name : t -> string -> string option
        (** Returns the parameter with the given name as a URL value. *)

    val register_out_parameter_idx : t -> int -> jdbc_type -> unit
        (** [register_out_parameter_idx cs idx t] registers the output parameter at
            index [idx] with the type [t]. *)

    val register_out_parameter_scale_idx : t -> int -> jdbc_type -> int -> unit
        (** [register_out_parameter_scale_idx cs idx t s] registers the output
            parameter at index [idx] with the type [t] and scale [s]. *)

    val register_out_parameter_name : t -> string -> jdbc_type -> unit
        (** [register_out_parameter_name cs name t] registers the output parameter
            named [name] with the type [t]. *)

    val register_out_parameter_scale_name : t -> string -> jdbc_type -> int -> unit
        (** [register_out_parameter_scale_name cs name t s] registers the output
            parameter named [name] with the type [t] and scale [s]. *)

    val set_ascii_stream_name : t -> string -> in_channel option -> unit
        (** Sets the parameter with the given name to the value read from the passed stream. *)

    val set_ascii_stream_length_name : t -> string -> in_channel option -> int32 -> unit
        (** Sets the parameter with the given name to the value read from the passed stream.
            The passed int32 indicates how many bytes should be read from the stream. *)

    val set_ascii_stream_length64_name : t -> string -> in_channel option -> int64 -> unit
        (** Sets the parameter with the given name to the value read from the passed stream.
            The passed int64 indicates how many bytes should be read from the stream. *)

    val set_big_decimal_name : t -> string -> CadmiumMath.BigDecimal.t option -> unit
        (** Sets the parameter with the given name to the passed big decimal value. *)

    val set_binary_stream_name : t -> string -> in_channel option -> unit
        (** Sets the parameter with the given name to the value read from the passed stream. *)

    val set_binary_stream_length_name : t -> string -> in_channel option -> int32 -> unit
        (** Sets the parameter with the given name to the value read from the passed stream.
            The passed int32 indicates how many bytes should be read from the stream. *)

    val set_binary_stream_length64_name : t -> string -> in_channel option -> int64 -> unit
        (** Sets the parameter with the given name to the value read from the passed stream.
            The passed int64 indicates how many bytes should be read from the stream. *)

    val set_blob_name : t -> string -> Blob.t option -> unit
        (** Sets the parameter with the given name to the passed blob value. *)

    val set_boolean_name : t -> string -> bool option -> unit
        (** Sets the parameter with the given name to the passed boolean value. *)

    val set_byte_name : t -> string -> int option -> unit
        (** Sets the parameter with the given name to the passed byte value. *)

    val set_bytes_name : t -> string -> string option -> unit
        (** Sets the parameter with the given name to the passed byte array value. *)

    val set_clob_name : t -> string -> Clob.t option -> unit
        (** Sets the parameter with the given name to the passed clob value. *)

    val set_date_name : t -> string -> Date.t option -> unit
        (** Sets the parameter with the given name to the passed date value. *)

    val set_double_name : t -> string -> float option -> unit
        (** Sets the parameter with the given name to the passed double value. *)

    val set_float_name : t -> string -> float option -> unit
        (** Sets the parameter with the given name to the passed float value. *)

    val set_int_name : t -> string -> int32 option -> unit
        (** Sets the parameter with the given name to the passed int value. *)

    val set_long_name : t -> string -> int64 option -> unit
        (** Sets the parameter with the given name to the passed long value. *)

    val set_nclob_name : t -> string -> NClob.t option -> unit
        (** Sets the parameter with the given name to the passed nclob value. *)

    val set_nstring_name : t -> string -> string option -> unit
        (** Sets the parameter with the given name to the passed nstring value. *)

    val set_object_name : t -> string -> Cadmium.java_object option -> unit
        (** Sets the parameter with the given name to the passed object value. *)

    val set_object_type_name : t -> string -> Cadmium.java_object option -> jdbc_type -> unit
        (** Sets the parameter with the given name to the passed object value. *)

    val set_object_type_scale_name : t -> string -> Cadmium.java_object option -> jdbc_type -> int -> unit
        (** Sets the parameter with the given name to the passed object value. *)

    val set_row_id_name : t -> string -> RowId.t option -> unit
        (** Sets the parameter with the given name to the passed row identifier value. *)

    val set_short_name : t -> string -> int option -> unit
        (** Sets the parameter with the given name to the passed short value. *)

    val set_sqlxml_name : t -> string -> SQLXML.t option -> unit
        (** Sets the parameter with the given name to the passed SQL XML value. *)

    val set_string_name : t -> string -> string option -> unit
        (** Sets the parameter with the given name to the passed string value. *)

    val set_time_name : t -> string -> Time.t option -> unit
        (** Sets the parameter with the given name to the passed time value. *)

    val set_timestamp_name : t -> string -> Timestamp.t option -> unit
        (** Sets the parameter with the given name to the passed timestamp value. *)

    val set_url_name : t -> string -> string option -> unit
        (** Sets the parameter with the given name to the passed URL value. *)
  end
(** Callable statements. *)


(** {6 Connections & Drivers} *)

module DatabaseMetaData :
  sig
    type t
        (** Type of metadata for databases. *)

    val all_procedures_are_callable : t -> bool
        (** Tests whether the current user can call all the procedures. *)

    val all_tables_are_selectable : t -> bool
        (** Tests whether the current user can select all the procedures. *)

    val auto_commit_failure_closes_all_result_sets : t -> bool
        (** Tests whether an exception at commit indicates that all open result sets are closed. *)

    val data_definition_causes_transaction_commit : t -> bool
        (** Tests whether data definition causes commit. *)

    val data_definition_ignored_in_transactions : t -> bool
        (** Tests whether data definitions are ignored in transactions. *)

    val deletes_are_detected : t -> jdbc_type -> bool
        (** Tests whether row deletion can be detected. *)

    val does_max_row_size_include_blobs : t -> bool
        (** Tests whether the maximum row size include the size of embedded blobs. *)

    val get_attributes : t -> string -> string -> string -> string -> ResultSet.t
        (** [get_attributes dbm catalog schema_patt type_name_patt attr_name_patt] returns a
            result set containing information for user-defined types for the passed catalog,
            schema pattern, type name pattern, and attribute name pattern. *)

    val get_best_row_identifier : t -> string -> string -> string -> result_scope -> bool -> ResultSet.t
        (** [get_best_row_identifier dbm catalog schema table scope] returns a result set
            containing information regarding the optimal sets of columns that uniquely
            identify a row for the passed catalog, schema, table, and scope. *)

    val get_catalogs : t -> ResultSet.t
        (** Returns the list of catalog names. *)

    val get_catalog_separator : t -> string
        (** Returns the separator used between catalog and table name. *)

    val get_catalog_term : t -> string
        (** Returns the preferred term for 'catalog'. *)

    val get_client_info_properties : t -> ResultSet.t
        (** Returns the client info properties for the driver currently in use. *)

    val get_column_privileges : t -> string -> string -> string -> string -> ResultSet.t
        (** [get_column_privileges dbm catalog schema table column_name_patt] returns a
            result set containing information about access permissions for the passed
            catalog, schema, table, and column name pattern. *)

    val get_columns : t -> string -> string -> string -> string -> ResultSet.t
        (** [get_columns dbm catalog schema_patt table_name_patt column_name_patt] returns a
            result set containing information about the column available for the passed
            catalog, schema pattern, table name pattern, and column name pattern. *)

    val get_cross_reference : t -> string -> string -> string -> string -> string -> string -> ResultSet.t
        (** [get_cross_reference dbm parent_catalog parent_schema parent_table foreign_catalog foreign_schema foreign_table]
            returns a result set containing information about cross references between passed
            parent ({i i. e.} primary) and foreign catalogs, schemas, and tables. *)

    val get_database_major_version : t -> int32
        (** Returns the major component of the database's version. *)

    val get_database_minor_version : t -> int32
        (** Returns the minor component of the database's version. *)

    val get_database_product_name : t -> string
        (** Returns the name of the database. *)

    val get_database_product_version : t -> string
        (** Returns the version of the database. *)

    val get_default_transaction_isolation : t -> transaction_isolation
        (** Returns the default transaction level. *)

    val get_driver_major_version : t -> int32
        (** Returns the major component of the driver's version. *)

    val get_driver_minor_version : t -> int32
        (** Returns the minor component of the driver's version. *)

    val get_driver_name : t -> string
        (** Returns the name of the driver. *)

    val get_driver_version : t -> string
        (** Returns the version of the driver. *)

    val get_exported_keys : t -> string -> string -> string -> ResultSet.t
        (** [get_exported_keys dbm catalog schema table] returns a result set containing
            information about the foreign key columns referencing the passed table in the
            given catalog and schema. *)

    val get_extra_name_characters : t -> string
        (** Returns the characters (besides letters, numbers, and underscore) that can be
            used in identifiers. *)

    val get_function_columns : t -> string -> string -> string -> string -> ResultSet.t
        (** [get_function_columns dbm catalog schema_patt function_name_patt column_name_patt]
            returns a result set containing information about all the function parameters for
            the passed catalog, schema pattern, function name pattern, and column name
            pattern. *)

    val get_functions : t -> string -> string -> string -> ResultSet.t
        (** [get_functions dbm catalog schema_patt function_name_patt] returns a result set
            containing information about all the functions for the passed catalog, schema
            pattern, function name pattern, and column name pattern. *)

    val get_identifier_quote_string : t -> string
        (** Returns the string used to quote identifiers. *)

    val get_imported_keys : t -> string -> string -> string -> ResultSet.t
        (** [get_imported_keys dbm catalog schema table] returns a result set containing
            information about imported keys for the passed catalog, schema, and table. *)

    val get_index_info : t -> string -> string -> string -> bool -> bool -> ResultSet.t
        (** [get_index_info dbm catalog schema table unique approximate] returns a result set
            containing information about indexes for the passed catalog, schema, and table.
            [unique] indicates whether information should be provided only for indexes over
            unique values, while [approximate] indicates whether result may include
            approximate values. *)

    val get_jdbc_major_version : t -> int32
        (** Returns the major component of the JDBC version. *)

    val get_jdbc_minor_version : t -> int32
        (** Returns the minor component of the JDBC version. *)

    val get_max_binary_literal_length : t -> int32
        (** Returns the maximum length that can be used for a binary literal.  *)

    val get_max_catalog_name_length : t -> int32
        (** Returns the maximum length that can be used for a catalog name.  *)

    val get_max_char_literal_length : t -> int32
        (** Returns the maximum length that can be used for a literal.  *)

    val get_max_column_name_length : t -> int32
        (** Returns the maximum length that can be used for a column name.  *)

    val get_max_columns_in_group_by : t -> int32
        (** Returns the maximum number of columns that can be used in a {i group by}. *)

    val get_max_columns_in_index : t -> int32
        (** Returns the maximum number of columns that can be used in an index. *)

    val get_max_columns_in_order_by : t -> int32
        (** Returns the maximum number of columns that can be used in an {i order by}. *)

    val get_max_columns_in_select : t -> int32
        (** Returns the maximum number of columns that can be used in a {i select}. *)

    val get_max_columns_in_table : t -> int32
        (** Returns the maximum number of columns that can be used in a table. *)

    val get_max_connections : t -> int32
        (** Returns the maximum number of connections to the database. *)

    val get_max_cursor_name_length : t -> int32
        (** Returns the maximum length for a cursor name.  *)

    val get_max_index_length : t -> int32
        (** Returns the maximum length of an index.  *)

    val get_max_procedure_name_length : t -> int32
        (** Returns the maximum length for a procedure name.  *)

    val get_max_row_size : t -> int32
        (** Returns the maximum row size.  *)

    val get_max_schema_name_length : t -> int32
        (** Returns the maximum length for a schema name.  *)

    val get_max_statement_length : t -> int32
        (** Returns the maximum length of a statement.  *)

    val get_max_statements : t -> int32
        (** Returns the maximum number of statements to the database. *)

    val get_max_table_name_length : t -> int32
        (** Returns the maximum length for a table name.  *)

    val get_max_tables_in_select : t -> int32
        (** Returns the maximum number of tables that can be used in a {i select}. *)

    val get_max_user_name_length : t -> int32
        (** Returns the maximum length for a user name.  *)

    val get_numeric_functions : t -> string
        (** Returns the list of available numeric functions.
            Names are separated by commas. *)

    val get_primary_keys : t -> string -> string -> string -> ResultSet.t
        (** [get_primary_keys dbm catalog schema table] returns a result set containing
            information about primary keys for the passed catalog, schema, and table. *)

    val get_procedure_columns : t -> string -> string -> string -> string -> ResultSet.t
        (** [get_procedure_columns dbm catalog schema_patt procedure_name_patt column_name_patt]
            returns a result set containing information about all the procedure parameters
            for the passed catalog, schema pattern, procedure name pattern, and column name
            pattern. *)

    val get_procedures : t -> string -> string -> string -> ResultSet.t
        (** [get_procedures dbm catalog schema_patt procedure_name_patt] returns a
            result set containing information about all the procedures for the
            passed catalog, schema pattern, and procedure name pattern. *)

    val get_procedure_term : t -> string
        (** Returns the preferred term for 'procedure'. *)

    val get_result_set_holdability : t -> result_set_holdability
        (** Returns the default holdability of the returned result sets. *)

    val get_row_id_lifetime : t -> rowid_lifetime
        (** Returns the lifetime of a row identifier. *)

    val get_schemas : t -> ResultSet.t
        (** Returns a result set containing the list of available schemas. *)

    val get_schemas_params : t -> string -> string -> ResultSet.t
        (** [get_schemas_params dbm catalog schema_patt] returns a result set
            containing information about the available schemas for the passed
            catalog, and schema pattern. *)

    val get_schema_term : t -> string
        (** Returns the preferred term for 'schema'. *)

    val get_search_string_escape : t -> string
        (** Returns the string used to escape wildcard characters. *)

    val get_sql_keywords : t -> string
        (** Returns the list of SQL keywords recognized by the database.
            Names are separated by commas. *)

    val get_sql_state_type : t -> sql_state_type
        (** Returns the kind of SQL state used by the database. *)

    val get_string_functions : t -> string
        (** Returns the list of available string functions.
            Names are separated by commas. *)

    val get_super_tables : t -> string -> string -> string -> ResultSet.t
        (** [get_super_tables dbm catalog schema_patt table_name_patt] returns a
            result set containing information about the table hierarchy for the
            passed catalog, schema pattern, and table name pattern. *)

    val get_super_types : t -> string -> string -> string -> ResultSet.t
        (** [get_super_types dbm catalog schema_patt type_name_patt] returns a
            result set containing information about the user-defined types for the
            passed catalog, schema pattern, and type name pattern. *)

    val get_system_functions : t -> string
        (** Returns the list of available system function.
            Names are separated by commas. *)

    val get_table_privileges : t -> string -> string -> string -> ResultSet.t
        (** [get_table_privileges dbm catalog schema_patt type_name_patt] returns a
            result set containing information about access permissions for the passed
            catalog, schema pattern, and type name pattern. *)

    val get_tables : t -> string -> string -> string -> string list -> ResultSet.t
        (** [get_tables dbm catalog schema_patt table_name_patt types] returns a
            result set containing information about all the available tables for
            the passed catalog, schema pattern, table name pattern, and types. *)

    val get_table_types : t -> ResultSet.t
        (** Returns a result set containing information about all the available
            table types. *)

    val get_time_date_functions : t -> string
        (** Returns the list of available time/date functions.
            Names are separated by commas. *)

    val get_type_info : t -> ResultSet.t
        (** Returns a result set containing information about all the available types. *)

    val get_udts : t -> string -> string -> string -> jdbc_type list -> ResultSet.t
        (** [get_udts dbm catalog schema_patt type_name_patt types] returns a
            result set containing information about the user-defined types for the
            passed catalog, schema pattern, type name pattern, and types. *)

    val get_url : t -> string
        (** Returns the database URL. *)

    val get_user_name : t -> string
        (** Returns the user name. *)

    val get_version_columns : t -> string -> string -> string -> ResultSet.t
        (** [get_version_columns dbm catalog schema table] returns a result set
            containing information about the automatic updates for the passed
            catalog, schema, and table. *)

    val inserts_are_detected : t -> jdbc_type -> bool
        (** Tests whether a row addition caon be detected for the passed type. *)

    val is_catalog_at_start : t -> bool
        (** Tests whether the catalog is used as prefix for a fully-qualified table name. *)

    val is_read_only : t -> bool
        (** Tests whether the database is in read-only mode. *)

    val locators_update_copy : t -> bool
        (** Tests whether updates to {i x}LOBs are made on copies. *)

    val null_plus_non_null_is_null : t -> bool
        (** Tests whether {i null + null} equals {i null}. *)

    val nulls_are_sorted_at_end : t -> bool
        (** Tests whether null values are sorted at end. *)

    val nulls_are_sorted_at_start : t -> bool
        (** Tests whether null values are sorted at start. *)

    val nulls_are_sorted_high : t -> bool
        (** Tests whether null values are sorted high. *)

    val nulls_are_sorted_low : t -> bool
        (** Tests whether null values are sorted low. *)

    val others_deletes_are_visible : t -> jdbc_type -> bool
        (** Tests whether deletes made by others are visible for the passed type. *)

    val others_inserts_are_visible : t -> jdbc_type -> bool
        (** Tests whether inserts made by others are visible for the passed type. *)

    val others_updates_are_visible : t -> jdbc_type -> bool
        (** Tests whether updates made by others are visible for the passed type. *)

    val own_deletes_are_visible : t -> jdbc_type -> bool
        (** Tests whether deletes made by the same result set are visible for the passed type. *)

    val own_inserts_are_visible : t -> jdbc_type -> bool
        (** Tests whether inserts made by the same result set are visible for the passed type. *)

    val own_updates_are_visible : t -> jdbc_type -> bool
        (** Tests whether updates made by the same result set are visible for the passed type. *)

    val stores_lower_case_identifiers : t -> bool
        (** Tests whether the database stores identifiers using lower case form. *)

    val stores_lower_case_quoted_identifiers : t -> bool
        (** Tests whether the database stores quoted identifiers using lower case form. *)

    val stores_mixed_case_identifiers : t -> bool
        (** Tests whether the database stores identifiers using mixed case form. *)

    val stores_mixed_case_quoted_identifiers : t -> bool
        (** Tests whether the database stores quoted identifiers using mixed case form. *)

    val stores_upper_case_identifiers : t -> bool
        (** Tests whether the database stores identifiers using upper case form. *)

    val stores_upper_case_quoted_identifiers : t -> bool
        (** Tests whether the database stores quoted identifiers using upper case form. *)

    val supports_alter_table_with_add_column : t -> bool
        (** Tests whether the database supports an {i alter} statement adding columns. *)

    val supports_alter_table_with_drop_column : t -> bool
        (** Tests whether the database supports an {i alter} statement removing columns. *)

    val supports_ansi92_entry_level_sql : t -> bool
        (** Tests whether the database supports the entry level of ANSI 92. *)

    val supports_ansi92_full_sql : t -> bool
        (** Tests whether the database fully supports ANSI 92. *)

    val supports_ansi92_intermediate_sql : t -> bool
        (** Tests whether the database supports the intermediate level of ANSI 92. *)

    val supports_batch_updates : t -> bool
        (** Tests whether the database supports batch updates. *)

    val supports_catalogs_in_data_manipulation : t -> bool
        (** Tests whether the database supports catalogs in data manipulation. *)

    val supports_catalogs_in_index_definitions : t -> bool
        (** Tests whether the database supports catalogs in index definitions. *)

    val supports_catalogs_in_privilege_definitions : t -> bool
        (** Tests whether the database supports catalogs in privilege definitions. *)

    val supports_catalogs_in_procedure_calls : t -> bool
        (** Tests whether the database supports catalogs in procedure calls. *)

    val supports_catalogs_in_table_definitions : t -> bool
        (** Tests whether the database supports catalogs in table definitions. *)

    val supports_column_aliasing : t -> bool
        (** Tests whether the database supports column aliasing. *)

    val supports_convert : t -> bool
        (** Tests whether the database supports conversions. *)

    val supports_convert_types : t -> jdbc_type -> jdbc_type -> bool
        (** [supports_convert_types dbm from to] tests whether the database supports
            conversions from type [from] to type [to]. *)

    val supports_core_sql_grammar : t -> bool
        (** Tests whether the database supports the core SQL grammar. *)

    val supports_correlated_subqueries : t -> bool
        (** Tests whether the database supports correlated subqueries. *)

    val supports_data_definition_and_data_manipulation_transactions : t -> bool
        (** Tests whether the database supports both data definition and manipulation in transactions. *)

    val supports_data_manipulation_transactions_only : t -> bool
        (** Tests whether the database supports data manipulation only in transactions. *)

    val supports_different_table_correlation_names : t -> bool
        (** Tests whether the database supports different table correleation names. *)

    val supports_expressions_in_order_by : t -> bool
        (** Tests whether the database supports expressions in {i order by} constructs. *)

    val supports_extended_sql_grammar : t -> bool
        (** Tests whether the database supports the extended SQL grammar. *)

    val supports_full_outer_joins : t -> bool
        (** Tests whether the database supports full outer joins. *)

    val supports_get_generated_keys : t -> bool
        (** Tests whether the database supports the retrieval of auto-generated values. *)

    val supports_group_by : t -> bool
        (** Tests whether the database supports {i group by} constructs. *)

    val supports_group_by_beyond_select : t -> bool
        (** Tests whether the database supports {i group by} beyond {i select}
            (that is grouping using a column not in the {i select} clause,
            provided that all the columns in the {i select} clause also appear
            in the {i group by} clause). *)

    val supports_group_by_unrelated : t -> bool
        (** Tests whether the database supports unrelated {i group by}
            (that is grouping using a column not in the {i select} clause). *)

    val supports_integrity_enhancement_facility : t -> bool
        (** Tests whether the database supports integrity enhancement facilities. *)

    val supports_like_escape_clause : t -> bool
        (** Tests whether the database supports {i like} constructs. *)

    val supports_limited_outer_joins : t -> bool
        (** Tests whether the database supports limited outer joins. *)

    val supports_minimum_sql_grammar : t -> bool
        (** Tests whether the database supports the minimum SQL grammar. *)

    val supports_mixed_case_identifiers : t -> bool
        (** Tests whether the database supports mixed case identifiers. *)

    val supports_mixed_case_quoted_identifiers : t -> bool
        (** Tests whether the database supports mixed case quoted identifiers. *)

    val supports_multiple_open_results : t -> bool
        (** Tests whether the database supports multiple open result sets. *)

    val supports_multiple_result_sets : t -> bool
        (** Tests whether the database supports multiple result sets. *)

    val supports_multiple_transactions : t -> bool
        (** Tests whether the database supports multiple transactions. *)

    val supports_named_parameters : t -> bool
        (** Tests whether the database supports named parameters. *)

    val supports_non_nullable_columns : t -> bool
        (** Tests whether the database supports non nullable columns. *)

    val supports_open_cursors_across_commit : t -> bool
        (** Tests whether the database supports open cursors across commits. *)

    val supports_open_cursors_across_rollback : t -> bool
        (** Tests whether the database supports open cursors across rollbacks. *)

    val supports_open_statements_across_commit : t -> bool
        (** Tests whether the database supports open statements across commits. *)

    val supports_open_statements_across_rollback : t -> bool
        (** Tests whether the database supports open statements across rollbacks. *)

    val supports_order_by_unrelated : t -> bool
        (** Tests whether the database supports unrelared {i order by}
            (that is ordering using a column not in the {i select} clause). *)

    val supports_outer_joins : t -> bool
        (** Tests whether the database supports outer joins. *)

    val supports_positioned_delete : t -> bool
        (** Tests whether the database supports positioned deletes. *)

    val supports_positioned_update : t -> bool
        (** Tests whether the database supports positioned upodates. *)

    val supports_result_set_concurrency : t -> result_set_type -> result_set_concurrency -> bool
        (** Tests whether the database supports the conjunction of passed result
            set type and concurrency. *)

    val supports_result_set_holdability : t -> result_set_holdability -> bool
        (** Tests whether the database supports the passed holdability. *)

    val supports_result_set_type : t -> result_set_type -> bool
        (** Tests whether the database supports the passed result set type. *)

    val supports_savepoints : t -> bool
        (** Tests whether the database supports save points. *)

    val supports_schemas_in_data_manipulation : t -> bool
        (** Tests whether the database supports schemas in data manipulation. *)

    val supports_schemas_in_index_definitions : t -> bool
        (** Tests whether the database supports schemas in index definitions. *)

    val supports_schemas_in_privilege_definitions : t -> bool
        (** Tests whether the database supports schemas in privilege definitions. *)

    val supports_schemas_in_procedure_calls : t -> bool
        (** Tests whether the database supports schemas in procedure calls. *)

    val supports_schemas_in_table_definitions : t -> bool
        (** Tests whether the database supports schemas in table definitions. *)

    val supports_select_for_update : t -> bool
        (** Tests whether the database supports {i select for update} constructs. *)

    val supports_statement_pooling : t -> bool
        (** Tests whether the database supports statement pooling. *)

    val supports_stored_functions_using_call_syntax : t -> bool
        (** Tests whether the database supports invocation of stored functions using the call syntax. *)

    val supports_stored_procedures : t -> bool
        (** Tests whether the database supports stored procedures. *)

    val supports_subqueries_in_comparisons : t -> bool
        (** Tests whether the database supports subqueries in comparisons. *)

    val supports_subqueries_in_exists : t -> bool
        (** Tests whether the database supports subqueries in {i exists} expressions. *)

    val supports_subqueries_in_ins : t -> bool
        (** Tests whether the database supports subqueries in {i in} expressions. *)

    val supports_subqueries_in_quantifieds : t -> bool
        (** Tests whether the database supports subqueries in quantified expressions. *)

    val supports_table_correlation_names : t -> bool
        (** Tests whether the database supports table correlation names. *)

    val supports_transaction_isolation_level : t -> transaction_isolation -> bool
        (** Tests whether the database supports the passed transaction isolation level. *)

    val supports_transactions : t -> bool
        (** Tests whether the database supports transactions. *)

    val supports_union : t -> bool
        (** Tests whether the database supports the {i union} construct. *)

    val supports_union_all : t -> bool
        (** Tests whether the database supports the {i union all} construct. *)

    val updates_are_detected : t -> jdbc_type -> bool
        (** Tests whether a row modification can be detected for the passed type. *)

    val uses_local_file_per_table : t -> bool
        (** Tests whether the database uses a local file per table to store data. *)

    val uses_local_files : t -> bool
        (** Tests whether the database uses local files to store data. *)
  end
(** Metadata for databases. *)

module Connection :
  sig
    type t
        (** Type of connections. *)

    val clear_warnings : t -> unit
        (** Clears the list of warnings currently reported by the passed connection. *)

    val close : t -> unit
        (** Closes the passed connection. *)

    val commit : t -> unit
        (** Commits all the changes made since the previous rollback/commit to
            the database. *)

    val create_array_of : t -> string -> Cadmium.java_object list -> Array.t
        (** Creates an array value. The type of elements is given by the passed
            string while the passed list contains the elements that will be used
            to fill the array. *)

    val create_blob : t -> Blob.t
        (** Creates a blob value. *)

    val create_clob : t -> Clob.t
        (** Creates a clob value. *)

    val create_nclob : t -> NClob.t
        (** Creates a nclob value. *)

    val create_sqlxml : t -> SQLXML.t
        (** Creates a XML value. *)

    val create_statement : t -> Statement.t
        (** Creates a SQL statement. *)

    val create_statement_type_conc : t -> result_set_type -> result_set_concurrency -> Statement.t
        (** [create_statement_type_conc conn t c] creates a SQL statement with
            type [t] and concurrency [c]. *)

    val create_statement_type_conc_hold : t -> result_set_type -> result_set_concurrency -> result_set_holdability -> Statement.t
        (** [create_statement_type_conc_hold conn t c h] creates a SQL statement
            with type [t], concurrency [c], and holdability [h]. *)

    val create_struct : t -> string -> Cadmium.java_object list -> Struct.t
        (** Creates an structured value. The type of the value is given by the
            passed string while the passed list contains the attribute values
            that will be used to build the value. *)

    val get_auto_commit : t -> bool
        (** Tests whether auto-commit mode is enabled for the passed connection. *)

    val get_catalog : t -> string
        (** Returns the catalog name for the passed connection. *)

    val get_client_infos : t -> (string * string) list
        (** Returns the client info properties, as a (key, value) list. *)

    val get_client_info : t -> string -> string
        (** Returns the value of the client info property whose name is passed. *)

    val get_holdability : t -> result_set_holdability
        (** Returns the holdability for the passed connection. *)

    val get_meta_data : t -> DatabaseMetaData.t
        (** Returns the database metadata for the passed connection. *)

    val get_transaction_isolation : t -> transaction_isolation
        (** Returns the isolation for the passed connection. *)

    val get_warnings : t -> sql_warning list
        (** Returns the list of warnings currently reported by the passed connection. *)

    val is_closed : t -> bool
        (** Tests whether the passed connection is closed. *)

    val is_read_only : t -> bool
        (** Tests whether the passed connection is in read-only mode. *)

    val is_valid : t -> int32 -> bool
        (** Tests whether the passed connection is both still opened and valid,
            the passed integer being the maximum timeout. *)

    val native_sql : t -> string -> string
        (** Converts the passed SQL statement into the dialect used by the
            database related to the passed connection. *)

    val prepare_call : t -> string -> CallableStatement.t
        (** [prepare_call conn sql] returns a callable statement from the passed
            SQL statement [sql]. *)

    val prepare_call_type_conc : t -> string -> result_set_type -> result_set_concurrency -> CallableStatement.t
        (** [prepare_call_type_conc conn sql t c] returns a callable statement
            from the passed SQL statement [sql]. The result sets returned by the
            callable statement will have a type of [t] and a concurrency of [c]. *)

    val prepare_call_type_conc_hold : t -> string -> result_set_type -> result_set_concurrency -> result_set_holdability -> CallableStatement.t
        (** [prepare_call_type_conc_hold conn sql t c h] returns a callable
            statement from the passed SQL statement [sql]. The result sets
            returned by the callable statement will have a type of [t],
            a concurrency of [c], and a holdability of [h]. *)

    val prepare_statement : t -> string -> PreparedStatement.t
        (** [prepare_statement conn sql] returns a prepared statement from the
            passed SQL statement [sql]. *)

    val prepare_statement_autogen_keys : t -> string -> statement_autogen_keys -> PreparedStatement.t
        (** [prepare_statement_autogen_keys conn sql ag] returns a prepared
            statement from the passed SQL statement [sql], using autogen setting [ag]. *)

    val prepare_statement_indexes : t -> string -> int list -> PreparedStatement.t
        (** [prepare_statement_indexes conn sql indexes] returns a prepared statement
            from the passed SQL statement [sql]. The list [indexes] contains
            the column indexes indicating the columns that should be returned from
            the inserted row(s). *)

    val prepare_statement_type_conc : t -> string -> result_set_type -> result_set_concurrency -> PreparedStatement.t
        (** [prepare_statement_type_conc conn sql t c] returns a prepared statement
            from the passed SQL statement [sql]. The result sets returned by the
            built statement will have a type of [t] and a concurrency of [c]. *)

    val prepare_statement_type_conc_hold : t -> string -> result_set_type -> result_set_concurrency -> result_set_holdability -> PreparedStatement.t
        (** [prepare_statement_type_conc_hold conn sql t c h] returns a
            prepared statement from the passed SQL statement [sql]. The result sets
            returned by the built statement will have a type of [t],
            a concurrency of [c], and a holdability of [h]. *)

    val prepare_statement_names : t -> string -> string list -> PreparedStatement.t
        (** [prepare_statement_names conn sql names] returns a prepared statement
            from the passed SQL statement [sql]. The list [names] contains
            the column names indicating the columns that should be returned from
            the inserted row(s). *)

    val release_savepoint : t -> Savepoint.t -> unit
        (** Removes the passed save point, as well as all the subsequently
            defined ones. *)

    val rollback : t -> unit
        (** Rollbacks any change made since the last commit. *)

    val rollback_to : t -> Savepoint.t -> unit
        (** Rollbacks to the database state defined by the passed save point. *)

    val set_auto_commit : t -> bool -> unit
        (** Sets the auto-commit mode for the passed connection. *)

    val set_catalog : t -> string -> unit
        (** Sets the catalog name to the passed value. *)

    val set_client_infos : t -> (string * string) list -> unit
        (** Changes the client info properties. *)

    val set_client_info : t -> string -> string -> unit
        (** [set_client_info conn n v] changes the value of the client info
            property whose named [n] to the value [v]. *)

    val set_holdability : t -> result_set_holdability -> unit
        (** Sets the default holdabilty for result sets generated by the passed
            connection. *)

    val set_read_only : t -> bool -> unit
        (** Sets the read-only mode for the passed connection. *)

    val set_savepoint : t -> Savepoint.t
        (** Creates an unnamed savepoint. *)

    val set_savepoint_with_name : t -> string -> Savepoint.t
        (** Creates a savepoint with the passed name. *)

    val set_transaction_isolation : t -> transaction_isolation -> unit
        (** Sets the transation isolation level for the passed connection. *)
  end
(** Connections. *)

module Driver :
  sig
    type property_info = {
        choices : string list; (** Possible property values. *)
        description : string; (** Description of the property. *)
        name : string; (** Name of the property. *)
        required : bool; (** Whether the property must be set at connection. *)
        value : string; (** Current value of property. *)
      }
        (** Type of property information for drivers. *)

    type t
        (** Type of drivers. *)

    val accepts_url : t -> string -> bool
        (** [accepts_url d url] tests whether driver [d] can connect to URL [url]. *)

    val connect : t -> string -> (string * string) list -> Connection.t
        (** [connect d url props] attemps to establish a connection to the passed
            database URL [url], using [props] as a (key,value) list of
            properties (acting as connection arguments). *)

    val get_major_version : t -> int32
        (** Returns the major component of the driver's version. *)

    val get_minor_version : t -> int32
        (** Returns the minor component of the driver's version. *)

    val get_property_info : t -> string -> (string * string) list -> property_info list
        (** [get_property_info d url props] returns information about the
            properties understood by driver [d]. [props] is a (key,value) list
            of properties (acting as connection arguments) that would be sent
            when trying to connect to URL [url]. *)

    val is_jdbc_compliant : t -> bool
        (** Tests whether the passed driver is JDBC compliant. *)
  end
(** Drivers. *)

module DriverManager :
  sig
    val deregister_driver : Driver.t -> unit
        (** Removes the passed driver from the list of available ones. *)

    val get_connection : string -> Connection.t
        (** [get_connection url] attemps to establish a connection to the passed
            database URL. *)

    val get_connection_with_properties : string -> (string * string) list -> Connection.t
        (** [get_connection_with_properties url props] attemps to establish a
            connection to the passed database URL [url], using [props] as a
            (key,value) list of properties (acting as connection arguments). *)

    val get_connection_with_user_passwd : string -> string -> string -> Connection.t
        (** [get_connection_with_user_passwd url user passwd] attemps to
            establish a connection to the passed database URL [url] using login
            [user] and password [passwd]. *)

    val get_driver : string -> Driver.t
        (** [get_driver url] returns a driver able to connect to the URL [url]. *)

    val get_drivers : unit -> Driver.t list
        (** Returns the list of available drivers. *)

    val get_login_timeout : unit -> int32
        (** Returns the maximum timeout for database connection. *)

    val print_log : string -> unit
        (** Prints the passed string onto the JDBC log. *)

    val register_driver : Driver.t -> unit
        (** Adds the passed driver to the list of available drivers. *)

    val set_login_timeout : int32 -> unit
        (** Sets the maximum timeout for database connection to the passed value. *)
  end
(** Driver manager. *)

module Iterators :
  sig
    val iter : ('a -> unit) -> < next : 'a; .. > -> unit
        (** [iter f o] applies [f] to the values successively returned by
            [o#next] until this method call raises a [Not_found] exception. *)

    val iter_close : ('a -> unit) -> < close : unit; next : 'a; .. > -> unit
        (** Same as [iter] except that [close] is called on the passed object
            when the iteration is done. *)

    val iteri : (int -> 'a -> unit) -> < next : 'a; .. > -> unit
        (** Same as [iter] except that the function is applied to the index of
            the element (starting with 0) as first argument,
            and the element as second argument. *)

    val iteri_close : (int -> 'a -> unit) -> < close : unit; next : 'a; .. > -> unit
        (** Same as [iteri] except that [close] is called on the passed object
            when the iteration is done. *)

    val map : ('a -> 'b) -> < next : 'a; .. > -> 'b list
        (** [map f o] applies [f] to the values successively returned by
            [o#next] until this method call raises a [Not_found] exception.
            The results of these applications are used to build the returned
            list. *)

    val map_close : ('a -> 'b) -> < close : unit; next : 'a; .. > -> 'b list
        (** Same as [map] except that [close] is called on the passed object
            when the iteration is done. *)

    val mapi : (int -> 'a -> 'b) -> < next : 'a; .. > -> 'b list
        (** Same as [map] except that the function is applied to the index of
            the element (starting with 0) as first argument,
            and the element as second argument. *)

    val mapi_close : (int -> 'a -> 'b) -> < close : unit; next : 'a; .. > -> 'b list
        (** Same as [mapi] except that [close] is called on the passed object
            when the iteration is done. *)

    val fold : ('a -> 'b -> 'a) -> 'a -> < next : 'b; .. > -> 'a
        (** [fold f z o] is equivalent to [List.fold_left f z l] where l is the
            list built from the values successively returned by [o#next]. *)

    val fold_close : ('a -> 'b -> 'a) -> 'a -> < close : unit; next : 'b; .. > -> 'a
        (** Same as [fold] except that [close] is called on the passed object
            when the iteration is done. *)

    val foldi : (int -> 'a -> 'b -> 'a) -> 'a -> < next : 'b; .. > -> 'a
        (** Same as [fold] except that the function is applied to the index of
            the element (starting with 0) as first argument,
            the accumulator as second argument, and the element as third
            argument. *)

    val foldi_close : (int -> 'a -> 'b -> 'a) -> 'a -> < close : unit; next : 'b; .. > -> 'a
        (** Same as [foldi] except that [close] is called on the passed object
            when the iteration is done. *)

    val to_list : < next : 'a; .. > -> 'a list
        (** Builds the list of the values successively returned by [o#next]. *)

    val to_list_close : < close : unit; next : 'a; .. > -> 'a list
        (** Same as [to_list] except that [close] is called on the passed object
            when the iteration is done. *)
  end
(** Iterator functions to be used with Nickel-generated database modules. *)
