(*
 * This file is part of Cadmium.
 * Copyright (C) 2007-2010 Xavier Clerc.
 *
 * Cadmium is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * Cadmium is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *)

(** Access to Java math routines and arbitrary-precision numbers. *)


module Float :
  sig
    val e : float
        (** The value of {i e}. *)

    val pi  : float
        (** The value of {i pi}. *)

    val abs : float -> float
        (** [abs x] returns the absolute value of [x]. *)

    val acos : float -> float
        (** [acos x] returns the arc cosine value of [x].
            Returned value is in {i 0..pi}. *)

    val asin : float -> float
        (** [asin x] returns the arc sine value of [x].
            Returned value is in {i -pi/2..pi/2}. *)

    val atan : float -> float
        (** [atan x] returns the arc tangent value of [x].
            Returned value is in {i -pi/2..pi/2}. *)

    val atan2 : float -> float -> float
        (** [atan2 x y] returns the value of the angle resulting from the
            conversion from rectangular coordinates {i (x, y)} into polar
            coordinates. *)

    val cbrt : float -> float
        (** [cbrt x] returns the cube root of [x]. *)

    val ceil : float -> float
        (** [ceil x] returns the smallest integer value that is greater than or
            equal to [x]. *)

    val copy_sign : float -> float -> float
        (** [copy_sign x y] returns the value [x], with the sign of [y]. *)

    val cos : float -> float
        (** [cos x] returns the cosine value of [x]. *)

    val cosh : float -> float
        (** [cosh x] returns the hyperbolic cosine value of [x]. *)

    val exp : float -> float
        (** [exp x] returns exponential value of [x]. *)

    val exp_m1 : float -> float
        (** [exp_m1 x] returns [(exp x) -. 1.0]. *)

    val floor : float -> float
        (** [floor x] returns the largest integer value that is less than or
            equal to [x]. *)

    val get_exponent : float -> int
        (** [get_exponent x] returns the exponent of [x]. *)

    val hypot : float -> float -> float
        (** [hypot x y] returns [sqrt ((x *. x) +. (y *. y))]. *)

    val ieee_remainder : float -> float -> float
        (** [ieee_remainder x y] returns the remainder of [x] and [y]. *)

    val log : float -> float
        (** [log x] returns the (base {i e}) logarithm of [x]. *)

    val log10 : float -> float
        (** [log10 x] returns the (base 10) logarithm of [x]. *)

    val log1p : float -> float
        (** [log1p x] returns [(log x) +. 1.0]. *)

    val max : float -> float -> float
        (** [max x y] returns the maximum of [x] and [y]. *)

    val min : float -> float -> float
        (** [min x y] returns the minimum of [x] and [y]. *)

    val next_after : float -> float -> float
        (** [next_after x y] returns value that is the closest to [x] in the direction of [y]. *)

    val next_up : float -> float
        (** [next_up x] returns value that is the closest to [x] in the direction of positive infinity. *)

    val pow : float -> float -> float
        (** [pow x y] returns [x ** y]. *)

    val random : unit -> float
        (** [random x] returns a random value betwwen 0.0 (inclusive) and 1.0 (exclusive). *)

    val rint : float -> float
        (** [rint x] returns the integer value that is the closest to [x]. *)

    val round : float -> int64
        (** [round x] returns the int64 value that is the closest to [x]. *)

    val scalb : float -> int -> float
        (** [scalb x n] returns [x *. (2.0 ** (float_of_int n))]. *)

    val signum : float -> float
        (** [signum x] returns the signum of [x], that is:
            - 0.0 if [x] is zero;
            - 1.0 if [x] is positive;
            - and -1.0 if [x] is negative *)

    val sin : float -> float
        (** [sin x] returns the sine value of [x]. *)

    val sinh : float -> float
        (** [sinh x] returns the hyperbolic sine value of [x]. *)

    val sqrt : float -> float
        (** [sqrt x] returns the square root of [x]. *)

    val tan : float -> float
        (** [tan x] returns the tangent value of [x]. *)

    val tanh : float -> float
        (** [tanh x] returns the hyperbolic tangent value of [x]. *)

    val to_degrees : float -> float
        (** [to_degrees x] converts the angle [x] measured in radians into
            an equivalent angle measured in degrees. *)

    val to_radians : float -> float
        (** [to_radians x] converts the angle [x] measured in degrees into
            an equivalent angle measured in radians. *)

    val ulp : float -> float
        (** [ulp x] returns size of an ulp for [x]. *)
  end
(** Float routines from {i java.lang.Math}. *)

module StrictFloat :
  sig
    val e : float
        (** The value of {i e}. *)

    val pi  : float
        (** The value of {i pi}. *)

    val abs : float -> float
        (** [abs x] returns the absolute value of [x]. *)

    val acos : float -> float
        (** [acos x] returns the arc cosine value of [x].
            Returned value is in {i 0..pi}. *)

    val asin : float -> float
        (** [asin x] returns the arc sine value of [x].
            Returned value is in {i -pi/2..pi/2}. *)

    val atan : float -> float
        (** [atan x] returns the arc tangent value of [x].
            Returned value is in {i -pi/2..pi/2}. *)

    val atan2 : float -> float -> float
        (** [atan2 x y] returns the value of the angle resulting from the
            conversion from rectangular coordinates {i (x, y)} into polar
            coordinates. *)

    val cbrt : float -> float
        (** [cbrt x] returns the cube root of [x]. *)

    val ceil : float -> float
        (** [ceil x] returns the smallest integer value that is greater than or
            equal to [x]. *)

    val copy_sign : float -> float -> float
        (** [copy_sign x y] returns the value [x], with the sign of [y]. *)

    val cos : float -> float
        (** [cos x] returns the cosine value of [x]. *)

    val cosh : float -> float
        (** [cosh x] returns the hyperbolic cosine value of [x]. *)

    val exp : float -> float
        (** [exp x] returns exponential value of [x]. *)

    val exp_m1 : float -> float
        (** [exp_m1 x] returns [(exp x) -. 1.0]. *)

    val floor : float -> float
        (** [floor x] returns the largest integer value that is less than or
            equal to [x]. *)

    val get_exponent : float -> int
        (** [get_exponent x] returns the exponent of [x]. *)

    val hypot : float -> float -> float
        (** [hypot x y] returns [sqrt ((x *. x) +. (y *. y))]. *)

    val ieee_remainder : float -> float -> float
        (** [ieee_remainder x y] returns the remainder of [x] and [y]. *)

    val log : float -> float
        (** [log x] returns the (base {i e}) logarithm of [x]. *)

    val log10 : float -> float
        (** [log10 x] returns the (base 10) logarithm of [x]. *)

    val log1p : float -> float
        (** [log1p x] returns [(log x) +. 1.0]. *)

    val max : float -> float -> float
        (** [max x y] returns the maximum of [x] and [y]. *)

    val min : float -> float -> float
        (** [min x y] returns the minimum of [x] and [y]. *)

    val next_after : float -> float -> float
        (** [next_after x y] returns value that is the closest to [x] in the direction of [y]. *)

    val next_up : float -> float
        (** [next_up x] returns value that is the closest to [x] in the direction of positive infinity. *)

    val pow : float -> float -> float
        (** [pow x y] returns [x ** y]. *)

    val random : unit -> float
        (** [random x] returns a random value betwwen 0.0 (inclusive) and 1.0 (exclusive). *)

    val rint : float -> float
        (** [rint x] returns the integer value that is the closest to [x]. *)

    val round : float -> int64
        (** [round x] returns the int64 value that is the closest to [x]. *)

    val scalb : float -> int -> float
        (** [scalb x n] returns [x *. (2.0 ** (float_of_int n))]. *)

    val signum : float -> float
        (** [signum x] returns the signum of [x], that is:
            - 0.0 if [x] is zero;
            - 1.0 if [x] is positive;
            - and -1.0 if [x] is negative *)

    val sin : float -> float
        (** [sin x] returns the sine value of [x]. *)

    val sinh : float -> float
        (** [sinh x] returns the hyperbolic sine value of [x]. *)

    val sqrt : float -> float
        (** [sqrt x] returns the square root of [x]. *)

    val tan : float -> float
        (** [tan x] returns the tangent value of [x]. *)

    val tanh : float -> float
        (** [tanh x] returns the hyperbolic tangent value of [x]. *)

    val to_degrees : float -> float
        (** [to_degrees x] converts the angle [x] measured in radians into
            an equivalent angle measured in degrees. *)

    val to_radians : float -> float
        (** [to_radians x] converts the angle [x] measured in degrees into
            an equivalent angle measured in radians. *)

    val ulp : float -> float
        (** [ulp x] returns size of an ulp for [x]. *)
  end
(** Float routines from {i java.lang.StrictMath}. *)

module Random :
  sig
    type t
          (** Type of random number generators. *)

    val make : unit -> t
        (** Constructs a random number generator. *)

    val make_of_int64 : int64 -> t
        (** Constructs a random number generator, using the passed value as the seed. *)

    val next_boolean : t -> bool
        (** [next_boolean r] returns the next boolean generated by [r]. *)

    val next_bytes : t -> string -> unit
        (** [next_bytes r b] populates the byte array [b] with values generated by [r]. *)

    val next_float : t -> float
        (** [next_float r] returns the next float generated by [r].
            The returned value is between 0.0 (inclusive) and 1.0 (exclusive). *)

    val next_gaussian : t -> float
        (** [next_gaussian r] returns the next {i Gaussian} float generated by [r].
            The distribution has a mean of 0.0 and a standard deviation of 1.0. *)

    val next_int : t -> int
        (** [next_int r] returns the next int value generated by [r]. *)

    val next_int_n : t -> int -> int
        (** [next_int_n r n] returns the next int value generated by [r].
            The returned value is between 0 (inclusive) and [n] (exclusive). *)

    val next_int32 : t -> int32
        (** [next_int32 r] returns the next int32 value generated by [r]. *)

    val next_int32_n : t -> int32 -> int32
        (** [next_int32_n r n] returns the next int value generated by [r].
            The returned value is between 0 (inclusive) and [n] (exclusive). *)

    val next_int64 : t -> int64
        (** [next_int64 r] returns the next int64 value generated by [r]. *)

    val set_seed : t -> int64 -> unit
        (** [set_seed r s] sets the seed of random number generator [r] to [s]. *)
  end
(** Random number generators ({i java.util.Random} class). *)

module BigInteger :
  sig
    type t
          (** Type of arbitrary-precision integers. *)

    val one : t
        (** The value {i one} as a big integer. *)

    val ten : t
        (** The value {i ten} as a big integer. *)

    val zero : t
        (** The value {i zero} as a big integer. *)

    val make_of_bytes : string -> t
        (** Constructs a big integer, using the passed byte array as a
            big-endian two-complement representation. *)

    val make_of_signum_bytes : int -> string -> t
        (** Constructs a big integer, using the passed byte array as a
            big-endian representation of the absolute value, and using the
            passed int as the sign (-1 for negative, 0 for zero, and 1 for positive). *)

    val make_prime_of_random : int -> int -> Random.t -> t
        (** Constructs a big integer using the passed random number generator.
            The big integer value is between [0] (inclusive) and [2**n] (exclusive)
            where [n] is the first passed int. The value is a prime with a probability
            greater than [1 - 0.5 ** cert] where [cert] is the second passed int. *)

    val make_of_random : int -> Random.t -> t
        (** Constructs a big integer using the passed random number generator.
            The big integer value is between [0] (inclusive) and [2**n] (exclusive)
            where [n] is the passed int. *)

    val make_of_string : string -> t
        (** Constructs a big integer from the passed string. *)

    val make_of_string_radix : string -> int -> t
        (** Constructs a big integer from the passed string, using the passed
            int as the radix of the string representation. *)

    val make_of_int64 : int64 -> t
        (** Constructs a big integer from an int64. *)

    val abs : t -> t
        (** [abs x] returns the absolute value of [x]. *)

    val add : t -> t -> t
        (** [add x y] returns the addition of [x] and [y]. *)

    val and_ : t -> t -> t
        (** [and_ x y] returns the logical and of [x] and [y]. *)

    val and_not : t -> t -> t
        (** [and_not x y] returns [and_ x (not  y)]. *)

    val bit_count : t -> int
        (** [bit_count x] returns the number of bits in the two-complement
            representation of [x]. *)

    val bit_length : t -> int
        (** [bit_length x] returns the number of bits in the two-complement
            representation of [x], excluding sign bit. *)

    val clear_bit : t -> int -> t
        (** [clear_bit x n] return a value equal to [x] with the [n]th bit cleared. *)

    val compare_to : t -> t -> int
        (** [compare_to x y] compares [x] and [y] returning:
            - -1 if [x] is less than [y];
            - 0 if [x] is equal to [y];
            - 1 if [x] is greater than [y]. *)

    val divide : t -> t -> t
        (** [divide x y] returns the quotient of [x] and [y]. *)

    val divide_and_remainder : t -> t -> (t * t)
        (** [divide_and_remainder x y] returns [(q, r)] where:
            - [q] is the quotient of [x] and [y];
            - [r] is the remainder of [x] and [y]. *)

    val equals : t -> t -> bool
        (** Tests whether the passed values are equal. *)

    val flip_bit : t -> int -> t
        (** [flip_bit x n] return a value equal to [x] with the [n]th bit flipped. *)

    val float_value : t -> float
        (** Converts the passed value into a float. *)

    val gcd : t -> t -> t
        (** [gcd x y] returns the greatest common divisor of [x] and [y]. *)

    val get_lowest_set_bit : t -> int
        (** [get_lowest_set_bit x] returns the index of the lowest set bit of [x]. *)

    val hash_code : t -> int32
        (** [hash_code x] returns the hash value of [x]. *)

    val int_value : t -> int
        (** Converts the passed value into an int. *)

    val int32_value : t -> int32
        (** Converts the passed value into an int32. *)

    val int64_value : t -> int64
        (** Converts the passed value into an int64. *)

    val is_probably_prime : t -> int -> bool
        (** [is_probably_prime x cert] tests if [x] is prime. Returns:
            - [false] if [x] is definitely composite;
            - [true] if [x] is prime (with probability greater than [1 - 0.5**cert]). *)

    val max : t -> t -> t
        (** [max x y] returns the maximum of [x] and [y]. *)

    val min : t -> t -> t
        (** [min x y] returns the minimum of [x] and [y]. *)

    val mod_ : t -> t -> t
        (** [mod_ x y] returns the modulo of [x] and [y]. *)

    val mod_inverse : t -> t -> t
        (** [mod_inverse x y] returns [(x ** -1) mod y]. *)

    val mod_pow : t -> t -> t -> t
        (** [mod_pow x y z] returns [(x ** y) mod z]. *)

    val multiply : t -> t -> t
        (** [multiply x y] returns the multiplication of [x] and [y]. *)

    val negate : t -> t
        (** [negate x] returns the opposite of [x]. *)

    val next_probable_prime : t -> t
        (** [next_probable_prime x] returns the first integer greater then [x] that is
            probably prime. *)

    val not : t -> t
        (** [not x] returns the negation of [x]. *)

    val or_ : t -> t -> t
        (** [or_ x y] returns the logical or of [x] and [y]. *)

    val pow : t -> int -> t
        (** [pow x y] returns [x] raised to the power of [y]. *)

    val remainder : t -> t -> t
        (** [remainder x y] returns the remainder of [x] and [y]. *)

    val set_bit : t -> int -> t
        (** [set_bit x n] return a value equal to [x] with the [n]th bit set. *)

    val shift_left : t -> int -> t
        (** [shift_left x n] returns [floor (x * 2**n)]. *)

    val shift_right : t -> int -> t
        (** [shift_right x n] returns [floor (x / 2**n)]. *)

    val signum : t -> int
        (** [signum x] returns the signum of [x], that is:
            - -1 if [x] is negative;
            - 0 if [x] is zero;
            - 1 if [x] is positive. *)

    val subtract : t -> t -> t
        (** [subtract x y] returns the subtraction of [x] and [y]. *)

    val test_bit : t -> int -> bool
        (** [test_bit x n] tests whether [x] has its [n]th bit set. *)

    val to_bytes : t -> string
        (** Converts the passed value into a byte array
            (using two-complement representation). *)

    val to_string : t -> string
        (** Converts the passed value into a string. *)

    val to_string_radix : t -> int -> string
        (** Converts the passed value into a string, using the passed int as radix. *)

    val xor : t -> t -> t
        (** [xor x y] returns the logical exclusive or of [x] and [y]. *)
  end
(** Arbitrary-precision integers ({i java.math.BigInteger} class). *)

module BigDecimal :
  sig
    type rounding_mode =
      | CEILING (** Round towards positive infinity. *)
      | DOWN (** Round towards zero. *)
      | FLOOR (** Round towards negative infinity. *)
      | HALF_DOWN (** Round towards "nearest neighbour", towards down if equidistant. *)
      | HALF_EVEN (** Round towards "nearest neighbour", towards even value if equidistant. *)
      | HALF_UP (** Round towards "nearest neighbour", towards up if equidistant. *)
      | UNNECESSARY (** No rounding, assert that the the operation has exact result. *)
      | UP (** Round away from zero. *)
    (** Rouding modes for decimal operations. *)

    type math_context =
      | Decimal32 (** Precision of 7 digits, and rouding mode half even. *)
      | Decimal64 (** Precision of 16 digits, and rouding mode half even. *)
      | Decimal128 (** Precision of 34 digits, and rouding mode half even. *)
      | Unlimited (** Unlimited precision. *)
      | Custom of int * rounding_mode (** Custom context, parameters being precision in bits, and rounding mode. *)
    (** Math contexts for decimal operations. *)

    type t
          (** Type of arbitrary-precision decimals. *)

    val one : t
        (** The value {i one} as a big decimal. *)

    val ten : t
        (** The value {i ten} as a big decimal. *)

    val zero : t
        (** The value {i zero} as a big decimal. *)

    val make_of_big_integer : BigInteger.t -> t
        (** Constructs a big decimal from a big integer. *)

    val make_of_big_integer_ctxt : math_context -> BigInteger.t -> t
        (** Constructs a big decimal from a big integer,
            using the passed math context. *)

    val make_of_big_integer_scale : BigInteger.t -> int -> t
        (** Constructs a big decimal from a big integer,
            using the passed int as scale. *)

    val make_of_big_integer_scale_ctxt : math_context -> BigInteger.t -> int -> t
        (** Constructs a big decimal from a big integer,
            using the passed int as scale and the passed math context. *)

    val make_of_float : float -> t
        (** Constructs a big decimal from a float. *)

    val make_of_float_ctxt : math_context -> float -> t
        (** Constructs a big decimal from a float, using the passed math context. *)

    val make_of_int32 : int32 -> t
        (** Constructs a big decimal from an int32. *)

    val make_of_int32_ctxt : math_context -> int32 -> t
        (** Constructs a big decimal from an int32, using the passed math context. *)

    val make_of_int64 : int64 -> t
        (** Constructs a big decimal from an int64. *)

    val make_of_int64_ctxt : math_context -> int64 -> t
        (** Constructs a big decimal from an int64, using the passed math context. *)

    val make_of_string : string -> t
        (** Constructs a big decimal from the passed string. *)

    val make_of_string_ctxt : math_context -> string -> t
        (** Constructs a big decimal from the passed string, using the passed math context. *)

    val abs : t -> t
        (** [abs x] returns the absolute value of [x]. *)

    val abs_ctxt : math_context -> t -> t
        (** [abs_ctxt mc x] returns the absolute value of [x],
            using math context [mc]. *)

    val add : t -> t -> t
        (** [add x y] returns the addition of [x] and [y]. *)

    val add_ctxt : math_context -> t -> t -> t
        (** [add_ctxt mc x y] returns the addition of [x] and [y],
            using math context [mc]. *)

    val byte_value_exact : t -> int
        (** Converts the passed value into a byte, checking for lost information. *)

    val compare_to : t -> t -> int
        (** [compare_to x y] compares [x] and [y] returning:
            - -1 if [x] is less than [y];
            - 0 if [x] is equal to [y];
            - 1 if [x] is greater than [y]. *)

    val divide : t -> t -> t
        (** [divide x y] returns the quotient of [x] and [y]. *)

    val divide_scale_mode : int -> rounding_mode -> t -> t -> t
        (** [divide_scale_mode s rm x y] returns the quotient of [x] and [y] with scale [s],
            using rounding mode [rm]. *)

    val divide_ctxt : math_context -> t -> t -> t
        (** [divide_ctxt mc x y] returns the quotient of [x] and [y],
            using math context [mc]. *)

    val divide_mode : rounding_mode -> t -> t -> t
        (** [divide_mode rm x y] returns the quotient of [x] and [y],
            using rounding mode [rm]. *)

    val divide_and_remainder : t -> t -> (t * t)
        (** [divide_and_remainder x y] returns [(q, r)] where:
            - [q] is the quotient of [x] and [y];
            - [r] is the remainder of [x] and [y]. *)

    val divide_and_remainder_ctxt : math_context -> t -> t -> (t * t)
        (** [divide_and_remainder_ctxt mc x y] returns [(q, r)] where:
            - [q] is the quotient of [x] and [y];
            - [r] is the remainder of [x] and [y].
            The math context [mc] is used. *)

    val divide_to_integral_value : t -> t -> t
        (** [divide_to_integral_value x y] returns the integer division of [x] and [y]. *)

    val divide_to_integral_value_ctxt : math_context -> t -> t -> t
        (** [divide_to_integral_value_ctxt mc x y] returns the integer division of [x] and [y],
            using math context [mc]. *)

    val equals : t -> t -> bool
        (** Tests whether the passed values are equal. *)

    val float_value : t -> float
        (** Converts the passed value into a float. *)

    val hash_code : t -> int32
        (** [hash_code x] returns the hash value of [x]. *)

    val int_value : t -> int
        (** Converts the passed value into an int. *)

    val int_value_exact : t -> int
        (** Converts the passed value into an int, checking for lost information. *)

    val int32_value : t -> int32
        (** Converts the passed value into an int32. *)

    val int32_value_exact : t -> int32
        (** Converts the passed value into an int32, checking for lost information. *)

    val int64_value : t -> int64
        (** Converts the passed value into an int64. *)

    val int64_value_exact : t -> int64
        (** Converts the passed value into an int64, checking for lost information. *)

    val max : t -> t -> t
        (** [max x y] returns the maximum of [x] and [y]. *)

    val min : t -> t -> t
        (** [min x y] returns the minimum of [x] and [y]. *)

    val move_point_left : t -> int -> t
        (** [move_point_left x n] moves the decimal point [n] places to the left. *)

    val move_point_right : t -> int -> t
        (** [move_point_right x n] moves the decimal point [n] places to the right. *)

    val multiply : t -> t -> t
        (** [multiply x y] returns the multiplication of [x] and [y]. *)

    val multiply_ctxt : math_context -> t -> t -> t
        (** [multiply_ctxt mc x y] returns the multiplication of [x] and [y],
            using math context [mc]. *)

    val negate : t -> t
        (** [negate x] returns the opposite of [x]. *)

    val negate_ctxt : math_context -> t -> t
        (** [negate_ctxt mc x] returns the opposite of [x],
            using math context [mc]. *)

    val plus : t -> t
        (** [plus x] returns [+x]. *)

    val plus_ctxt : math_context -> t -> t
        (** [plus_ctxt mc x] returns [+x], using math context [mc]. *)

    val pow : t -> int -> t
        (** [pow x n] returns [x] raised to the power of [n]. *)

    val pow_ctxt : math_context -> t -> int -> t
        (** [pow_ctxt mc x n] returns [x] raised to the power of [n],
            using math context [mc]. *)

    val precision : t -> int
        (** [precision x] returns the precision of [x]. *)

    val remainder : t -> t -> t
        (** [remainder x y] returns the remainder of [x] and [y]. *)

    val remainder_ctxt : math_context -> t -> t -> t
        (** [remainder_ctxt mc x y] returns the remainder of [x] and [y],
            using math context [mc]. *)

    val round : math_context -> t -> t
        (** [round mc x] returns [plus_ctxt mc x]. *)

    val scale : t -> int
        (** [scale x] returns the scale of [x]. *)

    val scale_by_power_of_ten : t -> int -> t
        (** [scale_by_power_of_ten x n] returns [x * (10 ** n)]. *)

    val set_scale : t -> int -> t
        (** [set_scale x s] returns a value equal to [x] with scale [s]. *)

    val set_scale_mode : rounding_mode -> t -> int -> t
        (** [set_scale_mode rm x s] returns a value equal to [x] with scale [s],
            using rounding mode [rm]. *)

    val short_value_exact : t -> int
        (** Converts the passed value into a short, checking for lost information. *)

    val signum : t -> int
        (** [signum x] returns the signum of [x], that is:
            - -1 if [x] is negative;
            - 0 if [x] is zero;
            - 1 if [x] is positive. *)

    val strip_trailing_zeros : t -> t
        (** [strip_trailing_zeros x] returns a value equal to [x], with no
            trailing zero. *)

    val subtract : t -> t -> t
        (** [subtract x y] returns the subtraction of [x] and [y]. *)

    val subtract_ctxt : math_context -> t -> t -> t
        (** [subtract_ctxt mc x y] returns the subtraction of [x] and [y],
            using math context [mc]. *)

    val to_big_integer : t -> BigInteger.t
        (** Converts the passed value into a big integer. *)

    val to_big_integer_exact : t -> BigInteger.t
        (** Converts the passed value into a big integer, checking for lost information. *)

    val to_engineering_string : t -> string
        (** Converts the passed value into a string (using engineering notation). *)

    val to_plain_string : t -> string
        (** Converts the passed value into a string (without exponent field). *)

    val to_string : t -> string
        (** Converts the passed value into a string. *)

    val ulp : t -> t
        (** [ulp x] returns size of an ulp for [x]. *)

    val unscaled_value : t -> BigInteger.t
        (** [unscaled_value x] returns the unscaled value of [x]. *)
  end
(** Arbitrary-precision decimals ({i java.math.BigDecimal} class). *)
