(*
 * This file is part of Cadmium.
 * Copyright (C) 2007-2010 Xavier Clerc.
 *
 * Cadmium is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * Cadmium is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *)

(* Types *)

type 'c java_type = 'c Cadmium.java_type =
  | Boolean
  | Byte
  | Char
  | Double
  | Float
  | Int
  | Long
  | Short
  | String
  | Class of 'c
  | Void
  | Array of 'c java_type

type 'o java_value = 'o Cadmium.java_value =
  | Boolean_val of bool
  | Byte_val of int
  | Char_val of int
  | Double_val of float
  | Float_val of float
  | Int_val of int32
  | Long_val of int64
  | Short_val of int
  | String_val of string
  | Object_val of 'o

type java_modifier = Cadmium.java_modifier =
  | Abstract
  | Final
  | Interface
  | Native
  | Private
  | Protected
  | Public
  | Static
  | StrictFP
  | Synchronized
  | Transcient
  | Volatile


(* Base functions *)

let is_available = Cadmium.is_available

let java_char_of_char = Cadmium.java_char_of_char

let char_of_java_char = Cadmium.char_of_java_char


(* Classes *)

class jObject p =
  let this = match p with
  | `Void -> Cadmium.Object.make ()
  | `Null -> Cadmium.get_null ()
  | `String s -> Cadmium.java_string_of_string s
  | `Cd'initObj o -> o#cd'this
  | `Cd'init t -> t in
  object
    method cd'this = this
    method cd'equal (o : jObject) = Cadmium.equal this o#cd'this
    method cd'notequal (o : jObject) = Cadmium.not_equal this o#cd'this
    method cd'instanceof (c : jClass) = Cadmium.instance_of this c#cd'this
    method cd'cast (c : jClass) = new jObject (`Cd'init (Cadmium.cast c#cd'this this))
    method cd'isnull = Cadmium.is_null this
    method clone = new jObject (`Cd'init (Cadmium.Object.clone this))
    method equals (obj : jObject) = Cadmium.Object.equals this obj#cd'this
    method getClass = new jClass (`Cd'init (Cadmium.Object.get_class this))
    method hashCode = Cadmium.Object.hash_code this
    method notify = Cadmium.Object.notify this
    method notifyAll = Cadmium.Object.notify_all this
    method toString = Cadmium.Object.to_string this
    method wait = Cadmium.Object.wait this
    method wait'1 t = Cadmium.Object.wait_time this t
    method wait'2 t n = Cadmium.Object.wait_time' this t n
  end
and jClass p =
  let rec conv_type = function
  | Boolean -> Cadmium.Boolean
  | Byte -> Cadmium.Byte
  | Char -> Cadmium.Char
  | Double -> Cadmium.Double
  | Float -> Cadmium.Float
  | Int -> Cadmium.Int
  | Long -> Cadmium.Long
  | Short -> Cadmium.Short
  | String -> Cadmium.String
  | Class (c : jClass) -> Cadmium.Class c#cd'this
  | Void -> Cadmium.Void
  | Array t -> Cadmium.Array (conv_type t) in
  let conv_types = List.map conv_type in
  let this = match p with
  | `For_name s -> Cadmium.Class.for_name s
  | `Cd'initObj o -> o#cd'this
  | `Cd'init t  -> t in
  object
    inherit (jObject (`Cd'init this))
    method getCanonicalName = Cadmium.Class.get_canonical_name this
    method getConstructor t = new jConstructor (`Cd'init (Cadmium.Class.get_constructor this (conv_types t)))
    method getConstructors = List.map (fun x -> new jConstructor (`Cd'init x)) (Cadmium.Class.get_constructors this)
    method getDeclaredConstructor t = new jConstructor (`Cd'init (Cadmium.Class.get_declared_constructor this (conv_types t)))
    method getDeclaredConstructors = List.map (fun x -> new jConstructor (`Cd'init x)) (Cadmium.Class.get_declared_constructors this)
    method getField s = new jField (`Cd'init (Cadmium.Class.get_field this s))
    method getFields = List.map (fun x -> new jField (`Cd'init x)) (Cadmium.Class.get_fields this)
    method getDeclaredField s = new jField (`Cd'init (Cadmium.Class.get_declared_field this s))
    method getDeclaredFields = List.map (fun x -> new jField (`Cd'init x)) (Cadmium.Class.get_declared_fields this)
    method getMethod s t = new jMethod (`Cd'init (Cadmium.Class.get_method this s (conv_types t)))
    method getMethods = List.map (fun x -> new jMethod (`Cd'init x)) (Cadmium.Class.get_methods this)
    method getDeclaredMethod s t = new jMethod (`Cd'init (Cadmium.Class.get_declared_method this s (conv_types t)))
    method getDeclaredMethods = List.map (fun x -> new jMethod (`Cd'init x)) (Cadmium.Class.get_declared_methods this)
    method getInterfaces = List.map (fun x -> new jClass (`Cd'init x)) (Cadmium.Class.get_interfaces this)
    method getModifiers = Cadmium.Class.get_modifiers this
    method getName = Cadmium.Class.get_name this
    method getPackage = Cadmium.Class.get_package this
    method getSimpleName = Cadmium.Class.get_simple_name this
    method getSuperclass = match Cadmium.Class.get_superclass this with
    | Some c -> Some (new jClass (`Cd'init c))
    | None -> None
    method isArray = Cadmium.Class.is_array this
    method isInterface = Cadmium.Class.is_interface this
    method isPrimitive = Cadmium.Class.is_primitive this
    method newInstance = new jObject (`Cd'init (Cadmium.Class.new_instance this))
  end
and jConstructor p =
  let rec conv_type = function
    | Boolean -> Cadmium.Boolean
    | Byte -> Cadmium.Byte
    | Char -> Cadmium.Char
    | Double -> Cadmium.Double
    | Float -> Cadmium.Float
    | Int -> Cadmium.Int
    | Long -> Cadmium.Long
    | Short -> Cadmium.Short
    | String -> Cadmium.String
    | Class (c : Cadmium.java_class) -> Cadmium.Class (new jClass (`Cd'init c))
    | Void -> Cadmium.Void
    | Array t -> Cadmium.Array (conv_type t) in
  let conv_types = List.map conv_type in
  let conv_value = function
    | Boolean_val b -> Cadmium.Boolean_val b
    | Byte_val i -> Cadmium.Byte_val i
    | Char_val i -> Cadmium.Char_val i
    | Double_val f -> Cadmium.Double_val f
    | Float_val f -> Cadmium.Float_val f
    | Int_val i -> Cadmium.Int_val i
    | Long_val i -> Cadmium.Long_val i
    | Short_val i -> Cadmium.Short_val i
    | String_val s -> Cadmium.String_val s
    | Object_val (o : jObject)  -> Cadmium.Object_val o#cd'this in
  let conv_values = List.map conv_value in
  let this = match p with
  | `Cd'initObj o -> o#cd'this
  | `Cd'init t -> t in
  object
    inherit (jObject (`Cd'init this))
    method getDeclaringClass = new jClass (`Cd'init (Cadmium.Constructor.get_declaring_class this))
    method getExceptionTypes = List.map (fun x -> new jClass (`Cd'init x)) (Cadmium.Constructor.get_exception_types this)
    method getModifiers = Cadmium.Constructor.get_modifiers this
    method getParameterTypes = conv_types (Cadmium.Constructor.get_parameter_types this)
    method newInstance p = new jObject (`Cd'init (Cadmium.Constructor.new_instance this (conv_values p)))
  end
and jField p =
  let this = match p with
  | `Cd'initObj o -> o#cd'this
  | `Cd'init t -> t in
  object
    inherit (jObject (`Cd'init this))
    method getDeclaringClass = new jClass (`Cd'init (Cadmium.Field.get_declaring_class this))
    method getModifiers = Cadmium.Field.get_modifiers this
    method getName = Cadmium.Field.get_name this
    method get (o : jObject) = new jObject (`Cd'init (Cadmium.Field.get this o#cd'this))
    method getBoolean (o : jObject) = Cadmium.Field.get_boolean this o#cd'this
    method getByte (o : jObject) = Cadmium.Field.get_byte this o#cd'this
    method getChar (o : jObject) = Cadmium.Field.get_char this o#cd'this
    method getDouble (o : jObject) = Cadmium.Field.get_double this o#cd'this
    method getFloat (o : jObject) = Cadmium.Field.get_float this o#cd'this
    method getInt (o : jObject) = Cadmium.Field.get_int this o#cd'this
    method getLong (o : jObject) = Cadmium.Field.get_long this o#cd'this
    method getShort (o : jObject) = Cadmium.Field.get_short this o#cd'this
    method set (o : jObject) (x : jObject) = Cadmium.Field.set this o#cd'this x#cd'this
    method setBoolean (o : jObject) x = Cadmium.Field.set_boolean this o#cd'this x
    method setByte (o : jObject) x = Cadmium.Field.set_byte this o#cd'this x
    method setChar (o : jObject) x = Cadmium.Field.set_char this o#cd'this x
    method setDouble (o : jObject) x = Cadmium.Field.set_double this o#cd'this x
    method setFloat (o : jObject) x = Cadmium.Field.set_float this o#cd'this x
    method setInt (o : jObject) x = Cadmium.Field.set_int this o#cd'this x
    method setLong (o : jObject) x = Cadmium.Field.set_long this o#cd'this x
    method setShort (o : jObject) x = Cadmium.Field.set_short this o#cd'this x
  end
and jMethod p =
  let rec conv_type = function
    | Boolean -> Cadmium.Boolean
    | Byte -> Cadmium.Byte
    | Char -> Cadmium.Char
    | Double -> Cadmium.Double
    | Float -> Cadmium.Float
    | Int -> Cadmium.Int
    | Long -> Cadmium.Long
    | Short -> Cadmium.Short
    | String -> Cadmium.String
    | Class (c : Cadmium.java_class) -> Cadmium.Class (new jClass (`Cd'init c))
    | Void -> Cadmium.Void
    | Array t -> Cadmium.Array (conv_type t) in
  let conv_types = List.map conv_type in
  let conv_value = function
    | Boolean_val b -> Cadmium.Boolean_val b
    | Byte_val i -> Cadmium.Byte_val i
    | Char_val i -> Cadmium.Char_val i
    | Double_val f -> Cadmium.Double_val f
    | Float_val f -> Cadmium.Float_val f
    | Int_val i -> Cadmium.Int_val i
    | Long_val i -> Cadmium.Long_val i
    | Short_val i -> Cadmium.Short_val i
    | String_val s -> Cadmium.String_val s
    | Object_val (o : jObject)  -> Cadmium.Object_val o#cd'this in
  let conv_value' = function
    | Cadmium.Boolean_val b -> Boolean_val b
    | Cadmium.Byte_val i -> Byte_val i
    | Cadmium.Char_val i -> Char_val i
    | Cadmium.Double_val f -> Double_val f
    | Cadmium.Float_val f -> Float_val f
    | Cadmium.Int_val i -> Int_val i
    | Cadmium.Long_val i -> Long_val i
    | Cadmium.Short_val i -> Short_val i
    | Cadmium.String_val s -> String_val s
    | Cadmium.Object_val o -> Object_val (new jObject (`Cd'init o)) in
  let conv_values = List.map conv_value in
  let this = match p with
  | `Cd'initObj o -> o#cd'this
  | `Cd'init t -> t in
  object
    inherit (jObject (`Cd'init this))
    method getDeclaringClass = new jClass (`Cd'init (Cadmium.Method.get_declaring_class this))
    method getExceptionTypes = List.map (fun x -> new jClass (`Cd'init x)) (Cadmium.Method.get_exception_types this)
    method getModifiers = Cadmium.Method.get_modifiers this
    method getName = Cadmium.Method.get_name this
    method getParameterTypes = conv_types (Cadmium.Method.get_parameter_types this)
    method getReturnType = conv_type (Cadmium.Method.get_return_type this)
    method invoke (o : jObject) p = conv_value' (Cadmium.Method.invoke this o#cd'this (conv_values p))
  end
and jArray p =
  let rec conv_type = function
  | Boolean -> Cadmium.Boolean
  | Byte -> Cadmium.Byte
  | Char -> Cadmium.Char
  | Double -> Cadmium.Double
  | Float -> Cadmium.Float
  | Int -> Cadmium.Int
  | Long -> Cadmium.Long
  | Short -> Cadmium.Short
  | String -> Cadmium.String
  | Class (c : jClass) -> Cadmium.Class c#cd'this
  | Void -> Cadmium.Void
  | Array t -> Cadmium.Array (conv_type t) in
  let this = match p with
  | `Cd'initObj o -> o#cd'this
  | `Cd'init t -> t
  | `New (t, s) -> Cadmium.Array.new_instance (conv_type t) s
  | `New_dims (t, d) -> Cadmium.Array.new_instance_dims (conv_type t) d
  | `From_bool_array a -> Cadmium.ArrayUtils.bool_array_to_java a
  | `From_byte_array a -> Cadmium.ArrayUtils.byte_array_to_java a
  | `From_char_array a -> Cadmium.ArrayUtils.char_array_to_java a
  | `From_double_array a -> Cadmium.ArrayUtils.double_array_to_java a
  | `From_float_array a -> Cadmium.ArrayUtils.float_array_to_java a
  | `From_int_array a -> Cadmium.ArrayUtils.int_array_to_java a
  | `From_long_array a -> Cadmium.ArrayUtils.long_array_to_java a
  | `From_short_array a -> Cadmium.ArrayUtils.short_array_to_java a
  | `From_string_array a -> Cadmium.ArrayUtils.string_array_to_java a in
  object
    inherit (jObject (`Cd'init this))
    method cd'to_bool_array = Cadmium.ArrayUtils.bool_array_from_java this
    method cd'to_byte_array = Cadmium.ArrayUtils.byte_array_from_java this
    method cd'to_char_array = Cadmium.ArrayUtils.char_array_from_java this
    method cd'to_double_array = Cadmium.ArrayUtils.double_array_from_java this
    method cd'to_float_array = Cadmium.ArrayUtils.float_array_from_java this
    method cd'to_int_array = Cadmium.ArrayUtils.int_array_from_java this
    method cd'to_long_array = Cadmium.ArrayUtils.long_array_from_java this
    method cd'to_short_array = Cadmium.ArrayUtils.short_array_from_java this
    method cd'to_string_array = Cadmium.ArrayUtils.string_array_from_java this
    method getLength = Cadmium.Array.get_length this
    method get i = new jObject (`Cd'init (Cadmium.Array.get this i))
    method getBoolean i = Cadmium.Array.get_boolean this i
    method getByte i = Cadmium.Array.get_byte this i
    method getChar i = Cadmium.Array.get_char this i
    method getDouble i = Cadmium.Array.get_double this i
    method getFloat i = Cadmium.Array.get_float this i
    method getInt i = Cadmium.Array.get_int this i
    method getLong i = Cadmium.Array.get_long this i
    method getShort i = Cadmium.Array.get_short this i
    method set i (x : jObject) = Cadmium.Array.set this i x#cd'this
    method setBoolean i x = Cadmium.Array.set_boolean this i x
    method setByte i x = Cadmium.Array.set_byte this i x
    method setChar i x = Cadmium.Array.set_char this i x
    method setDouble i x = Cadmium.Array.set_double this i x
    method setFloat i x = Cadmium.Array.set_float this i x
    method setInt i x = Cadmium.Array.set_int this i x
    method setLong i x = Cadmium.Array.set_long this i x
    method setShort i x = Cadmium.Array.set_short this i x
  end


(* Exceptions *)

exception Java_exception = Cadmium.Java_exception

let print_stack_trace out e = match e with
| Java_exception (_, _, o) -> Cadmium.print_stack_trace out o
| e -> output_string out (Printexc.to_string e)

let get_exception_instance e = match e with
| Java_exception (_, _, o) -> new jObject (`Cd'init o)
| _ -> invalid_arg "CadmiumObj.get_exception_instance"

let throw_exception e = Cadmium.throw_exception e#cd'this


(* Access to Graphics elements *)

let get_frame () = new jObject (`Cd'init (Cadmium.get_frame ()))

let get_canvas () = new jObject (`Cd'init (Cadmium.get_canvas ()))

let get_buffer () = new jObject (`Cd'init (Cadmium.get_buffer ()))


(* Callbacks *)

let encode_value = function
  | Cadmium.Boolean_val b -> Boolean_val b
  | Cadmium.Byte_val i -> Byte_val i
  | Cadmium.Char_val i -> Char_val i
  | Cadmium.Double_val f -> Double_val f
  | Cadmium.Float_val f -> Float_val f
  | Cadmium.Int_val i -> Int_val i
  | Cadmium.Long_val i -> Long_val i
  | Cadmium.Short_val i -> Short_val i
  | Cadmium.String_val s -> String_val s
  | Cadmium.Object_val o -> Object_val (new jObject (`Cd'init o))

let decode_value = function
  | Boolean_val b -> Cadmium.Boolean_val b
  | Byte_val i -> Cadmium.Byte_val i
  | Char_val i -> Cadmium.Char_val i
  | Double_val f -> Cadmium.Double_val f
  | Float_val f -> Cadmium.Float_val f
  | Int_val i -> Cadmium.Int_val i
  | Long_val i -> Cadmium.Long_val i
  | Short_val i -> Cadmium.Short_val i
  | String_val s -> Cadmium.String_val s
  | Object_val (o : jObject) -> Cadmium.Object_val o#cd'this

let () =
  Callback.register "cadmium_encode_value" encode_value;
  Callback.register "cadmium_decode_value" decode_value

let rec conv_type = function
  | Boolean -> Cadmium.Boolean
  | Byte -> Cadmium.Byte
  | Char -> Cadmium.Char
  | Double -> Cadmium.Double
  | Float -> Cadmium.Float
  | Int -> Cadmium.Int
  | Long -> Cadmium.Long
  | Short -> Cadmium.Short
  | String -> Cadmium.String
  | Class (c : jClass) -> Cadmium.Class c#cd'this
  | Void -> Cadmium.Void
  | Array t -> Cadmium.Array (conv_type t)

let conv_types = List.map conv_type

let create_interface l =
  let res = Cadmium.create_interface
      (List.map (fun x -> let s, p, r = x in s, (conv_types p), (conv_type r)) l) in
  new jClass (`Cd'init res)

let create_named_interface l n =
  let res = Cadmium.create_named_interface
      (List.map (fun x -> let s, p, r = x in s, (conv_types p), (conv_type r)) l)
      n in
  new jClass (`Cd'init res)

external register_object : Cadmium.java_class list -> < .. > -> Cadmium.java_object =
  "cadmium_register_object"

external register_object_transl : Cadmium.java_class list -> < .. > -> (Cadmium.java_method * string) list -> Cadmium.java_object =
  "cadmium_register_object_transl"

let register (c : jClass list) o =
  let c' = List.map (fun (x : jClass) -> x#cd'this) c in
  let res = register_object c' o in new jObject (`Cd'init res)

let register_transl (c : jClass list) o (t : (jMethod * string) list) =
  let c' = List.map (fun (x : jClass) -> x#cd'this) c in
  let t' = List.map (fun (m, s) -> (m#cd'this, s)) t in
  let res = register_object_transl c' o t' in
  new jObject (`Cd'init res)


(* Synchronization *)

let synchronized o f =
  Cadmium.synchronized o#cd'this f


(* Applet *)

type applet_parameter = string * string * string

type applet_handler = {
    info : string;
    parameters : applet_parameter array;
    init : jObject -> unit;
    start : jObject -> unit;
    run : jObject -> unit;
    stop : jObject -> unit;
    destroy : jObject -> unit;
  }

let register_applet h =
  Cadmium.register_applet
    { Cadmium.info = h.info;
      Cadmium.parameters = h.parameters;
      Cadmium.init = (fun o -> h.init (new jObject (`Cd'init o)));
      Cadmium.start = (fun o -> h.start (new jObject (`Cd'init o)));
      Cadmium.run = (fun o -> h.run (new jObject (`Cd'init o)));
      Cadmium.stop = (fun o -> h.stop (new jObject (`Cd'init o)));
      Cadmium.destroy = (fun o -> h.destroy (new jObject (`Cd'init o))); }

let get_applet_parameter applet name =
  Cadmium.get_applet_parameter applet#cd'this name


(* Runtime *)

type interpreter_parameters = Cadmium.interpreter_parameters = {
    file : string;
    arguments : string array;
    init_stack_size : int;
    max_stack_size : int;
    backtrace : bool;
    exit_stops_jvm : bool;
    providers : string array;
    compile_dispatchers : bool;
    use_awt : bool;
    bare_canvas : bool;
    use_javax_sound : bool;
    use_jdbm : bool;
    os : string;
    unix_emulation : bool
  }

let is_native = Cadmium.is_native

let get_coderunner () = new jObject (`Cd'init (Cadmium.get_coderunner ()))

let get_context () = new jObject (`Cd'init (Cadmium.get_context ()))

let get_parameters = Cadmium.get_parameters

class interpreter p =
  let this = match p with
  | `Params (params, file, dir, stdin, stdout, stderr) ->
      let interp = Cadmium.make_interpreter params file dir stdin stdout stderr
      in Cadmium.java_object_of_interpreter interp
  | `Cd'initObj o -> o#cd'this
  | `Cd'init t -> t in
  let that = Obj.magic this in
  object
    inherit (jObject (`Cd'init this))
    method execute : 'a . 'a = Cadmium.interpreter_execute that
    method executeCallback : 'a 'b . string -> 'a list -> 'b =
      fun n p -> Cadmium.interpreter_execute_callback that n p
    method getContext = new jObject (`Cd'init (Cadmium.interpreter_get_context that))
    method getParameters = Cadmium.interpreter_get_parameters that
  end


(* Scripts *)

external is_binding_object : string -> bool =
  "cadmium_is_binding_object"

let get_binding s =
  if is_binding_object s then
    let obj : Cadmium.java_object = Cadmium.get_binding s in
    Obj.magic (new jObject (`Cd'init obj))
  else
    Cadmium.get_binding s
