(*
 * This file is part of Cadmium.
 * Copyright (C) 2007-2010 Xavier Clerc.
 *
 * Cadmium is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * Cadmium is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *)

(** Access to HTTP Servlets. *)

(** The following modules (except {i Listeners}) are "translations" of Java
   classes from the {i javax.servlet} and {i javax.servlet.http} packages.
   Detailed information about the various functions can hence be found in the
   Javadoc of these packages and classes.

   The {i Listeners} module provides means to register functions as
   listeners. *)


module Cookie :
  sig
    type t
          (** Type of cookies. *)

    val make : string -> string -> t
        (** [make n v] constructs a new cookie with name [n] and value [v].
           Raises [Invalid_argument] if [n] contains illegal characters. *)

    val clone : t -> t
        (** Returns a copy of the passed cookie. *)

    val get_comment : t -> string
        (** Returns the comment of the passed cookie. *)

    val get_domain : t -> string
        (** Returns the domain of the passed cookie. *)

    val get_max_age : t -> int32
        (** Returns the maximum age of the passed cookie. *)

    val get_name : t -> string
        (** Returns the name of the passed cookie. *)

    val get_path : t -> string
        (** Returns the path of the passed cookie. *)

    val get_secure : t -> bool
        (** Tests whether the cookie is sent using a secure protocol. *)

    val get_value : t -> string
        (** Returns the value of the passed cookie. *)

    val get_version : t -> int32
        (** Returns the version of the passed cookie. *)

    val set_comment : t -> string -> unit
        (** Sets the comment of the passed cookie. *)

    val set_domain : t -> string -> unit
        (** Sets the domain of the passed cookie. *)

    val set_max_age : t -> int32 -> unit
        (** Sets the maximum age of the passed cookie. *)

    val set_path : t -> string -> unit
        (** Sets the path of the passed cookie. *)

    val set_secure : t -> bool -> unit
        (** Sets whether the passed cookie should be sent using a secure protocol. *)

    val set_value : t -> string -> unit
        (** Sets the value of the passed cookie. *)

    val set_version : t -> int32 -> unit
        (** Sets the version of the passed cookie. *)
  end
(** Bindings for {i javax.servlet.http.Cookie}. *)


module Context :
  sig
    type t
          (** Type of contexts. *)

    val get_attribute : t -> string -> Cadmium.java_object
        (** [get_attribute c n] returns the attribute named [n] in context [c].
           Raises [Not_found] if such an attribute does not exist. *)

    val get_attribute_obj : t -> string -> CadmiumObj.jObject
        (** [get_attribute c n] returns the attribute named [n] in context [c].
           Raises [Not_found] if such an attribute does not exist. *)

    val get_attribute_names : t -> string list
        (** Returns the list of attribute names. *)

    val get_context : t -> string -> t
        (** [get_context c p] returns the context for path [p] in context [c].
           Raises [Not_found] if such a context does not exist. *)

    val get_context_path : t -> string
        (** Returns the context path. *)

    val get_init_parameter : t -> string -> string
        (** [get_init_parameter c n] returns the parameter named [n] in context [c].
           Raises [Not_found] if such a parameter does not exist. *)

    val get_init_parameter_names : t -> string list
        (** Returns the list of parameter names. *)

    val get_major_version : t -> int32
        (** Returns the major version of the Servlet API of the container. *)

    val get_mime_type : t -> string -> string
        (** [get_mime_type c f] returns the MIME type of file [f] in context [c].
           Raises [Not_found] if the MIME type is unknown. *)

    val get_minor_version : t -> int32
        (** Returns the minor version of the Servlet API of the container. *)

    val get_real_path : t -> string -> string
        (** [get_real_path c p] returns the real path for virtual path [p].
           Raises [Invalid_argument] if translation fails. *)

    val get_resource_as_url : t -> string -> string
        (** [get_resource_as_url c p] returns the URL for the resource at path [p].
           Raises [Invalid_argument] if such a resource does not exist. *)

    val get_resource_as_stream : t -> string -> in_channel
        (** [get_resource_as_stream c p] returns the channel for the resource at path [p].
           Raises [Invalid_argument] if such a resource does not exist. *)

    val get_resource_paths : t -> string -> string list
        (** [get_resource_paths c p] returns the resource paths for path [p].
           Raises [Invalid_argument] if no resource exist in the passed path. *)

    val get_server_info : t -> string
        (** Returns the server info string. *)

    val get_servlet_context_name : t -> string
        (** Returns the servlet context name. *)

    val log : t -> string -> unit
        (** [log c m] logs message [m]. *)

    val log_exn : t -> string -> Cadmium.java_object -> unit
        (** [log c m t] logs message [m] and {i Throwable} [t].
           Raises [Invalid_argument] if [t] is not a {i Throwable} instance. *)

    val log_exn_obj : t -> string -> CadmiumObj.jObject -> unit
        (** [log c m t] logs message [m] and {i Throwable} [t].
           Raises [Invalid_argument] if [t] is not a {i Throwable} instance. *)

    val remove_attribute : t -> string -> unit
        (** [remove_attribute c n] removes the attribute named [n] from context [c]. *)

    val set_attribute : t -> string -> Cadmium.java_object -> unit
        (** [set_attribute c n v] adds an attribute with name [n] and value [v]
           to context [c]. *)

    val set_attribute_obj : t -> string -> CadmiumObj.jObject -> unit
        (** [set_attribute c n v] adds an attribute with name [n] and value [v]
           to context [c]. *)
  end
(** Bindings for {i javax.servlet.ServletContext}. *)


module Session :
  sig
    type t
          (** Type of sessions. *)

    val get_attribute : t -> string -> Cadmium.java_object
        (** [get_attribute s n] returns the attribute named [n] in session [s].
           Raises [Not_found] if such an attribute does not exist. *)

    val get_attribute_obj : t -> string -> CadmiumObj.jObject
        (** [get_attribute s n] returns the attribute named [n] in session [s].
           Raises [Not_found] if such an attribute does not exist. *)

    val get_attribute_names : t -> string list
        (** Returns the list of attribute names. *)

    val get_creation_time : t -> int64
        (** Returns the creation time for the passed session. *)

    val get_id : t -> string
        (** Returns the ID for the passed session. *)

    val get_last_accessed_time : t -> int64
        (** Returns the last accessed time for the passed session. *)

    val get_max_inactive_interval : t -> int32
        (** Returns the max inactive interval for the passed session. *)

    val get_servlet_context : t -> Context.t
        (** Returns the context for the passed session. *)

    val invalidate : t -> unit
        (** Invalidates the passed session.
           Raises [Invalid_argument] is the session is already invalidated. *)

    val is_new : t -> bool
        (** Tests whether the client does not yet know about the session.
           Raises [Invalid_argument] is the session is already invalidated. *)

    val remove_attribute : t -> string -> unit
        (** [remove_attribute s n] removes the attribute named [n] from session [s]. *)

    val set_attribute : t -> string -> Cadmium.java_object -> unit
        (** [set_attribute s n v] adds an attribute with name [n] and value [v]
           to session [s]. *)

    val set_attribute_obj : t -> string -> CadmiumObj.jObject -> unit
        (** [set_attribute s n v] adds an attribute with name [n] and value [v]
           to session [s]. *)

    val set_max_inactive_interval : t -> int32 -> unit
        (** [set_max_inactive_interval s m] sets the max inactive interval
           to [m] for the session [s]. *)
  end
(** Bindings for {i javax.servlet.http.HttpSession}. *)


module Listeners :
  sig
    type session_activation_listener = {
        session_did_activate : Session.t -> unit; (** Called when the passed session has just been activated. *)
        session_will_passivate : Session.t -> unit; (** Called when the passed session is about to be passivated. *)
      }
          (** Type for {i javax.servlet.http.HttpSessionActivationListener}. *)

    val register_session_activation_listener : session_activation_listener -> unit
        (** Registers the passed listener, overriding any previous one. *)

    type session_attribute_listener = {
        attribute_added : Session.t -> string -> Cadmium.java_object -> unit; (** Called when the attribute with passed name and value has been added to the passed session. *)
        attribute_removed : Session.t -> string -> Cadmium.java_object -> unit; (** Called when the attribute with passed name and value has been removed to the passed session. *)
        attribute_replaced : Session.t -> string -> Cadmium.java_object -> unit; (** Called when the attribute with passed name and value has been replaced in the passed session. *)
      }
          (** Type for {i javax.servlet.http.HttpSessionAttributeListener}. *)

    val register_session_attribute_listener : session_attribute_listener -> unit
        (** Registers the passed listener, overriding any previous one. *)

    type session_binding_listener = {
        value_bound : Session.t -> string -> Cadmium.java_object -> unit; (** Called when the value with passed name is being bound to passed session. *)
        value_unbound : Session.t -> string -> Cadmium.java_object -> unit; (** Called when the value with passed name is being unbound from passed session. *)
      }
          (** Type for {i javax.servlet.http.HttpSessionBindingListener}. *)

    val register_session_binding_listener : session_binding_listener -> unit
        (** Registers the passed listener, overriding any previous one. *)

    type session_listener = {
        session_created : Session.t -> unit; (** Called when a session was created. *)
        session_destroyed : Session.t -> unit; (** Called when a session is about to be invalidated. *)
      }
          (** Type for {i javax.servlet.http.HttpSessionListener}. *)

    val register_session_listener : session_listener -> unit
        (** Registers the passed listener, overriding any previous one. *)
  end
(** Session listeners. *)


module Request :
  sig
    type t
          (** Type of requests. *)

    val get_attribute : t -> string -> Cadmium.java_object
        (** [get_attribute r n] returns the attribute named [n] in request [r].
           Raises [Not_found] if such an attribute does not exist. *)

    val get_attribute_obj : t -> string -> CadmiumObj.jObject
        (** [get_attribute r n] returns the attribute named [n] in request [r].
           Raises [Not_found] if such an attribute does not exist. *)

    val get_attribute_names : t -> string list
        (** Returns the list of attribute names. *)

    val get_auth_type : t -> string
        (** Returns the authentication scheme of the passed request. *)

    val get_character_encoding : t -> string
        (** Returns the character encoding of the passed request. *)

    val get_content_length : t -> int32
        (** Returns the content length of the passed request. *)

    val get_content_type : t -> string
        (** Returns the content type of the passed request. *)

    val get_context_path : t -> string
        (** Returns the context path of the passed request. *)

    val get_cookies : t -> Cookie.t list
        (** Returns the cookies of the passed request. *)

    val get_date_header : t -> string -> int64
        (** [get_date_header r n] returns the date header named [n] for request [r].
           Returns [-1] if no such header exist.
           Raises [Invalid_argument] if the header is not a date header. *)

    val get_header : t -> string -> string
        (** [get_header r n] returns the header named [n] for request [r].
           Raises [Not_found] if no such header exist. *)

    val get_header_names : t -> string list
        (** Returns the list of header names of the passed request. *)

    val get_headers : t -> string -> string list
        (** [get_headers r n] returns all headers named [n] for request [r]. *)

    val get_input_stream : t -> in_channel
        (** Returns the input stream of the passed request. *)

    val get_int_header : t -> string -> int32
        (** [get_int_header r n] returns the int header named [n] for request [r].
           Returns [-1] if no such header exist.
           Raises [Invalid_argument] if the header is not an int header. *)

    val get_local_addr : t -> string
        (** Returns the local address of the passed request. *)

    val get_locale : t -> string
        (** Returns the locale of the passed request. *)

    val get_locales : t -> string list
        (** Returns the locales of the passed request. *)

    val get_local_name : t -> string
        (** Returns the local name of the passed request. *)

    val get_local_port : t -> int32
        (** Returns the local port of the passed request. *)

    val get_method : t -> string
        (** Returns the method of the passed request. *)

    val get_parameter : t -> string -> string
        (** [get_parameter r n] returns the parameter named [n] in request [r].
           Raises [Not_found] if such a parameter does not exist. *)

    val get_parameter_list : t -> ((string * (string list)) list)
        (** Returns the parameter list of the passed request as an association
           list from parameter names to lists of parameter values. *)

    val get_parameter_names : t -> string list
        (** Returns the list of parameter names. *)

    val get_parameter_values : t -> string -> string list
        (** [get_parameter r n] returns all values for the parameter named [n]
           in request [r].
           Raises [Not_found] if such a parameter does not exist. *)

    val get_path_info : t -> string
        (** Returns the path info of the passed request. *)

    val get_path_translated : t -> string
        (** Returns the translated path of the passed request. *)

    val get_protocol : t -> string
        (** Returns the protocol of the passed request. *)

    val get_query_string : t -> string
        (** Returns the query string of the passed request. *)

    val get_remote_addr : t -> string
        (** Returns the remote address of the passed request. *)

    val get_remote_host : t -> string
        (** Returns the remote host of the passed request. *)

    val get_remote_port : t -> int32
        (** Returns the remote port of the passed request. *)

    val get_remote_user : t -> string
        (** Returns the remote user of the passed request. *)

    val get_requested_session_id : t -> string
        (** Returns the requested session ID of the passed request. *)

    val get_request_uri : t -> string
        (** Returns the request URI of the passed request. *)

    val get_request_url : t -> string
        (** Returns the request URL of the passed request. *)

    val get_scheme : t -> string
        (** Returns the scheme of the passed request. *)

    val get_server_name : t -> string
        (** Returns the server name of the passed request. *)

    val get_server_port : t -> int32
        (** Returns the server port of the passed request. *)

    val get_servlet_path : t -> string
        (** Returns the servlet path of the passed request. *)

    val get_session : t -> Session.t
        (** Returns the session of the passed request, creating one if there
           is no current session. *)

    val get_session_creating : t -> bool -> Session.t
        (** Returns the session of the passed request.
           Passed boolean indicates whether session should be created if there
           is no current session.
           Raises [Not_found] if passed boolean is [false] and there is no
           current session. *)

    val is_requested_session_id_from_cookie : t -> bool
        (** Tests whether the requested session ID is passed by a cookie. *)

    val is_requested_session_id_from_url : t -> bool
        (** Tests whether the requested session ID is passed by the URL. *)

    val is_requested_session_id_valid : t -> bool
        (** Tests whether the requested session ID is still valid. *)

    val is_secure : t -> bool
        (** Tests whether the request was made using a secure channel. *)

    val is_user_in_role : t -> string -> bool
        (** Tests whether the authenticated user is included in the specified
           logical {i role}. *)

    val remove_attribute : t -> string -> unit
        (** [remove_attribute r n] removes the attribute named [n] from request [r]. *)

    val set_attribute : t -> string -> Cadmium.java_object -> unit
        (** [set_attribute r n v] adds an attribute with name [n] and value [v]
           to request [r]. *)

    val set_attribute_obj : t -> string -> CadmiumObj.jObject -> unit
        (** [set_attribute r n v] adds an attribute with name [n] and value [v]
           to request [r]. *)

    val set_character_encoding : t -> string -> unit
        (** [set_character_encoding r e] sets the character encoding of request
           [r] to [e]. *)
  end
(** Bindings for {i javax.servlet.http.HttpServletRequest}. *)


module Response :
  sig
    val sc_accepted : int32
        (** Status code. *)

    val sc_bad_gateway : int32
        (** Status code. *)

    val sc_bad_request : int32
        (** Status code. *)

    val sc_conflict : int32
        (** Status code. *)

    val sc_continue : int32
        (** Status code. *)

    val sc_created : int32
        (** Status code. *)

    val sc_expectation_failed : int32
        (** Status code. *)

    val sc_forbidden : int32
        (** Status code. *)

    val sc_found : int32
        (** Status code. *)

    val sc_gateway_tiemout : int32
        (** Status code. *)

    val sc_gone : int32
        (** Status code. *)

    val sc_http_version_not_supported : int32
        (** Status code. *)

    val sc_internal_server_error : int32
        (** Status code. *)

    val sc_length_required : int32
        (** Status code. *)

    val sc_method_not_allowed : int32
        (** Status code. *)

    val sc_moved_permanently : int32
        (** Status code. *)

    val sc_moved_temporarily : int32
        (** Status code. *)

    val sc_multiple_choices : int32
        (** Status code. *)

    val sc_no_content : int32
        (** Status code. *)

    val sc_non_authoritative_information : int32
        (** Status code. *)

    val sc_not_acceptable : int32
        (** Status code. *)

    val sc_not_found : int32
        (** Status code. *)

    val sc_not_implemented : int32
        (** Status code. *)

    val sc_not_modified : int32
        (** Status code. *)

    val sc_ok : int32
        (** Status code. *)

    val sc_partial_content : int32
        (** Status code. *)

    val sc_payment_required : int32
        (** Status code. *)

    val sc_precondition_failed : int32
        (** Status code. *)

    val sc_proxy_authentication_required : int32
        (** Status code. *)

    val sc_request_entity_too_large : int32
        (** Status code. *)

    val sc_request_timeout : int32
        (** Status code. *)

    val sc_request_uri_too_long : int32
        (** Status code. *)

    val sc_requested_range_not_satisfiable : int32
        (** Status code. *)

    val sc_reset_content : int32
        (** Status code. *)

    val sc_see_other : int32
        (** Status code. *)

    val sc_service_unavailable : int32
        (** Status code. *)

    val sc_switching_protocols : int32
        (** Status code. *)

    val sc_temporary_redirect : int32
        (** Status code. *)

    val sc_unauthorized : int32
        (** Status code. *)

    val sc_unsupported_media_type : int32
        (** Status code. *)

    val sc_use_proxy : int32
        (** Status code. *)

    type t
          (** Type of responses. *)

    val add_cookie : t -> Cookie.t -> unit
        (** [add_cookie r c] adds cookie [c] to response [r]. *)

    val add_date_header : t -> string -> int64 -> unit
        (** [add_date_header r n v] adds a date header named [n] with value [v]
           to response [r]. *)

    val add_header : t -> string -> string -> unit
        (** [add_header r n v] adds a header named [n] with value [v]
           to response [r]. *)

    val add_int_header : t -> string -> int32 -> unit
        (** [add_int_header r n v] adds an int header named [n] with value [v]
           to response [r]. *)

    val contains_header : t -> string -> bool
        (** Tests whether the response contains a header with the passed name. *)

    val encode_redirect_url : t -> string -> string
        (** Encodes the passed URL for use in [send_redirect]. *)

    val encode_url : t -> string -> string
        (** Encodes the passed URL. *)

    val flush_buffer : t -> unit
        (** Forces the content in the buffer to be written to the client. *)

    val get_buffer_size : t -> int32
        (** Returns the buffer size for the passed response. *)

    val get_character_encoding : t -> string
        (** Returns the character encoding for the passed response. *)

    val get_content_type : t -> string
        (** Returns the content type for the passed response. *)

    val get_locale : t -> string
        (** Returns the locale for the passed response. *)

    val get_output_stream : t -> out_channel
        (** Returns the output stream for the passed response. *)

    val is_committed : t -> bool
        (** Tests whether the passed response has been committed. *)

    val reset : t -> unit
        (** Clears the buffer as well as the status code and headers.
           Raises [Invalid_arugment] if the passed reponse has already been
           committed. *)

    val reset_buffer : t -> unit
        (** Clears the buffer.
           Raises [Invalid_arugment] if the passed reponse has already been
           committed. *)

    val send_error : t -> int32 -> unit
        (** [send_error r c] sends an error for response [r] with code [c].
           Raises [Invalid_arugment] if the passed reponse has already been
           committed. *)

    val send_error_msg : t -> int32 -> string -> unit
        (** [send_error_msg r c] sends an error for response [r] with code [c]
           and message [m].
           Raises [Invalid_arugment] if the passed reponse has already been
           committed. *)

    val send_redirect : t -> string -> unit
        (** Sends a redirect reponse to the passed URL.
           Raises [Invalid_arugment] if the passed reponse has already been
           committed. *)

    val set_buffer_size : t -> int32 -> unit
        (** [set_buffer_size r s] sets the buffer size for response [r] to [s].
           Raises [Invalid_arugment] if the passed reponse has already been
           committed.*)

    val set_character_encoding : t -> string -> unit
        (** [set_character_encoding r e] sets the character encoding for response [r] to [e]. *)

    val set_content_length : t -> int32 -> unit
        (** [set_content_length r l] sets the content length for response [r] to [l]. *)

    val set_content_type : t -> string -> unit
        (** [set_content_type r t] sets the content type for response [r] to [t]. *)

    val set_date_header : t -> string -> int64 -> unit
        (** [set_date_header r n v] sets the date header named [n] to value [v]
           for response [r]. *)

    val set_header : t -> string -> string -> unit
        (** [set_header r n v] sets the header named [n] to value [v]
           for response [r]. *)

    val set_int_header : t -> string -> int32 -> unit
        (** [set_int_header r n v] sets the int header named [n] to value [v]
           for response [r]. *)

    val set_locale : t -> string -> unit
        (** [set_locale r l] sets the locale for response [r] to [l]. *)

    val set_status : t -> int32 -> unit
        (** [set_status r c] sets the status code for response [r] to [c]. *)
  end
(** Bindings for {i javax.servlet.http.HttpServletResponse}. *)


module Servlet :
  sig
    type t
          (** Type of servlets. *)

    val get_init_parameter : t -> string -> string
        (** [get_init_parameter s n] returns the parameter named [n] in servlet [s].
           Raises [Not_found] if such a parameter does not exist. *)

    val get_init_parameter_names : t -> string list
        (** Returns the list of parameter names. *)

    val get_servlet_context : t -> Context.t
        (** Returns the context of the passed servlet. *)

    val get_servlet_name : t -> string
        (** Returns the name of the passed servlet. *)

    val log : t -> string -> unit
        (** [log c m] logs message [m]. *)

    val log_exn : t -> string -> Cadmium.java_object -> unit
        (** [log c m t] logs message [m] and {i Throwable} [t].
           Raises [Invalid_argument] if [t] is not a {i Throwable} instance. *)

    val log_exn_obj : t -> string -> CadmiumObj.jObject -> unit
        (** [log c m t] logs message [m] and {i Throwable} [t].
           Raises [Invalid_argument] if [t] is not a {i Throwable} instance. *)

    type handler = {
        destroy : t -> unit; (** Called by the container when the servlet is disposed. *)
        init : t -> unit; (** Called by the container when the servlet is initialized. *)
        info : string; (** Information string returned by the servlet. *)
        do_delete : (t -> Request.t -> Response.t -> unit) option; (** Handles {i delete} requests. *)
        do_get : (t -> Request.t -> Response.t -> unit) option; (** Handles {i get} requests. *)
        do_head : (t -> Request.t -> Response.t -> unit) option; (** Handles {i head} requests. *)
        do_options : (t -> Request.t -> Response.t -> unit) option; (** Handles {i options} requests. *)
        do_post : (t -> Request.t -> Response.t -> unit) option; (** Handles {i post} requests. *)
        do_put : (t -> Request.t -> Response.t -> unit) option; (** Handles {i put} requests. *)
        do_trace : (t -> Request.t -> Response.t -> unit) option; (** Handles {i trace} requests. *)
        get_last_modified : (t -> Request.t -> int64) option; (** Returns the time the requested object was last modified. *)
      }
          (** Servlet request handler. [None] values should be used for
             unhandled requests. *)

    val register : handler -> unit
        (** Registers the passed handler, overriding any previous one. *)
  end
(** Bindings and handler for {i javax.servlet.http.HttpServlet}. *)
