(*
 * This file is part of Cadmium.
 * Copyright (C) 2007-2010 Xavier Clerc.
 *
 * Cadmium is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * Cadmium is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *)

(** Access to XML parsers and transformers. *)

(** The following modules (except {i Tree}) are "translations" of the Java classes
   having the same name in packages {i javax.xml.parsers}, {i javax.xml.transform},
   {i javax.xml.validation}, {i javax.xml.xpath}, as well as {i org.w3c.dom},
   and {i org.xml.sax}. Detailed information about the various functions can hence be
   found in the Javadoc of these packages and classes.

   The {i Tree} module is not the "translation" of a Java class, only providing a
   simplified representation of XML documents. *)


module SchemaFactory :
  sig
    type t
          (** Type of schema factories. *)

    val make : string -> t
        (** Constructs a schema factory, passed argument being the schema language. *)

    val set_feature : t -> string -> bool -> unit
        (** [set_feature f n v] sets value [v] to feature named [n] for factory [f]. *)

    val get_feature : t -> string -> bool
        (** [get_feature f n] gets value of feature named [n] for factory [f]. *)

    val set_property : t -> string -> Cadmium.java_object -> unit
        (** [set_property f n v] sets value [v] to property named [n] for factory [f]. *)

    val set_property_obj : t -> string -> CadmiumObj.jObject -> unit
        (** [set_property f n v] sets value [v] to property named [n] for factory [f]. *)

    val get_property : t -> string -> Cadmium.java_object
        (** [get_property f n] gets value of property named [n] for factory [f]. *)

    val get_property_obj : t -> string -> CadmiumObj.jObject
        (** [get_property f n] gets value of property named [n] for factory [f]. *)
  end
(** Schema factories. *)


module Schema :
  sig
    type t
          (** Type of schemas. *)

    val make_schema_from_file : SchemaFactory.t -> string -> t
        (** [make_schema_from_file f x] constructs a schema from file [x]
           using factory [f]. *)

    val make_schema_from_url : SchemaFactory.t -> string -> t
        (** [make_schema_from_url f x] constructs a schema from url [x]
           using factory [f]. *)
  end
(** Schema builders. *)


module SAXFactory :
  sig
    type t
          (** Type of SAX parser factories. *)

    val make : unit -> t
        (** Constructs a SAX parser factory. *)

    val set_feature : t -> string -> bool -> unit
        (** [set_feature f n v] sets value [v] to feature named [n] for factory [f]. *)

    val get_feature : t -> string -> bool
        (** [get_feature f n] gets value of feature named [n] for factory [f]. *)

    val set_schema : t -> Schema.t -> unit
        (** [set_schema f s] sets the schema for constructed parsers to [s]
           for factory [f]. *)

    val get_schema : t -> Schema.t
        (** [get_schema f] gets the schema for constructed parsers for factory [f]. *)

    val set_namespace_aware : t -> bool -> unit
        (** Sets the 'namespace-aware' property for constructed parsers. *)

    val is_namespace_aware : t -> bool
        (** Gets the 'namespace-aware' property for constructed parsers. *)

    val set_validating : t -> bool -> unit
        (** Sets the 'validating' property for constructed parsers. *)

    val is_validating : t -> bool
        (** Gets the 'validating' property for constructed parsers. *)

    val set_xinclude_aware : t -> bool -> unit
        (** Sets the 'xinclude-aware' property for constructed parsers. *)

    val is_xinclude_aware : t -> bool
        (** Gets the 'xinclude-aware' property for constructed parsers. *)
  end
(** SAX parser factories. *)


module SAXParser :
  sig
    type t
          (** Type of SAX parsers. *)

    type parse_error = int * int * string
          (** Type of parse error. First component is line number, while second one is
             column number. The last component is the exception message associated with
             the parse error. *)

    type handler = {
        characters : string -> unit;
          (** Called for notification of characters inside an element. *)
        ignorable_whitespace : string -> unit;
          (** Called for notification of ignorable whitespace inside an element. *)
        start_document : unit -> unit;
          (** Called for notification of document begin. *)
        end_document : unit -> unit;
          (** Called for notification of document end. *)
        start_element : string -> string -> string -> ((string * string) list) -> unit;
          (** Called for notification of element start. Parameters are:
             - URI;
             - local name for tag;
             - qualified name for tag;
             - association list from attribute names to attribute values. *)
        end_element : string -> string -> string -> unit;
          (** Called for notification of element end. Parameters are:
             - URI;
             - local name for tag;
             - qualified name for tag. *)
        start_prefix_mapping : string -> string -> unit;
          (** Called for notification of prefix mapping.
             Parameters are prefix and URI.*)
        end_prefix_mapping : string -> unit;
          (** Called for notification of prefix mapping. Parameter is prefix. *)
        error : parse_error -> unit;
          (** Called for notification of recoverable parse error. *)
        fatal_error : parse_error -> unit;
          (** Called for notification of fatal parse error. *)
        warning : parse_error -> unit;
          (** Called for notification of parse warning. *)
      }
          (** Type of SAX handlers, collection of callback functions. *)

    val default_handler : handler
        (** Default handler, all callbacks do nothing. To be used to construct handler
           using the {i with} notation:
           [{ default_handler with characters = print_endline }]. *)

    val make : SAXFactory.t -> t
        (** Constructs a new SAX parser. *)

    val set_property : t -> string -> Cadmium.java_object -> unit
        (** [set_property p n v] sets value [v] to property named [n] for parser [p]. *)

    val set_property_obj : t -> string -> CadmiumObj.jObject -> unit
        (** [set_property p n v] sets value [v] to property named [n] for parser [p]. *)

    val get_property : t -> string -> Cadmium.java_object
        (** [get_property p n] gets value of property named [n] for parser [p]. *)

    val get_property_obj : t -> string -> CadmiumObj.jObject
        (** [get_property p n] gets value of property named [n] for parser [p]. *)

    val get_schema : t -> Schema.t
        (** Gets the schema for the passed parser. *)

    val is_namespace_aware : t -> bool
        (** Gets the 'namespace-aware' property for the passed parser. *)

    val is_validating : t -> bool
        (** Gets the 'validating' property for the passed parser. *)

    val is_xinclude_aware : t -> bool
        (** Gets the 'xinclude-aware' property for the passed parser. *)

    val parse_file : t -> handler -> string -> unit
        (** [parse_file p h f] parses file [f] using parser [p] and handler [h]. *)

    val parse_channel : t -> handler -> in_channel -> unit
        (** [parse_channel p h ch] parses channel [ch] using parser [p] and handler [h]. *)

    val reset : t -> unit
        (** Resets the passed SAX parser. *)
  end
(** SAX parsers. *)


module Document :
  sig
        (** {6 Types} *)

    type attr
          (** Type of attribute nodes. *)

    type cdata
          (** Type of cdata nodes. *)

    type comment
          (** Type of comment nodes. *)

    type document
          (** Type of document nodes. *)

    type document_type
          (** Type of document type nodes. *)

    type element
          (** Type of element nodes. *)

    type entity
          (** Type of entity nodes. *)

    type entity_reference
          (** Type of entity reference nodes. *)

    type notation
          (** Type of notation nodes. *)

    type processing_instruction
          (** Type of processing instruction nodes. *)

    type text
          (** Type of text nodes. *)

    type node =
      | Attr of attr (** attribute as bare node *)
      | Cdata of cdata (** cdata as bare node *)
      | Comment of comment (** comment as bare node *)
      | Document of document (** document as bare node *)
      | Document_type of document_type (** document type as bare node *)
      | Element of element (** element as bare node *)
      | Entity of entity (** entity as bare node *)
      | Entity_reference of entity_reference (** entity reference as bare node *)
      | Notation of notation (** notation as bare node *)
      | Processing_instruction of processing_instruction (** processing instruction as bare node *)
      | Text of text (** text as bare node *)
          (** Type of nodes. *)


            (** {6 Document functions} *)

    val get_document_element : document -> element
        (** Returns the root element of the document. *)

    val get_xml_encoding : document -> string
        (** Returns the document encoding. *)

    val set_strict_error_checking : document -> bool -> unit
        (** Sets the 'strict-error-checking' property for passed document. *)

    val is_strict_error_checking : document -> bool
        (** Gets the 'strict-error-checking' property for passed document. *)

    val set_xml_standalone : document -> bool -> unit
        (** Sets the 'stand-alone' property for passed document. *)

    val is_xml_standalone : document -> bool
        (** Gets the 'stand-alone' property for passed document. *)

    val set_xml_version : document -> bool -> unit
        (** Sets the 'xml-version' property for passed document. *)

    val get_xml_version : document -> string
        (** Gets the 'xml-version' property for passed document. *)

    val normalize_document : document -> unit
        (** Normalizes the passed document. *)

    val node_of_document : document -> node
        (** Converts the passed document into a bare node. *)


        (** {6 Attribute functions} *)

    val get_attr_parent : attr -> element
        (** Returns the parent of the passed attribute. *)

    val get_attr_name : attr -> string
        (** Returns the name of the passed attribute. *)

    val get_attr_value : attr -> string
        (** Returns the value of the passed attribute. *)

    val set_attr_value : attr -> string -> unit
        (** Changes the value of the passed attribute. *)

    val node_of_attr : attr -> node
        (** Converts the passed attribute into a bare node. *)


        (** {6 CData functions} *)

    val get_cdata_parent : cdata -> node
        (** Returns the parent of the passed cdata. *)

    val get_cdata : cdata -> string
        (** Returns the data of the passed cdata. *)

    val set_cdata : cdata -> string -> unit
        (** Changes the data of the passed cdata. *)

    val node_of_cdata : cdata -> node
        (** Converts the passed cdata into a bare node. *)


        (** {6 Comment functions} *)

    val get_comment_parent : comment -> node
        (** Returns the parent of the passed comment. *)

    val get_comment : comment -> string
        (** Returns the data of the passed comment. *)

    val set_comment : comment -> string -> unit
        (** Changes the data of the passed comment. *)

    val node_of_comment : comment -> node
        (** Converts the passed comment into a bare node. *)


        (** {6 Document type functions} *)

    val get_document_type_parent : document_type -> node
        (** Returns the parent of the passed document type. *)

    val get_document_type_name : document_type -> string
        (** Returns the name of the passed document type. *)

    val get_document_type_public_id : document_type -> string
        (** Returns the public ID of the passed document type. *)

    val get_document_type_system_id : document_type -> string
        (** Returns the system ID of the passed document type. *)

    val node_of_document_type : document_type -> node
        (** Converts the passed document type into a bare node. *)


        (** {6 Element functions} *)

    val get_element_parent : element -> node
        (** Returns the parent of the passed element. *)

    val get_attribute : element -> string -> string
        (** [get_attribute e n] returns the value for the attribute of element [e]
           named [n]. *)

    val get_attribute_ns : element -> string -> string -> string
        (** [get_attribute_ns e ns n] returns the value for the attribute of element [e]
           named [n] in namespace [ns]. *)

    val get_attribute_node : element -> string -> attr
        (** [get_attribute_node e n] returns the attribute of element [e] named [n]. *)

    val get_attribute_node_ns : element -> string -> string -> attr
        (** [get_attribute_node_ns e ns n] returns the attribute of element [e] named [n]
           in namespace [ns]. *)

    val get_elements_by_tag_name : element -> string -> element list
        (** [get_elements_by_tag_name e t] returns the child elements of [e] whose tag
           is [t]. *)

    val get_elements_by_tag_name_ns : element -> string -> string -> element list
        (** [get_elements_by_tag_name_ns e ns t] returns the child elements of [e] whose
           tag is [t], using namespace ns. *)

    val get_tag_name : element -> string
        (** Returns the tag name of the passed element. *)

    val has_attribute : element -> string -> bool
        (** [has_attribute e n] tests if element [e] has an attribute named [n]. *)

    val has_attribute_ns : element -> string -> string -> bool
        (** [has_attribute_ns e ns n] tests if element [e] has an attribute named [n]
           in namespace [ns]. *)

    val remove_attribute : element -> string -> unit
        (** [remove_attribute e n] removes the attribute named [n] from element [e]. *)

    val remove_attribute_ns : element -> string -> string -> unit
        (** [remove_attribute_ns e ns n] removes the attribute named [n] from element [e]
           in namespace [ns]. *)

    val remove_attribute_node : element -> attr -> unit
        (** [remove_attribute_node e n] removes attribute [n] from element [e]. *)

    val set_attribute : element -> string -> string -> unit
        (** [set_attribute e n v] sets the value of the attribute named [n] to the value
           [v] for element [e]. *)

    val set_attribute_ns : element -> string -> string -> string -> unit
        (** [set_attribute e n v] sets the value of the attribute named [n] to the value
           [v] for element [e] in namespace [ns]. *)

    val node_of_element : element -> node
        (** Converts the passed element into a bare node. *)


        (** {6 Entity functions} *)

    val get_entity_parent : entity -> node
        (** Returns the parent of the passed entity. *)

    val get_entity_notation_name : entity -> string
        (** Returns the notation name of the passed entity. *)

    val get_entity_public_id : entity -> string
        (** Returns the public ID of the passed entity. *)

    val get_entity_system_id : entity -> string
        (** Returns the system ID of the passed entity. *)

    val node_of_entity : entity -> node
        (** Converts the passed entity into a bare node. *)


        (** {6 Notation functions} *)

    val get_notation_parent : notation -> node
        (** Returns the parent of the passed notation. *)

    val get_notation_public_id : notation -> string
        (** Returns the public ID of the passed notation. *)

    val get_notation_system_id : notation -> string
        (** Returns the system ID of the passed notation. *)

    val node_of_notation : notation -> node
        (** Converts the passed notation into a bare node. *)


        (** {6 Processing instruction functions} *)

    val get_processing_instruction_parent : processing_instruction -> node
        (** Returns the parent of the passed processing instruction. *)

    val get_processing_instruction_target : processing_instruction -> string
        (** Returns the target of the passed processing instruction. *)

    val get_processing_instruction_data : processing_instruction -> string
        (** Returns the data of the passed processing instruction. *)

    val set_processing_instruction_data : processing_instruction -> string -> unit
        (** Changes the data of the passed processing instruction. *)

    val node_of_processing_instruction : processing_instruction -> node
        (** Converts the passed processing instruction into a bare node. *)


        (** {6 Text functions} *)

    val get_text_parent : text -> node
        (** Returns the parent of the passed text. *)

    val get_text : text -> string
        (** Returns the data of the passed text. *)

    val set_text : text -> string -> unit
        (** Changes the data of the passed text. *)

    val node_of_text : text -> node
        (** Converts the passed text into a bare node. *)


        (** {6 Node functions} *)

    val get_parent : node -> node
        (** Returns the parent of the passed node.
           Raises [Not_found] if such a node does not exist. *)

    val get_document : node -> document
        (** Returns the enclosing document of the passed node. *)

    val append_attr_child : node -> string -> attr
        (** [append_attr_child n a] adds an attribute child with name [a] to node [n].
           Returns the added child. *)

    val append_attr_ns_child : node -> string -> string -> attr
        (** [append_attr_ns_child n ns a] adds an attribute child with name [a] to node
           [n], using namespace [ns].
           Returns the added child. *)

    val append_cdata_child : node -> string -> cdata
        (** [append_cdata_child n d] adds a cdata child with data [d] to node [n].
           Returns the added child. *)

    val append_comment_child : node -> string -> comment
        (** [append_comment_child n d] adds a comment child with data [d] to node [n].
           Returns the added child. *)

    val append_element_child : node -> string -> element
        (** [append_element_child n t] adds an element child with tag [t] to node [n].
           Returns the added child. *)

    val append_element_ns_child : node -> string -> string -> element
        (** [append_element_ns_child n t] adds an element child with tag [t] to node [n]
           in namespace [ns].
           Returns the added child. *)

    val append_entity_reference_child : node -> string -> entity_reference
        (** [append_entity_reference_child n e] adds an entity reference child for
           reference named [e] to node [n].
           Returns the added child. *)

    val insert_attr_child_before : node -> node -> string -> attr
        (** [insert_attr_child n b a] adds an attribute child with name [a] to node [n]
           before child [b].
           Returns the added child. *)

    val insert_attr_ns_child_before : node -> node -> string -> string -> attr
        (** [insert_attr_ns_child n b ns a] adds an attribute child with name [a] to node
           [n] before child [b], using namespace [ns].
           Returns the added child. *)

    val insert_cdata_child_before : node -> node -> string -> cdata
        (** [insert_cdata_child n b d] adds a cdata child with data [d] to node [n]
           before child [b].
           Returns the added child. *)

    val insert_comment_child_before : node -> node -> string -> comment
        (** [insert_comment_child n b d] adds a comment child with data [d] to node [n]
           before child [b].
           Returns the added child. *)

    val insert_element_child_before : node -> node -> string -> element
        (** [insert_element_child n b t] adds an element child with tag [t] to node [n]
           before child [b].
           Returns the added child. *)

    val insert_element_ns_child_before : node -> node -> string -> string -> element
        (** [insert_element_ns_child n b t] adds an element child with tag [t]
           to node [n] before child [b] in namespace [ns].
           Returns the added child. *)

    val insert_entity_reference_child_before : node -> node -> string -> entity_reference
        (** [insert_entity_reference_child n b e] adds an entity reference child for
           reference named [e] to node [n] before child [b].
           Returns the added child. *)

    val get_child_nodes : node -> node list
        (** Returns the child nodes of the passed node. *)

    val get_first_child : node -> node
        (** Returns the first child of the passed node.
           Raises [Not_found] if such a node does not exist. *)

    val get_last_child : node -> node
        (** Returns the last child of the passed node.
           Raises [Not_found] if such a node does not exist. *)

    val get_previous_sibling : node -> node
        (** Returns the previous sibling of the passed node.
           Raises [Not_found] if such a node does not exist. *)

    val get_next_sibling : node -> node
        (** Returns the next sibling of the passed node.
           Raises [Not_found] if such a node does not exist. *)

    val remove_child : node -> node -> unit
        (** [remove_child n c] removes child [c] from node [n]. *)

    val replace_child : node -> node -> node -> unit
        (** [replace_child n nw od] replaces child [od] with [nw] for node [n]. *)
  end
(** Document manipulation. *)


module DocumentFactory :
  sig
    type t
          (** Type of document factories. *)

    val make : unit -> t
        (** Constructs a new factory. *)

    val set_attribute : t -> string -> Cadmium.java_object -> unit
        (** [set_attribute f n v] sets value [v] to attribute named [n] for factory [f]. *)

    val set_attribute_obj  : t -> string -> CadmiumObj.jObject -> unit
        (** [set_attribute f n v] sets value [v] to attribute named [n] for factory [f]. *)

    val get_attribute : t -> string -> Cadmium.java_object
        (** [get_attribute f n] gets value of attribute named [n] for factory [f]. *)

    val get_attribute_obj : t -> string -> CadmiumObj.jObject
        (** [get_attribute f n] gets value of attribute named [n] for factory [f]. *)

    val set_feature : t -> string -> bool -> unit
        (** [set_feature f n v] sets value [v] to feature named [n] for factory [f]. *)

    val get_feature : t -> string -> bool
        (** [get_feature f n] gets value of feature named [n] for factory [f]. *)

    val set_schema : t -> Schema.t -> unit
        (** [set_schema f s] sets the schema for constructed parsers to [s]
           for factory [f]. *)

    val get_schema : t -> Schema.t
        (** [get_schema f] gets the schema for constructed parsers for factory [f]. *)

    val set_coalescing : t -> bool -> unit
        (** Sets the 'coalescing' property for constructed parsers. *)

    val is_coalescing : t -> bool
        (** Gets the 'coalescing' property for constructed parsers. *)

    val set_expand_entity_references : t -> bool -> unit
        (** Sets the 'expand-entity' property for constructed parsers. *)

    val is_expand_entity_references : t -> bool
        (** Gets the 'expand-entity' property for constructed parsers. *)

    val set_ignoring_comments : t -> bool -> unit
        (** Sets the 'ignoring-comments' property for constructed parsers. *)

    val is_ignoring_comments : t -> bool
        (** Gets the 'ignoring-comments' property for constructed parsers. *)

    val set_ignoring_element_content_whitespace : t -> bool -> unit
        (** Sets the 'ignoring-element-content-whitespace' property for constructed parsers. *)

    val is_ignoring_element_content_whitespace : t -> bool
        (** Gets the 'ignoring-element-content-whitespace' property for constructed parsers. *)

    val set_namespace_aware : t -> bool -> unit
        (** Sets the 'namespace-aware' property for constructed parsers. *)

    val is_namespace_aware : t -> bool
        (** Gets the 'namespace-aware' property for constructed parsers. *)

    val set_validating : t -> bool -> unit
        (** Sets the 'validating' property for constructed parsers. *)

    val is_validating : t -> bool
        (** Gets the 'validating' property for constructed parsers. *)

    val set_xinclude_aware : t -> bool -> unit
        (** Sets the 'xinclude-aware' property for constructed parsers. *)

    val is_xinclude_aware : t -> bool
        (** Gets the 'xinclude-aware' property for constructed parsers. *)
  end
(** Document builder factories. *)


module DocumentBuilder :
  sig
    type t
          (** Type of transformer builders. *)

    val make : DocumentFactory.t -> t
        (** Constructs a new builder from passed factory. *)

    val get_schema : t -> Schema.t
        (** [get_schema b] gets the schema for builder [b]. *)

    val is_namespace_aware : t -> bool
        (** Gets the 'namespace-aware' property for passed builder. *)

    val is_validating : t -> bool
        (** Gets the 'validating' property for passed builder. *)

    val is_xinclude_aware : t -> bool
        (** Gets the 'xinclude-aware' property for passed builder. *)

    val new_document : t -> Document.document
        (** Creates an empty document. *)

    val parse_file : t -> string -> Document.document
        (** [parse_file b f] parses file [f] using builder [b]. *)

    val parse_channel : t -> in_channel -> Document.document
        (** [parse_channel b ch] parses channel [ch] using builder [b]. *)

    val reset : t -> unit
        (** Resets the passed builder. *)
  end
(** Document builders. *)


module Tree :
  sig
    type t = Element of element | Cdata of string | Text of string
    and element = { tag : string; attributes : (string * string) list; children : t list; }

    val of_document : Document.document -> t
        (** Converts the passed document into its simplified representation. *)
  end
(** Simplified representation of XML documents. *)


module TransformerFactory :
  sig
    type t
          (** Type of transformer factories. *)

    val make : unit -> t
        (** Constructs a transformer factory. *)

    val set_attribute : t -> string -> Cadmium.java_object -> unit
        (** [set_attribute f n v] sets value [v] to attribute named [n] for factory [f]. *)

    val set_attribute_obj : t -> string -> CadmiumObj.jObject -> unit
        (** [set_attribute f n v] sets value [v] to attribute named [n] for factory [f]. *)

    val get_attribute : t -> string -> Cadmium.java_object
        (** [get_attribute f n] gets value of attribute named [n] for factory [f]. *)

    val get_attribute_obj : t -> string -> CadmiumObj.jObject
        (** [get_attribute f n] gets value of attribute named [n] for factory [f]. *)

    val set_feature : t -> string -> bool -> unit
        (** [set_feature f n v] sets value [v] to feature named [n] for factory [f]. *)

    val get_feature : t -> string -> bool
        (** [get_feature f n] gets value of feature named [n] for factory [f]. *)
  end
(** Transformer factories. *)


module Transformer :
  sig
    type t
          (** Type of transformers. *)

    val make_id : TransformerFactory.t -> t
        (** Constructs an identity transformer. *)

    val make_from_file : TransformerFactory.t -> string -> t
        (** [make_from_file f fn] constructs a transformer from file [fn],
           using factory [f]. *)

    val make_from_channel : TransformerFactory.t -> in_channel -> t
        (** [make_from_channel f ch] constructs a transformer from channel [ch],
           using factory [f]. *)

    val clear_parameters : t -> unit
        (** Clears transformer parameters. *)

    val set_parameter : t -> string -> Cadmium.java_object -> unit
        (** [set_parameter t n v] sets value [v] to parameter named [n]
           for transformer [t]. *)

    val set_parameter_obj : t -> string -> CadmiumObj.jObject -> unit
        (** [set_parameter t n v] sets value [v] to parameter named [n]
           for transformer [t]. *)

    val get_parameter : t -> string -> Cadmium.java_object
        (** [get_parameter t n] gets value of parameter named [n] for transformer [t]. *)

    val get_parameter_obj : t -> string -> CadmiumObj.jObject
        (** [get_parameter t n] gets value of parameter named [n] for transformer [t]. *)

    val set_output_property : t -> string -> string -> unit
        (** [set_output_property t n v] sets value [v] to property named [n]
           for transformer [t]. *)

    val get_output_property : t -> string -> string
        (** [get_output_property t n] gets value of property named [n]
           for transformer [t]. *)

    val transform_file_to_file : t -> string -> string -> unit
        (** [transform_file_to_file t src dst] applies transformer [t]
           to [src], producing [dst]. *)

    val transform_file_to_channel : t -> string -> out_channel -> unit
        (** [transform_file_to_channel t src dst] applies transformer [t]
           to [src], producing [dst]. *)

    val transform_channel_to_file : t -> in_channel -> string -> unit
        (** [transform_channel_to_file t src dst] applies transformer [t]
           to [src], producing [dst]. *)

    val transform_channel_to_channel : t -> in_channel -> out_channel -> unit
        (** [transform_channel_to_channel t src dst] applies transformer [t]
           to [src], producing [dst]. *)

    val reset : t -> unit
        (** Resets the passed transformer. *)
  end
(** Transformers. *)


module XPathFactory :
  sig
    type t
          (** Type of xpath factories. *)

    val make : unit -> t
        (** Constructs an xpath factory. *)

    val set_feature : t -> string -> bool -> unit
        (** [set_feature f n v] sets value [v] to feature named [n] for factory [f]. *)

    val get_feature : t -> string -> bool
        (** [get_feature f n] gets value of feature named [n] for factory [f]. *)
  end
(** XPath factories. *)


module XPath :
  sig
    type t
          (** Type of xpath instances. *)

    type expr
          (** Type of xpath expressions. *)

    val make : XPathFactory.t -> t
        (** Constructs an xpath instance. *)

    val compile : t -> string -> expr
        (** Compiles the passed xpath expression. *)

    val evaluate_file : t -> string -> string -> string
        (** [evaluate_file x e f] evaluate expression [e] in file [f] using [x]. *)

    val evaluate_channel : t -> string -> in_channel -> string
        (** [evaluate_channel x e c] evaluate expression [e] in channel [c] using [x]. *)

    val evaluate_document : t -> string -> Document.document -> string
        (** [evaluate_document x e d] evaluate expression [e] in document [d]
           using [x]. *)

    val expr_evaluate_file : expr -> string -> string
        (** [expr_evaluate_file e f] evaluate expression [e] in file [f]. *)

    val expr_evaluate_channel : expr -> in_channel -> string
        (** [expr_evaluate_channel e c] evaluate expression [e] in channel [c]. *)

    val expr_evaluate_document : expr -> Document.document -> string
        (** [expr_evaluate_document e d] evaluate expression [e] in document [d]. *)

    val reset : t -> unit
        (** Resets the passed xpath instance. *)
  end
(** XPath queries. *)
